package services.pescado.fusionservice.impl;

import Helper.*;
import java.util.*;


/**
 * Stores clocking information about a benchmark.
 * 
 * @author J&uuml;rgen Mo&szlig;graber
 */
public class Benchmark
{
	private final List<Pair<String, Long>>	benches				= new ArrayList<Pair<String, Long>> ();
	private long														currentStart	= System.currentTimeMillis ();


	/**
	 * Store the current bench with the given name and start the next one.
	 * 
	 * @param name A name for the bench
	 */
	public void next (final String name)
	{
		final long end = System.currentTimeMillis ();
		this.benches.add (new Pair<String, Long> (name, Long.valueOf (end - this.currentStart)));
		this.currentStart = end;
	}


	/**
	 * Formats the results as text.
	 * 
	 * @return The formatted text
	 */
	public String printResults ()
	{
		final StringBuilder sb = new StringBuilder ("Benchmark information (in ms):\n");
		long sum = 0;
		for (final Pair<String, Long> bench: this.benches)
		{
			final Long time = bench.getTwo ();
			sb.append (bench.getOne ()).append (": ").append (time).append ('\n');
			sum += time.longValue ();
		}
		sb.append ("Sum: ").append (sum);
		return sb.toString ();
	}
}
