package services.pescado.fusionservice.impl;

import java.io.IOException;
import java.io.StringReader;
import java.sql.Timestamp;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.StringTokenizer;
import java.util.TimeZone;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.jws.WebService;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import net.opengis.gml.AbstractFeatureType;
import net.opengis.gml.CodeType;
import net.opengis.ogc.BinarySpatialOpType;
import net.opengis.ogc.SpatialOperatorNameType;
import net.opengis.om._1.ObservationCollectionType;
import net.opengis.om._1.ObservationPropertyType;
import net.opengis.om._1.ObservationType;
import net.opengis.sos._1.Capabilities;
import net.opengis.sos._1.Contents;
import net.opengis.sos._1.GetObservation.EventTime;
import net.opengis.sos._1.ObservationOfferingType;
import net.sf.json.JSONArray;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import services.pescado.common.PESCaDOException;
import services.pescado.common.PESCaDOExceptionEx;
import services.pescado.common.ParameterType;
import services.pescado.fusionservice.FusionService;
import services.pescado.fusionservice.impl.Masks.Env;
import services.pescado.fusionservice.impl.Masks.MapPack;
//import services.pescado.fusionservice.impl.airM.AirMergeAccess; AIRMERGE disabled
import services.pescado.fusionservice.impl.database.DataBase;
import services.pescado.fusionservice.impl.database.Options;
import services.pescado.fusionservice.impl.datapack.DataPack;
import services.pescado.fusionservice.impl.datapack.Observation;
import services.pescado.fusionservice.impl.fuser.Datetime;
import services.pescado.fusionservice.impl.fuser.Ensemble;
import services.pescado.fusionservice.impl.fuser.FusionResult;
import services.pescado.kbas.KBAccess;
import services.pescado.knowledgebaseaccessservice.KnowledgeBaseAccessService;
import services.pescado.knowledgebaseaccessservice.StatementType;
import Helper.DateAccessor;
import Helper.Pair;
import Helper.io.Properties;
import Helper.net.ProxyInfo;
import de.fraunhofer.iosb.sosclient.GMLHelper;
import de.fraunhofer.iosb.sosclient.OWSException;
import de.fraunhofer.iosb.sosclient.SOSClient;
import de.fraunhofer.iosb.sosclient.ValueAndDuration;
import de.fraunhofer.iosb.sosclient.ValueBlockElement;


/**
 * 
 * 
 * @author J&uuml;rgen Mo&szlig;graber (Fraunhofer IOSB)
 * @author Desiree Hilbring (Fraunhofer IOSB)
 * @author Lasse Johansson (FMI)
 */
@WebService(endpointInterface = "services.pescado.fusionservice.FusionService", wsdlLocation = "FusionService.wsdl", targetNamespace = "http://pescado.services/FusionService/", serviceName = "FusionService", portName = "FusionServiceSOAP")
public class FusionServiceImpl implements FusionService
{
	private final static Logger														LOGGER										= Logger.getLogger ("FS");
	private final static String														SOS_ENDPOINT							= "DataRetrievalService";

	private static final String														XSD_DOUBLE								= "^^xsd:double";
	private static final String														XSD_INTEGER								= "^^xsd:integer";
	private static final String														XSD_DATE_TIME							= "^^xsd:dateTime";
	private ArrayList<String>															fsDump										= new ArrayList<String> ();
	private final static Pattern													DATETIMEPATTERN						= Pattern.compile ("(-?[0-9]{4,})-([0-9]{2})-([0-9]{2})T([0-9]{2}):([0-9]{2}):([0-9]{2})([.]([0-9]{1,3}))?((Z)|(([+]|-)([0-9]{2}):([0-9]{2})))?");
	private final static String														NS_PESCADO								= "http://www.pescado-project.eu/ontology/pescado.owl#";
	private final static String														NS_PESCADO_INSTANCES			= "http://www.pescado-project.eu/ontology/pescadoInstances.owl#";
	private final static String														RDF_TYPE									= "http://www.w3.org/1999/02/22-rdf-syntax-ns#type";

	private final static String														OBSERVED_PROPERTY_EXACT		= ".observed.exact";
	private final static String														OBSERVED_EXACT						= "#observed#exact";
	private final static String														EXACT											= "exact";
	private final static String														FORECASTED								= "http://www.pescado-project.eu/ontology/pescado.owl#forecasted";
	private final static String														OBSERVED									= "http://www.pescado-project.eu/ontology/pescado.owl#observed";
	private final static String														BOTH											= "both";

	private final static String														ENVIRONMENT_OF_INTEREST_QUERY;
	static
	{
		// Retrieve the geo area of the problem
		final StringBuilder q = new StringBuilder ();
		q.append ("PREFIX pescado: <http://www.pescado-project.eu/ontology/pescado.owl#>\n");
		q.append ("select ?env where ");
		q.append ("{ ");
		q.append ("  <%geoarea%> pescado:hasGeoAreaEnvironmentOfInterest ?env . ");
		q.append ("}");
		ENVIRONMENT_OF_INTEREST_QUERY = q.toString ();
	}

	private final static String														GEO_AREA_QUERY;
	static
	{
		// Retrieve the geo area of the problem
		final StringBuilder q = new StringBuilder ();
		q.append ("PREFIX pescado: <http://www.pescado-project.eu/ontology/pescado.owl#>\n");
		q.append ("select ?geoarea where ");
		q.append ("{ ");
		q.append ("  <%problemInstance%> pescado:hasProblemRequest ?r . ");
		q.append ("  ?r pescado:hasRequestGeoArea ?geoarea");
		q.append ("}");
		GEO_AREA_QUERY = q.toString ();
	}
	private final static String														GET_REQUEST_URI_QUERY;
	static
	{
		final StringBuilder q2 = new StringBuilder ();
		q2.append ("PREFIX pescado: <http://www.pescado-project.eu/ontology/pescado.owl#>\n");
		q2.append ("select ?request where ");
		q2.append ("{ ");
		q2.append ("<%problem%> pescado:hasProblemRequest ?request . ");
		q2.append ("}");
		GET_REQUEST_URI_QUERY = q2.toString ();
	}
	private final static String														GET_OLD_FROM_TIME_QUERY;
	static
	{
		final StringBuilder q2 = new StringBuilder ();
		q2.append ("PREFIX pescado: <http://www.pescado-project.eu/ontology/pescado.owl#>\n");
		q2.append ("select ?time where ");
		q2.append ("{ ");
		q2.append ("<%request%> pescado:hasFromDateTime ?time . ");
		q2.append ("}");
		GET_OLD_FROM_TIME_QUERY = q2.toString ();
	}
	private final static String														GET_OLD_TO_TIME_QUERY;
	static
	{
		final StringBuilder q2 = new StringBuilder ();
		q2.append ("PREFIX pescado: <http://www.pescado-project.eu/ontology/pescado.owl#>\n");
		q2.append ("select ?time where ");
		q2.append ("{ ");
		q2.append ("<%request%> pescado:hasToDateTime ?time . ");
		q2.append ("}");
		GET_OLD_TO_TIME_QUERY = q2.toString ();
	}
	private final static String														GET_SUB_TIME_QUERY;
	static
	{
		final StringBuilder q2 = new StringBuilder ();
		q2.append ("PREFIX pescado: <http://www.pescado-project.eu/ontology/pescado.owl#>\n");
		q2.append ("select ?time where ");
		q2.append ("{ ");
		q2.append ("<%request%> pescado:hasRequestSubmissionDateTime ?time . ");
		q2.append ("}");
		GET_SUB_TIME_QUERY = q2.toString ();
	}

	private final static Map<String, Collection<String>>	offerings									= new HashMap<String, Collection<String>> ();
	private static List<String>														serviceIdentificationType	= null;																																																																						// provides
																																																																																																																			// SOS
	private final KBAccess																kba;
	private final Properties															queryProperties						= new Properties ();
	private final SOSClient																sosClient;
	final Properties																			properties								= new Properties ().loadFromClasspath ("/FusionService.properties");

	private final MapPack																	maps;
	public final DataBase																	DB;
	private String fusionStrResult;
	final double minAreaInKM2 = 900;
	

	// ***************************

	/**
	 * Constructor.
	 * 
	 * @throws IOException Could not load the properties file
	 */
	// public FusionServiceImpl (MapPack maps, DataBase DB) throws IOException
	public FusionServiceImpl () throws IOException
	{
		// this.maps = maps;
		// this.DB = DB;

		// Init Fusion Service
		// final Properties properties = new Properties ().loadFromClasspath
		// ("/FusionService.properties");
		this.queryProperties.loadFromClasspath ("/queries.properties");

		// Proxy
		final String proxy = this.properties.getProperty ("LocalProxy");
		final int proxyport = this.properties.getInt ("LocalProxyPort");
		this.sosClient = new SOSClient (this.properties.getProperty (SOS_ENDPOINT), proxy != null && !proxy.equals ("") ? new ProxyInfo (proxy, proxyport) : null);
		this.kba = Services.getKnowledgeBaseAccess ();
		Services.init (this.properties);

		this.retrieveCapabilities ();

		// create mapPack and DB
		java.io.File file = new java.io.File (""); // Dummy file
		String abspath = file.getAbsolutePath ();
		String Z = System.getProperty ("file.separator");
		LOGGER.info ("mapPack initalization...");

		this.maps = MapPack.load (abspath + Z + "maps" + Z + "Finland_mapPack.dat");

		LOGGER.info ("Node database load...");

		this.DB = new DataBase (abspath + Z);

		// mapPack test==========
		double lat = 60.16803; // center of Helsinki
		double lon = 24.941098;

		Env env = this.maps.getEnvironment (lat, lon, Options.profRadius);
		// int pop = maps.getLocalPopulation(lat, lon);

		LOGGER.info ("extended Profile test with lat=" + lat + " lon =" + lon + "\t" + env.getSummary ());
		LOGGER.info ("Local population test...");
		for (int i = 0; i < 10; i++)
		{
			double area = 2.0 * i + 1;
			int pop2 = this.maps.getLocalPopulation (lat, lon, area);
			LOGGER.info (pop2 + "\t" + area + " km2");
		}
	}


	/** {@inheritDoc} */
	@Override
	public String fuseData (final String ontologyIdentifier, final String problemURI, final List<ParameterType> parameter, final List<ParameterType> racsParameters) throws PESCaDOException
	{
		LOGGER.info ("[FS - request");
		fusionStrResult = "";
        fsDump = new ArrayList<String>();
		
		synchronized (FusionServiceImpl.offerings)
		{
			if (FusionServiceImpl.offerings.isEmpty ())
				throw PESCaDOExceptionEx.with ("FS not yet initialised.", "", "FS");
		}

		final Iterator<ParameterType> paramIterator = parameter.iterator ();
		while (paramIterator.hasNext ())
		{
			final ParameterType paramtype = paramIterator.next ();
			final String pproperty = paramtype.getProperty ();
			LOGGER.info ("ppropety " + pproperty);
		}
		final Iterator<ParameterType> rparamIterator = racsParameters.iterator ();
		while (rparamIterator.hasNext ())
		{
			final ParameterType rparamtype = rparamIterator.next ();
			final String rpproperty = rparamtype.getProperty ();
			LOGGER.info ("rppropety " + rpproperty);
		}

		Parameters params = null;
		// population count evaluation and storing
		try
		{
			params = new Parameters (parameter);

			int polygonPop = this.getPolygonPopulation (params);
			LOGGER.info ("Population inside the polygon was estimated to be " + polygonPop + ".");
			fsDump.add ("Population inside the polygon was estimated to be " + polygonPop + ".");
			this.storePopulationInfo (polygonPop, ontologyIdentifier, problemURI);
		}
		catch (IOException e)
		{
			System.out.println ("IOE: Could not store population count with geoArea.");
			// e.printStackTrace ();
		}
		catch (NullPointerException exp)
		{
			System.out.println ("NullPointer: Could not store population count with params coordinates.");
		}

		catch (final ParseException ex)
		{
			LOGGER.log (Level.SEVERE, ex.getMessage (), ex);
			throw PESCaDOExceptionEx.with ("Exception raised during retrieving data of DRS or storing data into KBAS", "", "FS", ex);
		}

		// Find out if data is observed, forecasted ...
		final String environmentalDataNature = this.lookForEnvironmentalDataNature (ontologyIdentifier, problemURI);

		if (!environmentalDataNature.equals ("both"))
		{
			doFusion (ontologyIdentifier, problemURI, racsParameters, params, environmentalDataNature, new Pair<String, String> ("false", "false"));
		}
		else
		{
			try
			{
				LOGGER.info ("FS: Push Use Case: Calculate time ranges for forecasted and observed!");
				Parameters params1 = new Parameters (parameter);
				params1.setUseCase ("DetailedAirQuality");

				// Add information for DetailedAirQuality to ontology
				final String requestResult = this.kba.executeQuery (ontologyIdentifier, GET_REQUEST_URI_QUERY.replace ("%problem%", problemURI));
				final JSONArray jsonArray = this.kba.getJSONBinding (requestResult);
				String requestNode = extractStringProperty ("request", jsonArray);
				LOGGER.info ("requestNode " + requestNode);
				ArrayList<StatementType> statements2 = new ArrayList<StatementType> ();
				Calendar currentTimeC = Calendar.getInstance ();
				currentTimeC.setTimeZone (TimeZone.getDefault ());

				final String requestSubTime = this.kba.executeQuery (ontologyIdentifier, GET_SUB_TIME_QUERY.replace ("%request%", requestNode));
				final JSONArray jsonArray4 = this.kba.getJSONBinding (requestSubTime);
				String subTime = extractStringProperty ("time", jsonArray4);
				Calendar cFrom = DateAccessor.parseISODateTimeWithTimezone (subTime);
				Calendar cTo = DateAccessor.parseISODateTimeWithTimezone (subTime);
				LOGGER.info ("subTime " + subTime);

				LOGGER.info ("PASTHOURS " + params1.getHoursInPast () + "FUTUREHOURS " + params1.getHoursInFuture ());
				int ph = params1.getHoursInPast ();
				int fh = params1.getHoursInFuture ();
				if (ph == 0)
				{
					ph = 24;
					params1.setHoursInPast (ph);
				}
				if (fh == 0)
				{
					fh = 12;
					params1.setHoursInFuture (fh);
				}
				cFrom.add (Calendar.HOUR_OF_DAY, -ph);
				// go back to beginning of last day
				if (ph == 24)
				{
					cFrom.set (Calendar.HOUR_OF_DAY, 0);
					cFrom.set (Calendar.MINUTE, 0);
					cFrom.set (Calendar.SECOND, 0);
					cFrom.set (Calendar.MILLISECOND, 0);
				}
				cFrom.setTimeZone (TimeZone.getDefault ());
				String fromTime = DateAccessor.formatISODateTimeWithTimezone (cFrom);

				cTo.add (Calendar.HOUR_OF_DAY, fh);
				cTo.setTimeZone (TimeZone.getDefault ());
				String toTime = DateAccessor.formatISODateTimeWithTimezone (cTo);

				final String requestOldFromTime = this.kba.executeQuery (ontologyIdentifier, GET_OLD_FROM_TIME_QUERY.replace ("%request%", requestNode));
				final JSONArray jsonArray2 = this.kba.getJSONBinding (requestOldFromTime);
				String oldFromTime = extractStringProperty ("time", jsonArray2);

				final String requestOldToTime = this.kba.executeQuery (ontologyIdentifier, GET_OLD_TO_TIME_QUERY.replace ("%request%", requestNode));
				final JSONArray jsonArray3 = this.kba.getJSONBinding (requestOldToTime);
				String oldToTime = extractStringProperty ("time", jsonArray3);
				LOGGER.info ("oldTo " + oldToTime + " oldFrom " + oldFromTime);

				// this.kba.getService ().dumpDebugInfo (ontologyIdentifier,
				// "before.owl");
				this.kba.getService ().removeABoxStatement (ontologyIdentifier, requestNode, NS_PESCADO + "hasFromDateTime", oldFromTime + XSD_DATE_TIME);
				this.kba.getService ().removeABoxStatement (ontologyIdentifier, requestNode, NS_PESCADO + "hasToDateTime", oldToTime + XSD_DATE_TIME);
				// this.kba.getService ().dumpDebugInfo (ontologyIdentifier,
				// "remove.owl");
				addStatementToList (statements2, requestNode, NS_PESCADO + "hasFromDateTime", fromTime + XSD_DATE_TIME);
				addStatementToList (statements2, requestNode, NS_PESCADO + "hasToDateTime", toTime + XSD_DATE_TIME);

				this.kba.getService ().addABoxStatements (ontologyIdentifier, statements2);
				// this.kba.getService ().dumpDebugInfo (ontologyIdentifier,
				// "added.owl");
				LOGGER.info ("Scheduler: submission time: " + subTime);
				// Add information for DetailedAirQuality to ontology end

				LOGGER.info ("FIRST FUSION");
				params1.setStartDateTime (fromTime);
				params1.setEndDateTime (subTime);
				Pair<String, String> rootNodes = doFusion (ontologyIdentifier, problemURI, racsParameters, params1, OBSERVED, new Pair<String, String> ("false", "false"));
				LOGGER.info ("FIRST FUSION END");

				Parameters params2 = new Parameters (parameter);
				params2.setUseCase ("DetailedAirQuality");
				LOGGER.info ("PASTHOURS " + params2.getHoursInPast () + "FUTUREHOURS " + params2.getHoursInFuture ());
				int ph2 = params2.getHoursInPast ();
				int fh2 = params2.getHoursInFuture ();
				if (ph2 == 0)
				{
					ph2 = 24;
					params2.setHoursInPast (ph2);
				}
				if (fh2 == 0)
				{
					fh2 = 12;
					params2.setHoursInFuture (fh2);
				}

				LOGGER.info ("SECOND FUSION");
				params2.setStartDateTime (subTime);
				params2.setEndDateTime (toTime);
				doFusion (ontologyIdentifier, problemURI, racsParameters, params2, FORECASTED, rootNodes);
				LOGGER.info ("SECOND FUSION END");
				// this.kba.getService ().addABoxStatements (ontologyIdentifier,
				// statements2);
				LOGGER.info ("FS: Push Use Case: Fusion Done!");

			}
			catch (ParseException e)
			{
				LOGGER.log (Level.SEVERE, e.getMessage (), e);
				throw PESCaDOExceptionEx.with ("Exception raised parsing FS parameters.", "", "FS", e);
			}
			catch (IOException e)
			{
				LOGGER.log (Level.SEVERE, e.getMessage (), e);
				throw PESCaDOExceptionEx.with ("Exception raised looking for request node in ontology.", "", "FS", e);
			}
		}

		final KBAccess kba = Services.getKnowledgeBaseAccess ();
		final KnowledgeBaseAccessService kbas = kba.getService ();
		String dump = this.properties.getProperty ("dumpOntology");
		if (dump.equals ("true"))
			kbas.dumpDebugInfo (ontologyIdentifier, "dumpAfterFSFinished.owl");

		
		return fusionStrResult;
	}


	/**
	 * Do the fusion: Retrieve data, fuse it and store it to the ontology.
	 * 
	 * @param ontologyIdentifier identifier
	 * @param problemURI problem
	 * @param racsParameters parameters
	 * @param params parameters
	 * @param environmentalDataNature observed or forecasted
	 * @param rootNodes
	 * @param environmentalDataNature2
	 * @return
	 * @throws PESCaDOException exception
	 */
	private Pair<String, String> doFusion (final String ontologyIdentifier, final String problemURI, final List<ParameterType> racsParameters, Parameters params, final String environmentalDataNature, Pair<String, String> rootNodes) throws PESCaDOException
	{
		// Get the data from the DRS and store it to the ontology
		try
		{
			final Benchmark benchmark = new Benchmark ();

			// LOGGER.info ("[FS - Retrieve data from DRS");
			final Map<String, ObservationCollectionType> data = this.retrieveData (params, racsParameters, environmentalDataNature);
			// LOGGER.info ("Time FS - retrieve data: " + (retrieveEnd -
			// retrieveStart));
			benchmark.next ("retrieveData");

			// Fuse the data
			LOGGER.info ("Fusing data - creating ensemble");

			LOGGER.info ("Parsing retrieved data");
			String [][] parsed_data = this.parse_retrieveData (data); // {{type,ID,timestamp,value},{}...}
			benchmark.next ("parseRetrievedData");

			/*
			 * This method calculates fusion for all data types and timestamps that
			 * are given by parsed_data array.
			 * 
			 * @params: use_two enables causes the fusion service to use two values
			 * for all sources.This reduces imprecision but smoothens the data
			 * provided at the same time.
			 * 
			 * @returns: ArrayList<{type,timestamp,value,imprecisionSTD,...}>
			 */

			ArrayList<FusionResult> fusionResult = this.calculateFusion (parsed_data, params, ontologyIdentifier, problemURI);
			fusionStrResult = this.getFusionResultListForUI(fusionResult);
			benchmark.next ("calculateFusion");

			// final long fuseEnd = System.currentTimeMillis ();
			// LOGGER.info ("Time FS - fuse data: " + (fuseEnd - fuseStart));

			String fusedNode = this.storeFusedData (ontologyIdentifier, problemURI, fusionResult, environmentalDataNature, racsParameters, rootNodes.getOne ());
			benchmark.next ("storeFusedData");
			String exceedanceNode = this.storeStationData (ontologyIdentifier, problemURI, parsed_data, environmentalDataNature, racsParameters, rootNodes.getTwo ());
			benchmark.next ("storeStationData");
			Pair<String, String> nodeRoots = new Pair<String, String> (fusedNode, exceedanceNode);

			// TODO this method should be modified to utilize parsed_data to get
			// a performance boost

			LOGGER.info (benchmark.printResults ());

			// final long storeEnd = System.currentTimeMillis ();
			// LOGGER.info ("Time FS - store data: " + (storeEnd - storeStart));
			//
			// LOGGER.info ("Times needed in FS: ");
			// LOGGER.info ("Time FS - retrieve data: " + (retrieveEnd -
			// retrieveStart));
			// LOGGER.info ("Time FS - fuse data: " + (fuseEnd - fuseStart));
			// LOGGER.info ("Time FS - store data: " + (storeEnd - storeStart));
			// LOGGER.info ("Done.");

			return nodeRoots;
		}
		catch (final Throwable t)
		{
			LOGGER.log (Level.SEVERE, t.getMessage (), t);
			throw PESCaDOExceptionEx.with ("Exception raised during retrieving data of DRS or storing data into KBAS", "", "FS", t);
		}

	}


	// Stores all station VALUES for exeedance evaluation
	private void storeExceedanceDataOfStation (final String ontologyIdentifier, final KnowledgeBaseAccessService kbas, final String [][] parsed_data, final String geoArea, final String exceedanceNodeURI, final String environmentalDataNature, List<ParameterType> racsParameters) throws PESCaDOException
	{
		ArrayList<StatementType> statements = new ArrayList<StatementType> ();

		final List<String> missingArray = createObservedPropertyArray (racsParameters);

		for (int i = 0; i < parsed_data.length; i++)
		{
			String [] pElement = parsed_data[i];
			String observedProperty = NS_PESCADO + pElement[0];
			String time = pElement[2];
			String value = pElement[3];
			String opDataType = pElement[4];
			// duration is still needed as 5th element in the String blobs
			String duration = pElement[5];
			if (!(duration.equals ("00:00:59:59") || duration.equals ("00:01:00:00") || duration.equals ("00:07:59:59") || duration.equals ("00:08:00:00") || duration.equals ("00:23:59:59") || duration.equals ("01:00:00:00")))
			{
				LOGGER.info ("Duration other that 1,8 or 24 hours: " + duration);
				LOGGER.info ("PESCaDO cannot handle durations which differ from 1,8 or 24 hours!");

				continue;
			}

			if (value.contains ("noData") || value.contains ("-"))
			{
				// LOGGER.severe ("Retrieved value contains noData or negative value.");
				// LOGGER.severe ("Information will be stored in ontology");
				// final String missingDataNode = this.kba.createIndividualURI
				// (NS_PESCADO_INSTANCES, "MissingNode");
				//
				// addStatementToList (statements, missingDataNode, RDF_TYPE, NS_PESCADO
				// + "MissingData");
				// addStatementToList (statements, missingDataNode, NS_PESCADO +
				// "hasConclusionWeight", "1.0^^xsd:double");
				// addStatementToList (statements, missingDataNode, NS_PESCADO +
				// "hasMissingDataEnvironmentalDataType", NS_PESCADO +
				// observedProperty);
				//
				// LOGGER.info ("In store fusion: environmentalDataNature: " +
				// environmentalDataNature);
				// addStatementToList (statements, missingDataNode, NS_PESCADO +
				// "hasMissingDataEnvironmentalDataNature", environmentalDataNature);
				// // Add missing node to geoArea node
				// addStatementToList (statements, geoArea, NS_PESCADO +
				// "hasGeoAreaConclusion", missingDataNode);

				continue;
			}
			if (opDataType.contains ("AirPollutantValue"))
			{
				if (missingArray.contains (observedProperty))
					missingArray.remove (observedProperty);
				final String environmentalDataURI = this.kba.createIndividualURI (NS_PESCADO_INSTANCES, "EnvironmentalData");
				addStatementToList (statements, environmentalDataURI, RDF_TYPE, NS_PESCADO + "EnvironmentalData");
				addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasEnvironmentalDataEnvironmentalDataType", observedProperty);
				addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasEnvironmentalDataNature", environmentalDataNature);

				// Connect the environmental data to the exceedance node
				addStatementToList (statements, exceedanceNodeURI, NS_PESCADO + "hasEnvironmentalNodeEnvironmentalData", environmentalDataURI);

				// Information from Marco
				// - O3: 1h and 8h hour averages
				// - CO: 8h hour averages
				// - NO2: 1h hour averages
				// - SO2: 1h and 24h hour averages
				// - PM10: 24h hour averages
				// if (observedProperty.contains("O3"))
				// {
				//
				// } else if (observedProperty.contains("CO"))
				// {
				//
				// } else if (observedProperty.contains("NO2"))
				// {
				//
				// }

				addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasFromDateTime", time + XSD_DATE_TIME);
				addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasToDateTime", incTimeWithDuration (time, duration) + XSD_DATE_TIME);

				final String valueURI = this.kba.createIndividualURI (NS_PESCADO_INSTANCES, opDataType);
				addStatementToList (statements, valueURI, RDF_TYPE, NS_PESCADO + opDataType);
				String v = value + XSD_DOUBLE;
				LOGGER.fine ("v " + v);

				addStatementToList (statements, valueURI, NS_PESCADO + "hasValueValue", v);

				if (duration.equals ("00:00:59:59"))
					addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasAirPollutantTimeInterval", NS_PESCADO + "timeInterval1h");
				else if (duration.equals ("00:01:00:00"))
					addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasAirPollutantTimeInterval", NS_PESCADO + "timeInterval1h");
				else if (duration.equals ("00:07:59:59"))
					addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasAirPollutantTimeInterval", NS_PESCADO + "timeInterval8h");
				else if (duration.equals ("00:08:00:00"))
					addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasAirPollutantTimeInterval", NS_PESCADO + "timeInterval8h");
				else if (duration.equals ("00:23:59:59"))
					addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasAirPollutantTimeInterval", NS_PESCADO + "timeInterval24h");
				else if (duration.equals ("01:00:00:00"))
					addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasAirPollutantTimeInterval", NS_PESCADO + "timeInterval24h");
				addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "microgramPerMeterCubed");

				// Link it to the environmentaldata
				addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasEnvironmentalDataValue", valueURI);
			}

		}

		// Add missing nodes
		// fusion already added missing nodes! delete it here!
		// LOGGER.info ("Number of MissingNode: " + missingArray.size ());
		// for (int j = 0; j < missingArray.size (); j++)
		// {
		// LOGGER.info ("Add MissingNode for " + missingArray.get (j));
		// final String missingDataNode = this.kba.createIndividualURI
		// (NS_PESCADO_INSTANCES, "MissingNode");
		// addStatementToList (statements, missingDataNode, RDF_TYPE, NS_PESCADO +
		// "MissingData");
		// addStatementToList (statements, missingDataNode, NS_PESCADO +
		// "hasConclusionWeight", "1.0^^xsd:double");
		// addStatementToList (statements, missingDataNode, NS_PESCADO +
		// "hasMissingDataEnvironmentalDataType", NS_PESCADO + missingArray.get
		// (j));
		// LOGGER.info ("In store fusion: environmentalDataNature: " +
		// environmentalDataNature);
		// addStatementToList (statements, missingDataNode, NS_PESCADO +
		// "hasMissingDataEnvironmentalDataNature", environmentalDataNature);
		//
		// // Add missing node to geoArea node
		// addStatementToList (statements, geoArea, NS_PESCADO +
		// "hasGeoAreaConclusion", missingDataNode);
		// }

		// Add collected statements to ontology
		kbas.addABoxStatements (ontologyIdentifier, statements);
	}


	// Stores all un-fused input data, including values which are stored also in
	// storeExceedanceDataOfStation
	private String storeStationData (String ontologyIdentifier, String problemURI, String [][] parsed_data, String environmentalDataNature, List<ParameterType> racsParameters, String stationNode) throws IOException, PESCaDOException
	{

		String exceedanceNodeURI = "false";

		if (parsed_data == null || parsed_data.length == 0)
			return exceedanceNodeURI;

		final KBAccess kba = Services.getKnowledgeBaseAccess ();
		final KnowledgeBaseAccessService kbas = kba.getService ();

		String dump = this.properties.getProperty ("dumpOntology");
		if (dump.equals ("true"))
			kbas.dumpDebugInfo (ontologyIdentifier, "storeStationDataBefore.owl");

		// Link the geo area from the request to the node, first retrieve the
		// geoarea

		final String result = kba.executeQuery (ontologyIdentifier, GEO_AREA_QUERY.replace ("%problemInstance%", problemURI));
		LOGGER.info (" problemURI " + problemURI);
		LOGGER.info ("result " + result);

		// Exception Test
		final JSONArray jsonArray = kba.getJSONBinding (result);
		if (jsonArray.isEmpty ())
		{
			final String msg = "Geo reference not found for problem " + problemURI;
			LOGGER.severe ("        " + msg);
			throw new RuntimeException (msg);
		}
		final String geoArea = jsonArray.getJSONObject (0).getJSONObject ("geoarea").getString ("value");

		// Store all data sets
		if (this.lookForRequestType (ontologyIdentifier, problemURI))
		{
			// LOGGER.info ("Task: CheckAirQualityLimits!");
			if (stationNode.equals ("false"))
			{
				exceedanceNodeURI = this.createExceedanceEnvironmentalNode (ontologyIdentifier, kbas, parsed_data, geoArea);
			}
			else
			{
				exceedanceNodeURI = stationNode;
			}

			if (!exceedanceNodeURI.equals ("false"))
			{

				this.storeExceedanceDataOfStation (ontologyIdentifier, kbas, parsed_data, geoArea, exceedanceNodeURI, environmentalDataNature, racsParameters);

			}
		}

		if (dump.equals ("true"))
			kbas.dumpDebugInfo (ontologyIdentifier, "storeStationDataAfter.owl");

		return exceedanceNodeURI;
	}


	// Stores all fused data and missing nodes under the node 'fused'
	private String storeFusedData (String ontologyIdentifier, String problemURI, ArrayList<FusionResult> fusionResult, String environmentalDataNature, List<ParameterType> racsParameters, String fusedNode) throws IOException
	{
		// fusionResult contains: [0] type, [1] timestamp, [2] value, [3],
		// imprecision [4], WEIGHT?, [5] opDataType

		if (fusionResult == null || fusionResult.isEmpty ())
			return "false";

		// List for which data was asked for
		final List<String> missingArray = createObservedPropertyArray (racsParameters);

		final KBAccess kba = Services.getKnowledgeBaseAccess ();
		final KnowledgeBaseAccessService kbas = kba.getService ();

		String dump = this.properties.getProperty ("dumpOntology");
		if (dump.equals ("true"))
			kbas.dumpDebugInfo (ontologyIdentifier, "storeFusedDataBefore.owl");

		// Link the geo area from the request to the node, first retrieve the
		// geoarea
		final String result = kba.executeQuery (ontologyIdentifier, GEO_AREA_QUERY.replace ("%problemInstance%", problemURI));
		LOGGER.info (" problemURI " + problemURI);
		LOGGER.info (" result " + result);

		// Exception Test
		final JSONArray jsonArray = kba.getJSONBinding (result);
		if (jsonArray.isEmpty ())
		{
			final String msg = "Geo reference not found for problem " + problemURI;
			LOGGER.severe ("        " + msg);
			throw new RuntimeException (msg);
		}
		final String geoArea = jsonArray.getJSONObject (0).getJSONObject ("geoarea").getString ("value");

		// Add single fusion node
		ArrayList<StatementType> statements = new ArrayList<StatementType> ();
		String fusionNodeURI;
		if (fusedNode.equals ("false"))
			fusionNodeURI = kba.createIndividualURI (NS_PESCADO_INSTANCES, "fusedNode");
		else
			fusionNodeURI = fusedNode;
		addStatementToList (statements, fusionNodeURI, RDF_TYPE, NS_PESCADO + "EnvironmentalNode");
		addStatementToList (statements, fusionNodeURI, NS_PESCADO + "hasEnvironmentalNodeConfidenceValue", "1.0^^xsd:double");
		addStatementToList (statements, fusionNodeURI, NS_PESCADO + "hasEnvironmentalNodeName", "Fused Node^^xsd:string");
		addStatementToList (statements, fusionNodeURI, NS_PESCADO + "hasEnvironmentalNodeForm", NS_PESCADO + "fusedEnvironmentalNode");

		// Add fusion node to geoArea node
		addStatementToList (statements, geoArea, NS_PESCADO + "hasGeoAreaEnvironmentalNode", fusionNodeURI);
		kbas.addABoxStatements (ontologyIdentifier, statements);

		// Store all data sets
		statements = new ArrayList<StatementType> ();
		for (int i = 0; i < fusionResult.size (); i++)
		{
			FusionResult fr = fusionResult.get (i);
			String observedProperty = NS_PESCADO + fr.getType ();
			String startTime = fr.getDt ().getStringDate ();

			// if data for an observedProperty exist remove it from missing list
			final String [] oPArray = observedProperty.split ("#");
			// Remove from missingData list if data available
			if (missingArray.contains (oPArray[1]))
				missingArray.remove (oPArray[1]);

			final String environmentalDataURI = kba.createIndividualURI (NS_PESCADO_INSTANCES, "EnvironmentalData");
			addStatementToList (statements, environmentalDataURI, RDF_TYPE, NS_PESCADO + "EnvironmentalData");
			addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasEnvironmentalDataEnvironmentalDataType", observedProperty);
			addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasEnvironmentalDataNature", environmentalDataNature);

			// Connect the environmental data to the fusion node
			addStatementToList (statements, fusionNodeURI, NS_PESCADO + "hasEnvironmentalNodeEnvironmentalData", environmentalDataURI);

			addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasFromDateTime", startTime + XSD_DATE_TIME);
			// System.out.println("startTime " + startTime);
			addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasToDateTime", incTimeWithDuration (startTime, "00:01:00:00") + XSD_DATE_TIME);

			String opDataType = fr.getOpDataType ();

			// Code for 2nd prototype
			// TODO Rating or index values are not considered until now
			if (opDataType.contains ("Value")) // This should ALWAYS happen
			{
				// test to see if KBAS access only WindDirection insteadof windDirection
				if (opDataType.equals ("windDirectionValue"))
					opDataType = "WindDirectionValue";
				else if (opDataType.equals ("rainValue"))
					opDataType = "RainValue";
				else if (opDataType.equals ("airPollutantValue"))
					opDataType = "AirPollutantValue";
				else if (opDataType.equals ("windSpeedValue"))
					opDataType = "WindSpeedValue";
				else if (opDataType.equals ("temperatureValue"))
					opDataType = "TemperatureValue";
				else if (opDataType.equals ("skyConditionValue"))
					opDataType = "SkyConditionValue";
				else if (opDataType.equals ("pollenValue"))
					opDataType = "PollenValue";
				else if (opDataType.equals ("humidityValue"))
					opDataType = "HumidityValue";
				else if (opDataType.equals ("pressureValue"))
					opDataType = "PressureValue";

				String v = fr.getValueString ();
				double valueDouble = 0.0;
				int valueInt = 0;

				try
				{
					valueDouble = fr.getValue ();
					valueInt = (int) (valueDouble);

				}
				catch (NumberFormatException ex)
				{ // this should never happen!
					LOGGER.severe ("Invalid fusion value: " + v);
					continue;
				}

				if (opDataType.equals ("WindSpeedValue") && (valueDouble < 0)) // negative
																																				// values
																																				// discarded
				{
					LOGGER.severe ("Invalid WindSpeedValue: " + v);
					continue;
				}
				else if (opDataType.equals ("WindDirectionValue") && (valueDouble < 0 || valueDouble > 360)) // must
																																																			// be
																																																			// 0
																																																			// -360
																																																			// degrees
				{

					LOGGER.severe ("Invalid WindDirectionValue: " + v);
					continue;

				}
				else if (opDataType.equals ("RainValue") && (valueDouble < 0)) // negative
																																				// values
																																				// discarded
				{
					LOGGER.severe ("Invalid RainValue: " + v);
					continue;
				}
				else if (opDataType.equals ("AirPollutantValue") && (valueDouble < 0)) // negative
																																								// values
																																								// discarded
				{
					LOGGER.severe ("Invalid AirPollutantValue: " + v);
					continue;
				}
				else if (opDataType.equals ("SkyConditionValue") && (valueDouble < 0 || valueDouble > 9)) // must
																																																	// be
																																																	// 0-9
				{

					LOGGER.severe ("Invalid SkyConditionValue: " + v);
					continue;

				}
				else if (opDataType.equals ("PollenValue") && (valueDouble < 0)) // negative
																																					// values
																																					// discarded
				{
					LOGGER.severe ("Invalid PollenValue: " + v);
					continue;
				}
				else if (opDataType.equals ("PressureValue") && (valueDouble < 0)) // negative
																																						// values
																																						// discarded
				{
					LOGGER.severe ("Invalid PressureValue: " + v);
					continue;
				}
				else if (opDataType.equals ("HumidityValue") && (valueDouble < 0 || valueDouble > 100)) // must
																																																// be
																																																// 0-100
				{

					LOGGER.severe ("Invalid Humidity: " + v);
					continue;

				}

				final String valueURI = kba.createIndividualURI (NS_PESCADO_INSTANCES, opDataType);
				addStatementToList (statements, valueURI, RDF_TYPE, NS_PESCADO + opDataType);

				// System.out.println("Test value: " + v);
				// String v = value.toString ();

				if (opDataType.contains ("WindDirection") || opDataType.contains ("UVIndex") || opDataType.contains ("SkyCondition"))
				{

					v = valueInt + XSD_INTEGER;
				}
				else
					v = valueDouble + XSD_DOUBLE;

				LOGGER.info ("Fused storeValue! " + valueURI + " value " + v + " opDataType " + opDataType + "startTime " + startTime + " plus duration " + incTimeWithDuration (startTime, "00:01:00:00") + " duration " + fr.getDuration ());
				addStatementToList (statements, valueURI, NS_PESCADO + "hasValueValue", v);
				// test for unit
				if (opDataType.contains ("AirPollutantValue"))
				{
					addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasAirPollutantTimeInterval", NS_PESCADO + "timeInterval1h");
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "microgramPerMeterCubed");
				}
				else if (opDataType.contains ("WindSpeedValue"))
				{
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "meterPerSecond");
				}
				else if (opDataType.contains ("WindDirectionValue"))
				{
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "degree");
				}
				else if (opDataType.contains ("TemperatureValue"))
				{
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "degreeC");
				}
				else if (opDataType.contains ("RainValue"))
				{
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "millimeterPerHour");
				}
				else if (opDataType.contains ("SkyConditionValue"))
				{
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "eights");
				}
				else if (opDataType.contains ("PollenValue"))
				{
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "grainPerMeterCubed");
				}
				else if (opDataType.contains ("HumidityValue"))
				{
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "percent");
				}
				else if (opDataType.contains ("PressureValue"))
				{
					addStatementToList (statements, valueURI, NS_PESCADO + "hasUnit", NS_PESCADO + "hectoPascal");
				}
				// Link it to the environmentaldata
				addStatementToList (statements, environmentalDataURI, NS_PESCADO + "hasEnvironmentalDataValue", valueURI);
				continue;
			}
			// Add missing nodes
			// Why are the missing nodes added for each fusion result?
			// hil: try to remove them
			// LOGGER.info ("Number of MissingNode: " + missingArray.size ());
			// for (int j = 0; j < missingArray.size (); j++)
			// {
			// LOGGER.info ("Add MissingNode for " + missingArray.get (j));
			// final String missingDataNode = kba.createIndividualURI
			// (NS_PESCADO_INSTANCES, "MissingNode");
			// addStatementToList (statements, missingDataNode, RDF_TYPE, NS_PESCADO +
			// "MissingData");
			// addStatementToList (statements, missingDataNode, NS_PESCADO +
			// "hasConclusionWeight", "1.0^^xsd:double");
			// addStatementToList (statements, missingDataNode, NS_PESCADO +
			// "hasMissingDataEnvironmentalDataType", NS_PESCADO + missingArray.get
			// (j));
			// LOGGER.info ("In store fusion: environmentalDataNature: " +
			// environmentalDataNature);
			// addStatementToList (statements, missingDataNode, NS_PESCADO +
			// "hasMissingDataEnvironmentalDataNature", environmentalDataNature);
			//
			// // Add missing node to geoArea node
			// addStatementToList (statements, geoArea, NS_PESCADO +
			// "hasGeoAreaConclusion", missingDataNode);
			// }
			// kbas.addABoxStatements (ontologyIdentifier, statements);
			// if (dump.equals ("true"))
			// kbas.dumpDebugInfo (ontologyIdentifier, "storeDataAfter.owl");

		}

		// Add missing nodes
		LOGGER.info ("Number of MissingNode: " + missingArray.size ());
		for (int j = 0; j < missingArray.size (); j++)
		{
			LOGGER.info ("Add MissingNode for " + missingArray.get (j));
			final String missingDataNode = kba.createIndividualURI (NS_PESCADO_INSTANCES, "MissingNode");
			addStatementToList (statements, missingDataNode, RDF_TYPE, NS_PESCADO + "MissingData");
			addStatementToList (statements, missingDataNode, NS_PESCADO + "hasConclusionWeight", "1.0^^xsd:double");
			addStatementToList (statements, missingDataNode, NS_PESCADO + "hasMissingDataEnvironmentalDataType", NS_PESCADO + missingArray.get (j));
			LOGGER.info ("In store fusion alternative: environmentalDataNature: " + environmentalDataNature);
			addStatementToList (statements, missingDataNode, NS_PESCADO + "hasMissingDataEnvironmentalDataNature", environmentalDataNature);

			// Add missing node to geoArea node
			addStatementToList (statements, geoArea, NS_PESCADO + "hasGeoAreaConclusion", missingDataNode);
		}
		kbas.addABoxStatements (ontologyIdentifier, statements);
		if (dump.equals ("true"))
			kbas.dumpDebugInfo (ontologyIdentifier, "storeFusedDataAfter.owl");
		return fusionNodeURI;
	}


	/**
	 * Look in knowledge base which type of environmental data is requested:
	 * forecasted or observed.
	 * 
	 * @param ontologyIdentifier ontologyIdentifier for ontology version
	 * @param problemURI problem uri identifying the request
	 * @return String forecasted or observed
	 * @throws PESCaDOException
	 */
	private String lookForEnvironmentalDataNature (final String ontologyIdentifier, final String problemURI) throws PESCaDOException
	{
		try
		{
			String query = (String) this.queryProperties.get ("selectProblemRequest");
			final String problemURIForQuery = problemURI.split ("#")[1];
			query = query.replace ("problemURI", problemURIForQuery);
			final String resultQuery = this.kba.executeQuery (ontologyIdentifier, query);
			LOGGER.info ("Problem: " + resultQuery);
			final JSONArray jsonArray = this.kba.getJSONBinding (resultQuery);
			if (jsonArray.isEmpty ())
			{
				final String msg = "Property " + resultQuery + " not found!";
				LOGGER.severe ("        " + msg);
				return null;
			}
			// Fix added 2012-07-17
			String requestURI = jsonArray.getJSONObject (0).getJSONObject ("request").getString ("value");
			String query2 = (String) this.queryProperties.get ("selectEnvironmentalDataNature");
			final String requestURIForQuery = requestURI.split ("#")[1];
			LOGGER.info ("requestURIForQuery" + requestURIForQuery);
			query2 = query2.replace ("requestURI", requestURIForQuery);
			LOGGER.info ("Query2: " + query2);
			final String resultQuery2 = this.kba.executeQuery (ontologyIdentifier, query2);
			LOGGER.info ("Request: " + resultQuery2);
			final JSONArray jsonArray2 = this.kba.getJSONBinding (resultQuery2);
			String dataNatureElement = "";
			if (jsonArray2.isEmpty ())
			{
				// TODO FIX for Final Demonstrator!
				// Hi Lasse:
				// For the new scenario for the final demonstrator, no
				// pescado:hasRequestEnvironmentalDataNature exists.
				// Observed and forecasted data must be considered in the FS.
				// use the following data:
				// observed from StartDate until submission date
				// forecasted from submission date to EndDate
				// The following is just a workaround to test, if the scheduler
				// works.
				LOGGER.info ("PUSH USE CASE!" + dataNatureElement);
				dataNatureElement = BOTH;
				return dataNatureElement;
			}

			// Fix end

			dataNatureElement = jsonArray2.getJSONObject (0).getJSONObject ("dataNature").getString ("value");
			LOGGER.info ("data nature element: " + dataNatureElement);
			return dataNatureElement;
		}
		catch (final IOException e)
		{
			e.printStackTrace ();
		}
		return null;
	}


	/**
	 * Returns true, if request type in ontology contains "CheckAirQualityLimits".
	 * 
	 * @param ontologyIdentifier ontologyIdentifier for ontology version
	 * @param problemURI problem uri identifying the request
	 * @return boolean, true, if request type in ontology contains
	 *         "CheckAirQualityLimits".
	 * @throws PESCaDOException
	 */
	private boolean lookForRequestType (final String ontologyIdentifier, final String problemURI) throws PESCaDOException
	{
		try
		{
			String query = (String) this.queryProperties.get ("selectProblemRequest");
			final String problemURIForQuery = problemURI.split ("#")[1];
			query = query.replace ("problemURI", problemURIForQuery);
			final String resultQuery = this.kba.executeQuery (ontologyIdentifier, query);
			final String scndpart = resultQuery.substring (resultQuery.indexOf ("value"));
			final String [] resultArray = scndpart.split ("\"");
			if (resultArray.length <= 2)
				throw PESCaDOExceptionEx.with ("Error checking ontology for request URI", "", "FS", null);

			final String requestURI = resultArray[2];
			String query2 = (String) this.queryProperties.get ("selectRequestType");
			final String requestURIForQuery = requestURI.split ("#")[1];
			query2 = query2.replace ("requestURI", requestURIForQuery);
			final String resultQuery2 = this.kba.executeQuery (ontologyIdentifier, query2);
			final String scndpart2 = resultQuery2.substring (resultQuery2.indexOf ("value"));

			// return true if request type is check air quality limits or detailed air
			// quality report
			if (scndpart2.contains ("DetailedAirQualityReport"))
				return true;
			else if (scndpart2.contains ("CheckAirQualityLimits"))
				return true;
			else
				return false;
		}
		catch (final IOException e)
		{
			e.printStackTrace ();

		}
		return false;
	}


	/**
	 * Get the name of a feature.
	 * 
	 * @param foi The ID of the feature from which to get the name
	 * @param fois A map with features, the key is the ID of the feature
	 * @return The name or null if not found
	 */
	private static String getFeatureName (final String foi, final Map<String, AbstractFeatureType> fois)
	{
		final AbstractFeatureType featureType = fois.get (foi);
		if (featureType == null)
			return null;

		final List<JAXBElement<CodeType>> nameElement = featureType.getNameElement ();
		if (nameElement == null || nameElement.isEmpty ())
		{
			LOGGER.severe ("        No name for feature '" + foi + "' found.");
			return null;
		}
		return nameElement.get (0).getValue ().getValue ();
	}


	/**
	 * Increase the given date/time in ISO format by 1 hour.
	 * 
	 * @param startTime The start time
	 * @return The increased time
	 */
	public static String incTime (final String startTime)
	{
		// "2011-04-04T00:00:00+02:00"
		final Matcher matcher = DATETIMEPATTERN.matcher (startTime);
		matcher.matches ();

		final int year = Integer.parseInt (matcher.group (1));
		final int month = Integer.parseInt (matcher.group (2));
		final int date = Integer.parseInt (matcher.group (3));
		final int hourOfDay = Integer.parseInt (matcher.group (4));
		final int minute = Integer.parseInt (matcher.group (5));
		final int second = Integer.parseInt (matcher.group (6));
		final int millisecond = Integer.parseInt (matcher.group (8));
		final String sign = matcher.group (12);
		final int timeZoneHours = Integer.parseInt (matcher.group (13));
		final int timeZoneMinutes = Integer.parseInt (matcher.group (14));

		final String timeZone = matcher.group (11);

		final StringBuilder timeZoneID = new StringBuilder ("GMT").append (sign);
		final String h = Integer.toString (timeZoneHours);
		if (h.length () == 1)
			timeZoneID.append ('0').append (h);
		final String m = Integer.toString (timeZoneMinutes);
		if (m.length () == 1)
			timeZoneID.append ('0').append (m);
		final Calendar c = Calendar.getInstance (TimeZone.getTimeZone (timeZoneID.toString ()));
		c.set (year, month - 1, date, hourOfDay, minute, second);
		c.set (Calendar.MILLISECOND, millisecond);
		c.add (Calendar.HOUR_OF_DAY, 1);
		final Date time = c.getTime ();
		final String datetime = DateAccessor.format (time, DateAccessor.ISO);
		return datetime + timeZone;
	}


	/**
	 * Increase the given date/time in ISO format by 1 hour.
	 * 
	 * @param startTime The start time
	 * @param duration The duration to add
	 * @return The increased time
	 * @throws PESCaDOException if duration format is unknown
	 */
	public static String incTimeWithDuration (final String startTime, final String duration)
	{

		// "2011-04-04T00:00:00+02:00"
		final Matcher matcher = DATETIMEPATTERN.matcher (startTime);
		matcher.matches ();

		// final int year = Integer.parseInt(matcher.group(1));
		// final int month = Integer.parseInt(matcher.group(2));
		// final int date = Integer.parseInt(matcher.group(3));
		// final int hourOfDay = Integer.parseInt(matcher.group(4));
		// final int minute = Integer.parseInt(matcher.group(5));
		// final int second = Integer.parseInt(matcher.group(6));
		// final int millisecond = Integer.parseInt(matcher.group(8));
		final String sign = matcher.group (12);
		final int timeZoneHours = Integer.parseInt (matcher.group (13));
		final int timeZoneMinutes = Integer.parseInt (matcher.group (14));

		final String timeZone = matcher.group (11);

		final StringBuilder timeZoneID = new StringBuilder ("GMT").append (sign);
		final String h = Integer.toString (timeZoneHours);
		if (h.length () == 1)
			timeZoneID.append ('0').append (h);
		final String m = Integer.toString (timeZoneMinutes);
		if (m.length () == 1)
			timeZoneID.append ('0').append (m);
		final Calendar c = Calendar.getInstance (TimeZone.getTimeZone (timeZoneID.toString ()));

		try
		{
			Timestamp startTimeIso = DateAccessor.parseISO (startTime);
			long millisStart = startTimeIso.getTime ();
			c.setTimeInMillis (millisStart);
		}
		catch (ParseException e)
		{
			LOGGER.severe ("Error parsing time");
			e.printStackTrace ();
		}

		// old version
		// c.set(year, month - 1, date, hourOfDay, minute, second);
		// c.set(Calendar.MILLISECOND, millisecond);

		Date startDings = c.getTime ();

		final String [] durationArray = duration.split (":");
		if (!durationArray[3].equals ("00"))
		{
			final int dur4 = new Integer (durationArray[3]).intValue ();
			c.add (Calendar.SECOND, dur4);
		}
		if (!durationArray[2].equals ("00"))
		{
			final int dur3 = new Integer (durationArray[2]).intValue ();
			c.add (Calendar.MINUTE, dur3);
		}
		if (!durationArray[1].equals ("00"))
		{
			// LOGGER.info ("HOUR!");
			final int dur2 = new Integer (durationArray[1]).intValue ();
			c.add (Calendar.HOUR_OF_DAY, dur2);
			// c.add(Calendar.HOUR, dur2);
		}
		if (!durationArray[0].equals ("00"))
		{
			final int dur1 = new Integer (durationArray[0]).intValue ();
			c.add (Calendar.DAY_OF_MONTH, dur1);
		}
		/*
		 * if (duration.equals ("00:00:59:59")) { c.add (Calendar.HOUR_OF_DAY, 1); }
		 * else if (duration.equals ("00:07:59:59")) { c.add (Calendar.HOUR_OF_DAY,
		 * 8); } else if (duration.equals ("00:23:59:59")) { c.add
		 * (Calendar.DAY_OF_WEEK, 1); }
		 */
		final Date time = c.getTime ();
		final String datetime = DateAccessor.format (time, DateAccessor.ISO);
		// LOGGER.info ("startTime " + startTime);
		// LOGGER.info ("duration  " + duration);
		// LOGGER.info ("endTime   " + datetime + timeZone);
		return datetime + timeZone;
	}


	/**
	 * Reads capabilities of DRS (SOS).
	 */
	private void retrieveCapabilities ()
	{
		new Thread (new Runnable ()
		{
			@Override
			public void run ()
			{
				FusionServiceImpl.this.readCapabilities ();
			}
		}).start ();
	}


	/**
	 * Read the capabilities document from the SOS server and store the available
	 * data types.
	 */
	void readCapabilities ()
	{
		LOGGER.fine ("in readCapabilities");

		try
		{
			final Capabilities capabilities = this.sosClient.getCapabilities ();

			LOGGER.info ("Capabilities are received");
			final Contents contents = capabilities.getContents ();
			LOGGER.info ("Analyze offerings:");
			synchronized (FusionServiceImpl.offerings)
			{
				FusionServiceImpl.offerings.clear ();
				for (final ObservationOfferingType offering: contents.getObservationOfferingList ().getObservationOffering ())
				{
					final String dataType = offering.getId ();
					LOGGER.info ("dataType offeringid: " + dataType);
					final Collection<String> properties = new ArrayList<String> ();
					FusionServiceImpl.offerings.put (dataType, properties);
					LOGGER.info ("number of observedProp " + offering.getObservedProperty ().size ());

					for (final net.opengis.swe._1_0.PhenomenonPropertyType op: offering.getObservedProperty ())
					{
						// LOGGER.info ("href " + op.getHref ());
						properties.add (op.getHref ());
					}
				}
			}
			FusionServiceImpl.serviceIdentificationType = capabilities.getServiceIdentification ().getServiceTypeVersion ();
			LOGGER.info ("SOS Type: " + FusionServiceImpl.serviceIdentificationType.toString ());
		}
		catch (final Throwable t)
		{
			t.printStackTrace ();
			LOGGER.log (Level.WARNING, "Could not retrieve offerings from DRS. Retrying ...", t);

			try
			{
				Thread.sleep (10000);
			}
			catch (InterruptedException e)
			{
				LOGGER.log (Level.SEVERE, "Sleep for retry interrupted.", e);
			}
			this.readCapabilities ();
			return;
		}

		try
		{
			// query information about all nodes. If this is successful then node
			// information will be refreshed and stored.
			LOGGER.info ("DescribeSensor -all...");
			ArrayList<String> response = this.callDescribeSensorAllNew ();
			LOGGER.info ("Storing retrieved information...");
			this.DB.replaceAndStore (response);
		}
		catch (Exception e)
		{
			LOGGER.info ("DESCRIBE SENSOR-ALL FAILED! Previously stored database is used instead.");
			e.printStackTrace ();
		}

		LOGGER.info ("Fusion service is ready.");
	}


	/**
	 * Retrieve the data from the SOS server.
	 * 
	 * @param params The parameters
	 * @param racsParameters The relevant aspects to query for
	 * @param environmentalDataNature The environmental data nature
	 * @return The retrieved data indexed by the offering
	 * @throws IOException Crash during query
	 * @throws OWSException The SOS server returned an error
	 * @throws PESCaDOException
	 */
	protected Map<String, ObservationCollectionType> retrieveData (Parameters params, final List<ParameterType> racsParameters, final String environmentalDataNature) throws IOException, OWSException, PESCaDOException
	{
		// TODO Where do we get the SRS name from?
		// final String srsName = "urn:ogc:def:crs:EPSG:4326";
		// test for SOS2
		String sosType = "";
		if (FusionServiceImpl.serviceIdentificationType != null)
			sosType = FusionServiceImpl.serviceIdentificationType.toString ();
		else
			sosType = "1.0.0";
		String srsName = null;
		if (sosType.contains ("2.0.0"))
		{
			srsName = "urn:ogc:def:crs:EPSG::4326";
		}
		else
		{
			srsName = "urn:ogc:def:crs:EPSG:4326";
		}
		params = this.getModifiedParams (params);
		final BinarySpatialOpType cp = GMLHelper.containsPolygon (srsName, params.getPolygon ());
		String start = params.getStartDateTime (environmentalDataNature);
		String end = params.getEndDateTime (environmentalDataNature);

		LOGGER.info ("RetrieveData :::::::::::::::::::::::::: startTime " + start + " endTime " + end);
		// final EventTime eventTime = SOSClient.createDuringTimeRestriction
		// (params.getStartDateTime (), params.getEndDateTime (), true);
		final EventTime eventTime = SOSClient.createDuringTimeRestriction (start, end, true);

		final Map<String, ObservationCollectionType> result = new HashMap<String, ObservationCollectionType> ();
		final Iterator<ParameterType> racsParamIterator = racsParameters.iterator ();
		final Map<String, Collection<String>> offerings = this.getOfferings ();

		while (racsParamIterator.hasNext ())
		{
			final ParameterType paramType = racsParamIterator.next ();
			final String offering = paramType.getProperty ();
			LOGGER.info ("offering: " + offering);
			// Necessary for new version of drs, offerings don't have name
			// spaces
			// anymore
			String offeringWNS = "";
			final String [] oA = offering.split ("#");
			if (oA.length > 1)
				offeringWNS = oA[1];
			else
				offeringWNS = oA[0];
			final Collection<String> availableProperties = offerings.get (offeringWNS);
			LOGGER.info ("avProp: " + availableProperties);

			final List<Object> obspropv = paramType.getContent ();
			final Iterator<Object> obsiterator = obspropv.iterator ();
			while (obsiterator.hasNext ())
			{
				final String obsproperties = (String) obsiterator.next ();
				LOGGER.info ("obsprop: " + obsproperties);
				final String [] obspropArray = obsproperties.split (" ");
				final Collection<String> properties = new ArrayList<String> ();
				for (final String element: obspropArray)
				{
					if (availableProperties != null)
					{
						// Necessary for new version of drs, properties don't
						// have name
						// spaces anymore

						String propertiesWNS = "";
						// LOGGER.info ("element: " + element);
						final String [] pA = element.split ("#");
						if (pA.length > 1)
						{
							propertiesWNS = pA[1];
						}
						else
						{
							propertiesWNS = pA[0];
						}
						// LOGGER.info ("availableProperties: " +
						// availableProperties);
						// LOGGER.info ("propertiesWNS: " + propertiesWNS);

						// Differs between different prototype versions
						final String propertiesDot = propertiesWNS + OBSERVED_PROPERTY_EXACT;
						String propertiesFence = propertiesWNS + OBSERVED_EXACT;
						if (availableProperties.contains (propertiesDot))
						{
							properties.add (propertiesDot);
							// LOGGER.info ("ADD " + propertiesDot);
						}
						else if (availableProperties.contains (propertiesFence))
						{
							// Differs between different prototype versions: use
							// when DRS
							// contains forecasted data
							propertiesFence = propertiesWNS + "#" + environmentalDataNature.split ("#")[1] + "#" + EXACT;

							// Outcomment for tests with observed data
							// propertiesFence = propertiesWNS + OBSERVED_EXACT;

							properties.add (propertiesFence);
							// LOGGER.info ("ADD " + propertiesFence);
						}
						else if (availableProperties.contains (propertiesWNS))
						{
							properties.add (propertiesDot);
							// LOGGER.info ("ADD " + propertiesDot);
						}
						else
						{
							LOGGER.info (" G   No data for related aspect '" + offering + "' property '" + propertiesWNS + "'available.");
						}

					}
					else
						LOGGER.info ("    No data for related aspect '" + offering + "'available.");
				}

				if (properties.isEmpty ())
				{
					LOGGER.info ("   properties EMPTY!!!!!!!!!! ");
					continue;
				}

				LOGGER.info ("    Retrieving data for related aspect: " + offering);

				result.put (offeringWNS, this.sosClient.getObservation (srsName, offeringWNS, properties, null, SpatialOperatorNameType.INTERSECTS, cp, Collections.singleton (eventTime), null));

			}
		}
		return result;
	}


	/**
	 * Create array with requested observed properties.
	 * 
	 * @param racsParameters The relevant aspects to query for
	 * @return The observed properties
	 */
	private static List<String> createObservedPropertyArray (final List<ParameterType> racsParameters)
	{
		final List<String> properties = new ArrayList<String> ();
		for (final ParameterType paramType: racsParameters)
		{
			final String offering = paramType.getProperty ();
			LOGGER.info ("offering: " + offering);

			// Necessary for new version of drs, offerings don't have namespaces
			// anymore

			final String [] oA = offering.split ("#");
			final String offeringWNS = oA.length > 1 ? oA[1] : oA[0];
			final Collection<String> availableProperties = offerings.get (offeringWNS);
			if (availableProperties == null)
				continue;

			final List<Object> obspropv = paramType.getContent ();
			for (int j = 0; j < obspropv.size (); j++)
			{
				final String obsproperties = (String) obspropv.get (j);
				for (final String element: obsproperties.split (" "))
				{
					final String [] pA = element.split ("#");
					final String propertiesWNS = pA.length > 1 ? pA[1] : pA[0];

					// Differs between different prototype versions
					final String propertiesDot = propertiesWNS + OBSERVED_PROPERTY_EXACT;
					final String propertiesFence = propertiesWNS + OBSERVED_EXACT;
					if (availableProperties.contains (propertiesDot))
					{
						properties.add (propertiesWNS);
					}
					else if (availableProperties.contains (propertiesFence))
					{
						// Outcomment for tests with observed data
						// propertiesFence = propertiesWNS + OBSERVED_EXACT;
						properties.add (propertiesWNS);
					}
					else if (availableProperties.contains (propertiesWNS))
					{
						properties.add (propertiesWNS);
					}
					else
					{
						LOGGER.info (" G   No data for related aspect '" + offering + "' property '" + propertiesWNS + "'available.");
					}
				}
			}
		}
		return properties;
	}


	/**
	 * For the Arnold Admin scenario the air qualitiy exceedances need to be
	 * stored into the ontology. This operation created the exceedance
	 * environmental node necessary for that, if exceedances are available.
	 * 
	 * @param ontologyIdentifier identifier for ontology
	 * @param kbas access to KBAS
	 * @param resultValues contains values to be checked if exceedances occure
	 * @param geoArea affected geoarea
	 * @return false, if no exceedances have been reached; the uri of the
	 *         exceedance node, if exceedances have been reached.
	 */
	private String createExceedanceEnvironmentalNode (final String ontologyIdentifier, final KnowledgeBaseAccessService kbas, final String [][] parse_Result, final String geoArea)
	{
		boolean checkForExceedanceValues = false;
		for (int j = 0; j < parse_Result.length; j++)
		{
			String [] pElement = parse_Result[j];
			if (!pElement[3].equals ("noData"))
			{
				if (pElement[4].contains ("AirPollutantValue"))
				{
					String duration = pElement[5];
					if ((duration.equals ("00:00:59:59") || duration.equals ("00:01:00:00") || duration.equals ("00:07:59:59") || duration.equals ("00:08:00:00") || duration.equals ("00:23:59:59") || duration.equals ("01:00:00:00")))
					{
						LOGGER.info ("HUHUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU " + duration);
						checkForExceedanceValues = true;
						break;
					}
				}
			}
		}
		String exceedanceNodeURI = "false";
		if (checkForExceedanceValues)
		{
			exceedanceNodeURI = this.kba.createIndividualURI (NS_PESCADO_INSTANCES, "stationNode");

			ArrayList<StatementType> statements = new ArrayList<StatementType> ();
			addStatementToList (statements, exceedanceNodeURI, RDF_TYPE, NS_PESCADO + "EnvironmentalNode");
			addStatementToList (statements, exceedanceNodeURI, NS_PESCADO + "hasEnvironmentalNodeConfidenceValue", "1.0^^xsd:double");
			addStatementToList (statements, exceedanceNodeURI, NS_PESCADO + "hasEnvironmentalNodeName", "Station Node^^xsd:string");
			addStatementToList (statements, exceedanceNodeURI, NS_PESCADO + "hasEnvironmentalNodeForm", NS_PESCADO + "stationEnvironmentalNode");
			// Add exceedance node to geoArea node
			addStatementToList (statements, geoArea, NS_PESCADO + "hasGeoAreaEnvironmentalNode", exceedanceNodeURI);
			kbas.addABoxStatements (ontologyIdentifier, statements);
		}
		return exceedanceNodeURI;
	}


	/**
	 * Look for offerings offered by DRS (SOS).
	 * 
	 * @return Map with offerings
	 */
	private static Map<String, Collection<String>> getOfferings ()
	{
		synchronized (FusionServiceImpl.offerings)
		{
			return new HashMap<String, Collection<String>> (FusionServiceImpl.offerings);
		}
	}


	private static void addStatementToList (ArrayList<StatementType> statements, final String subject, final String predicate, final String object)
	{
		StatementType statement = new StatementType ();
		statement.setSubject (subject);
		statement.setPredicate (predicate);
		statement.setObject (object);
		statements.add (statement);
	}


	/*
	 * This method uses parsed data (from retrieveData) to perform fusion
	 * individually for each variable type.For each data provider (e.g. station-1
	 * or site-xx), describeSensor operation is used to retrieve coordinates and
	 * imprecision metrics.Also, the geographical area for data retrieval is
	 * modified (made larger) to ensure enough information for fusion.A profile
	 * describing the polygon centerpoint is evaluated and utilized in the fusion
	 * process.
	 * 
	 * @returns an ArrayList of strings which include fused value, imprecision
	 * (standard deviation) and relative weightings of used data sources.
	 */
	private ArrayList<FusionResult> calculateFusion (String [][] parsed_data, Parameters params, String ontologyIdentifier, String problemURI) throws ParseException
	{

		final Benchmark benchmarkFusionAlg = new Benchmark ();

		LOGGER.info ("Fusion  phase 1: Boundary coordinates modification and mask loading...");
		// ///////////////////////////////////
		String envOfInt = "";
		try
		{
			envOfInt = this.getEnvironmentOfInterest (ontologyIdentifier, problemURI);
		}
		catch (IOException e2)
		{
			e2.printStackTrace ();
		}

		// use the center of queried area
		double maxlat = -90.0;
		double minlat = 90.0;
		double maxlon = -90.0;
		double minlon = 90.0;

		Datetime startDate = null;
		Datetime endDate = null;

		double centlat = 0;
		double centlon = 0;

		String dateStart = "";
		String dateEnd = "";
		int timeZone = 0;

		try
		{
			dateStart = params.getStartDateTime ();
			dateEnd = params.getEndDateTime ();
			startDate = new Datetime (dateStart);
			endDate = new Datetime (dateEnd);
			timeZone = startDate.timeZone;
			String temp = "Date test: " + startDate.getStringDate () + "\t" + endDate.getStringDate () + timeZone;
			fsDump.add (temp);
			System.out.println (temp);

		}
		catch (Exception e1)
		{
			System.err.println ("ERROR: Parameter dates could not be parsed! => UTC-0 is the common time zone");
		}

		List<Double> polyCoords = params.getPolygon ();
		double index2 = 0;
		for (int i = 0; i < polyCoords.size (); i++)
		{
			double coordinate = polyCoords.get (i);
			// System.out.println ("polyCoordinate " + i + ": " + coordinate);
			if (i % 2 == 0)
			{// even = latitude
				centlat = centlat + coordinate;
				if (coordinate > maxlat)
					maxlat = coordinate;
				if (coordinate < minlat)
					minlat = coordinate;
				index2 = index2 + 1;
			}
			else
			{// odd =longitude
				centlon = centlon + coordinate;
				if (coordinate > maxlon)
					maxlon = coordinate;
				if (coordinate < minlon)
					minlon = coordinate;
			}
		}
		centlat = centlat / index2;
		centlon = centlon / index2;
		System.out.println ("Centerpoint " + centlat + ", " + centlon);
		System.out.println ("max/min boundaries :" + minlat + ", " + minlon + " - " + maxlat + ", " + maxlon);

		double [] boundaries = new double [4];
		boundaries[0] = minlon;
		boundaries[1] = maxlon;
		boundaries[2] = minlat;
		boundaries[3] = maxlat;

		// ////////////////////////////////////

		LOGGER.info ("Fusion phase 2 - structured dataPack instance creation...'");
		// CREATE a observation list using imprecisionDB and describe sensor
		// DB...
		ArrayList<Observation> obs = new ArrayList<Observation> ();

		for (int i = 0; i < parsed_data.length; i++)
		{

			String type = parsed_data[i][0];
			String ID = parsed_data[i][1];
			Datetime dt = new Datetime (parsed_data[i][2]);

			// set the time zone
			String originalDate = dt.getStringDate ();
			long origH = dt.hourCounter;
			dt.convertToGMT (timeZone); // timeZone is 0 by default if parameter
																	// parsing fails

			try
			{
				// if this datetime does not fit the selected time span +- 24h, the
				// observation
				// will be DISCCARDED
				if (dt.hourCounter > endDate.hourCounter + 24 || dt.hourCounter < startDate.hourCounter - 24)
				{
					System.out.println ("Rejected observation due to date: " + dt.getStringDate () + ", h_orig=" + origH + ", h=" + dt.hourCounter);
					System.out.println ("End: " + endDate.getStringDate () + ", h=" + endDate.hourCounter);

					continue;
				}
			}
			catch (NullPointerException e)
			{
			}
			;

			// now we read the actual value of observation
			// In some cases the value is a String rating and it can be converted to a
			// value with KB. IF this fails then the loop over Observations continues.
			double value = 0.0;
			String opDataType = parsed_data[i][4];
			try
			{
				value = Double.valueOf (parsed_data[i][3]);
			}
			catch (NumberFormatException e)
			{
				try
				{
					System.out.println ("Trying rating conversion..." + parsed_data[i][3]);
					Double converted = this.convertRatingToValue (ontologyIdentifier, parsed_data[i][3]);

					if (converted == null)
					{
						System.out.println ("Conversion returned null... skipped.");
						continue;
					}
					else
					// successful conversion
					{
						value = converted;
						opDataType = type + "Value";
					}

				}
				catch (Exception ex)
				{
					System.out.println ("FS parsing: NotDouble - " + parsed_data[i][3] + "... skipped.");
					continue;
				}
			}

			String duration = parsed_data[i][5];
			double durationDouble = 1.0; // 1h default
			String [] durTokens = duration.split (":");

			try
			{
				int durDays = Integer.valueOf (durTokens[0]);
				int durHours = Integer.valueOf (durTokens[1]);
				int durMins = Integer.valueOf (durTokens[2]);
				durationDouble = durDays * 24.0 + durHours + durMins / 60.0;

			}
			catch (ArrayIndexOutOfBoundsException e)
			{
				System.out.println ("Bad duration? " + (duration) + "  assumed to be hours.");
				int durHours = Integer.valueOf (durTokens[0]);
				durationDouble = durHours;
			}
			catch (NumberFormatException ex)
			{
				System.out.println ("Duration parsing: was not a number.");
			}

			if (durationDouble > 24)
				durationDouble = 24.0;
			if (durationDouble < 1)
				durationDouble = 1.0;

			double lat;
			double lon;
			double baseVar;
			double d_offset = 0;
			int t_offset = 0;
			boolean isForecasted = false;

			boolean hasExactCoords = this.DB.hasExactCoords (ID); // returns false if
																														// ID is not
																														// recognized
			if (!hasExactCoords)
			{
				isForecasted = true;
				d_offset = 3.0; // km
				t_offset = 2;
			}

			baseVar = DB.getBaseVar (ID, type, t_offset, durationDouble, isForecasted);

			// coordinates
			try
			{
				lat = Double.valueOf (parsed_data[i][6]);
				lon = Double.valueOf (parsed_data[i][7]);
			}
			catch (ArrayIndexOutOfBoundsException e)
			{
				try
				{
					lat = this.DB.getLatitude (ID);
					lon = this.DB.getLongitude (ID);

				}
				catch (Exception ex) // nullPointer or NumberFormat
				{// polygon centerpoint needs to be used
					// with a large penalty
					lat = centlat;
					lon = centlon;
					d_offset = 5.0;
					isForecasted = true;
					System.out.println ("Coordinates evaluation from parsed data AND describe sensor both FAILED => polygon center + penalty used instead.");
				}
			}

			// environment from mask
			Env forcedEnv = new Env ("vb"); // background assumed
			try
			{
				if (!isForecasted)
					forcedEnv = maps.getEnvironment (lat, lon, Options.profRadius);
			}
			catch (Exception exp)
			{
				System.err.println ("Could not retrieve Environment from mapPack.");
				forcedEnv = new Env ("vb");
			}

			// Now we have all the data we need to create an Observation
			// String type, String ID, double latitude, double longitude,
			// Datetime t_taken, double value, double baseVar, boolean model,
			// String relevance, double d_offset, int t_offset)
			Observation ob = new Observation (type, ID, lat, lon, dt, value, baseVar, isForecasted, forcedEnv, d_offset, t_offset, opDataType, durationDouble);
			obs.add (ob);
			String temp = "OBSERVATION ADDED :" + type + " " + ID + " " + lat + " " + lon + " " + dt.getStringDate () + " (" + originalDate + "),val " + value + ",bV " + baseVar + ",isM " + isForecasted + ",d_off " + d_offset + ",dur" + durationDouble + ",exProf " + forcedEnv.exProfile;
			// System.out.println (temp);
			fsDump.add (temp);
		}

		obs = this.addAirPollutantDataFromAirMerge(obs, startDate, centlat, centlon);
		
		System.out.println ("dataPack initalization...");
		DataPack datapack = new DataPack (obs); // now a DataPack is formed =>
		// the input data is structured
		// for efficient fusion

		LOGGER.info ("Fusion phase 3 - Fusion iteration over types and timestamps...");
		// FS needs to be threadsafe, ens cannot be a class variable
		// this.ens = new Ensemble (datapack, DB, maps);
		Ensemble ens = new Ensemble (datapack, DB, maps);
		int timeIncrement = 1;// 1h

		// ENVIRONMENT OF INTEREST or the center profile...
		Env centerEnv;
		if (envOfInt != null)
		{
		fsDump.add("EOI selected: "+envOfInt);	
			if (envOfInt.contains ("traffic") || envOfInt.contains ("Traffic"))
			{// urban area with road traffic
				centerEnv = new Env ("Ht");
			}
			else if (envOfInt.contains ("background")  || envOfInt.contains ("Background"))
			{// suburban background area
				centerEnv = new Env ("mb");
			}
			else if (envOfInt.contains ("park") || envOfInt.contains ("Park"))
			{// rural area or natural park
				centerEnv = new Env ("vb");
			}
			else
			{// not available => center coordinates represents environment
				fsDump.add("EOI not selected... evaluating... ");
				centerEnv = this.maps.getEnvironment (centlat, centlon, Options.profRadius);
				
			}

		}
		else
		{// not available => center coordinates represents environment
			fsDump.add("EOI not selected... evaluating... ");
			centerEnv = this.maps.getEnvironment (centlat, centlon, Options.profRadius);
					
		}
		
		String temp = centlat + ";" + centlon + ";" + centerEnv.getSummary ();
		fsDump.add (temp);	
		

		ArrayList<FusionResult> frList = ens.getCompleteFusionResult (centlat, centlon, timeIncrement, centerEnv,startDate, endDate);
		fsDump.add("===fused values for the selected environment======");
		for (int i=0;i<frList.size();i++) {
			FusionResult fr = frList.get(i);
			String temp3 = (fr.getType() + "\t" + fr.getValue() + "\t"
					+ fr.getDt().getStringDate() + "\t" + fr.getOpDataType());
			fsDump.add(temp3);
		}
		
		LOGGER.info ("Fusion phase 4 - Dump performance evaluation data to file...");
		ens.dumpObsToFile ();
		LOGGER.info ("Fusion phases - DONE!");

		LOGGER.info (benchmarkFusionAlg.printResults ());
		String temp2 = this.getFusionResultListForUI(frList);
		fsDump.add("=====retuned String ========");
		fsDump.add(temp2);
		ens.dumpALtoFile (fsDump);
		fsDump.clear ();
		return frList;

	}// end of doFusion


	String [][] parse_retrieveData (Map<String, ObservationCollectionType> data) throws ParseException
	{
		ArrayList<String []> al = new ArrayList<String []> ();
		String [][] parsed_data;

		for (final Entry<String, ObservationCollectionType> e: data.entrySet ())
		{
			final ObservationCollectionType oct = e.getValue ();

			final List<ObservationPropertyType> member = oct.getMember ();
			LOGGER.info ("    Reading data from " + member.size () + " observations.");

			for (final ObservationPropertyType type: member)
			{

				final ObservationType observation = type.getObservation ();
				LOGGER.info ("observation " + observation);
				// Adaption for 52north SOS 3.5
				if (observation == null)
				{
					continue;
				}

				// Store the feature of interests as stations
				final Map<String, AbstractFeatureType> fois = SOSClient.parseFeaturesOfInterest (observation);

				// Read all values of the observation
				// final Map<String, Map<String, List<Pair<String, Object>>>>

				final Vector<ValueBlockElement> resultValues = SOSClient.parseResultValues2ndPrototype (observation);

				// for (final Entry<String, Map<String, List<Pair<String,
				// Object>>>> rv:
				// resultValues.entrySet ())
				// {
				final Enumeration<ValueBlockElement> rvEnum = resultValues.elements ();

				while (rvEnum.hasMoreElements ())
				{

					// Start adding measurements by FS
					final ValueBlockElement ele = rvEnum.nextElement ();
					// LOGGER.info ("ValueBlockElement da!" + ele.getFeature () + " fois "
					// + fois.toString ());
					final String foiName = FusionServiceImpl.getFeatureName (ele.getFeature (), fois);
					if (foiName == null)
					{
						LOGGER.info ("FOI NULL !!!!!!!!!!!!!!!!!!!!!");
						continue;
					}
					// String feature = ele.getFeature ();
					final String time = ele.getTime ();

					final List<ValueAndDuration> valDurList = ele.getValDurList ();
					final Iterator<ValueAndDuration> valDurIterator = valDurList.iterator ();
					while (valDurIterator.hasNext ())
					{
						final ValueAndDuration valDur = valDurIterator.next ();
						String observedProperty = valDur.getFieldName ();
						final String duration = valDur.getDuration ();
						final Object value = valDur.getValue ();

						// LOGGER.info ("Original observed property " + observedProperty);
						final String [] oPArray = observedProperty.split ("_");
						observedProperty = NS_PESCADO + oPArray[0];
						final String opDataType = oPArray[1];
						// LOGGER.info ("opDataType " + opDataType);
						String href = type.getObservation ().getProcedure ().getHref ();
						String temp = "Reading data: " + foiName + " " + time + " " + href + " " + duration + " " + value.toString () + " " + opDataType;
						fsDump.add (temp);
						// ======data reorganizing - type ===================//
						String [] dataline = new String [6];

						StringTokenizer st = new StringTokenizer (observedProperty, "#");
						String [] tokenizedProperty = new String [2];
						int i = 0;
						while (st.hasMoreTokens ())
						{
							String token = st.nextToken ();
							// System.out.println(token);
							tokenizedProperty[i] = token;
							i++;
						}
						dataline[0] = tokenizedProperty[1]; // e.g. "PM2.5"
						// should be after #

						// ======data reorganizing - ID ===================//

						// String forecOrObs =
						// type.getObservation().getProcedure().
						dataline[1] = href;
						// ======data reorganizing - Timestamp
						// ===================//

						dataline[2] = time;

						// ======data reorganizing - value ===================//
						dataline[3] = value.toString ();
						dataline[4] = opDataType;
						dataline[5] = duration; // TODO COORDINATES IF
						// AVAILABLE!
						// System.out.println(dataline[0] + "\t" + dataline[1]
						// + "\t" + dataline[2] + "\t" + dataline[3]
						// + "\t" + dataline[4] + "\t" + dataline[5]);
						al.add (dataline);
					}
				}
			}
		}
		int len = al.size ();
		parsed_data = new String [len] [6];
		al.toArray (parsed_data);
		return parsed_data;
	}


	/**
	 * Call of describeSensor operation of DRS (SOS). This method parses the
	 * response xml given by DescribeSensor Returns a String[6] =
	 * {identifier1,identifier2,imprecision metrics, uncertainty, latitude,
	 * longitude}
	 * 
	 * @param id
	 * @return SensorML document
	 * @throws JAXBException
	 */
	private ArrayList<String> callDescribeSensorNew (String ID)
	{
		ArrayList<String> sensorResponse = new ArrayList<String> ();
		String describeSensorResponse = "";

		try
		{
			describeSensorResponse = this.sosClient.describeSensor (ID);

		}
		catch (final OWSException e)
		{
			e.printStackTrace ();
		}
		// DocumentBuilder needs the response to be given in InputSource, file
		// or InputStream. here we use InputSource
		InputSource s = new InputSource (new StringReader (describeSensorResponse));

		try
		{
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance ();
			DocumentBuilder db = dbf.newDocumentBuilder ();
			Document doc = db.parse (s);
			doc.getDocumentElement ().normalize ();
			// System.out.println("Root element " +
			// doc.getDocumentElement().getNodeName());

			// identifier - here we fetch identifier information
			NodeList nodeLst = doc.getElementsByTagName ("identifier");
			// System.out.println("Information of identifiers (list length = "+nodeLst.getLength());

			// iterate over items and search for "value"
			for (int i = 0; i < nodeLst.getLength (); i++)
			{
				Node fstNode = nodeLst.item (i);

				if (fstNode.getNodeType () == Node.ELEMENT_NODE)
				{
					try
					{
						Element fstElmnt = (Element) fstNode;
						NodeList fstNmElmntLst = fstElmnt.getElementsByTagName ("value");
						Element fstNmElmnt = (Element) fstNmElmntLst.item (0);
						NodeList fstNm = fstNmElmnt.getChildNodes ();

						// System.out.println("Identifier "+s+": " + ((Node)
						// fstNm.item(0)).getNodeValue());
						String item = ((Node) fstNm.item (0)).getNodeValue ();
						sensorResponse.add (item);
					}
					catch (NullPointerException e)
					{
						String item = "";
						sensorResponse.add (item);
					}
				}
			}

			// /////imprecision and coordinates==========================
			NodeList nodeLst_imprec = doc.getElementsByTagName ("swe:value");
			// System.out.println("Information about imprecision (list length = "
			// + nodeLst_imprec.getLength());

			for (int i = 0; i < nodeLst_imprec.getLength (); i++)
			{

				Node fstNode = nodeLst_imprec.item (i);
				String apu = fstNode.getTextContent ();
				// System.out.println(apu);
				sensorResponse.add (apu);
			}

		}
		catch (Exception e)
		{
			e.printStackTrace ();
		}

		return sensorResponse;
	}


	// This methods retrieves information from all listed sources and returns a
	// structured array of Strings.
	// These Strings are directly compatible with database.DataBase class
	public synchronized ArrayList<String> callDescribeSensorAllNew ()
	{
		ArrayList<String> results = new ArrayList<String> ();
		String describeSensorResponse = "";
		try
		{
			describeSensorResponse = this.sosClient.describeSensor ("all");
			// System.out.println(describeSensorResponse);

		}
		catch (final OWSException e)
		{
			e.printStackTrace ();
		}
		// DocumentBuilder needs the response to be given in InputSource, file
		// or InputStream. here we use InputSource
		InputSource s = new InputSource (new StringReader (describeSensorResponse));

		try
		{
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance ();
			DocumentBuilder db = dbf.newDocumentBuilder ();
			Document doc = db.parse (s);
			doc.getDocumentElement ().normalize ();
			System.out.println ("Root element " + doc.getDocumentElement ().getNodeName ());

			// identifier - here we fetch identifier information
			NodeList nodeLst = doc.getElementsByTagName ("System");
			System.out.println ("Information of identifiers (list length = " + nodeLst.getLength ());

			for (int i = 0; i < nodeLst.getLength (); i++)
			{

				// String temp = "";
				Node fstNode = nodeLst.item (i);

				NodeList systemChilds = fstNode.getChildNodes ();
				// 0 = identification 1=capabilities 2= position, these are all elements

				// ===========ID============================
				Node id = systemChilds.item (3);
				ArrayList<String> idArr = new ArrayList<String> ();
				Element idEle = (Element) id;
				NodeList ids = idEle.getElementsByTagName ("value");
				for (int z = 0; z < ids.getLength (); z++)
				{

					Node singleId = ids.item (z);
					String apu = singleId.getTextContent ();
					// temp = temp + apu + "; ";
					idArr.add (apu);

				}
				ArrayList<String> otherArr = new ArrayList<String> ();
				// Capabilities===========================
				Node cap = systemChilds.item (7);
				Element capEle = (Element) cap;
				NodeList caps = capEle.getElementsByTagName ("swe:value");
				for (int z = 0; z < caps.getLength (); z++)
				{

					Node singleCap = caps.item (z);
					String apu = singleCap.getTextContent ();
					// temp = temp + apu + "; ";
					otherArr.add (apu);

				}

				// Position===========================
				Node pos = systemChilds.item (11);
				Element posEle = (Element) pos;
				NodeList poss = posEle.getElementsByTagName ("swe:value");
				for (int z = 0; z < poss.getLength (); z++)
				{

					Node singlePos = poss.item (z);
					String apu = singlePos.getTextContent ();
					// temp = temp + apu + "; ";
					otherArr.add (apu);

				}

				String disc1 = "-";
				String disc2 = "-";
				if (idArr.size () >= 4)
				{
					disc1 = idArr.get (2);
					disc2 = idArr.get (3);
				}
				String line = idArr.get (0) + ";" + idArr.get (1) + ";" + disc1 + ";" + disc2 + ";" + otherArr.get (0) + ";" + otherArr.get (1) + ";" + otherArr.get (2) + ";" + otherArr.get (3) + ";" + otherArr.get (4) + ";";
				results.add (line);
				// System.out.println(line);

				/*
				 * boolean idIsElement=( id.getNodeType()==Node.ELEMENT_NODE); boolean
				 * capIsElement=( cap.getNodeType()==Node.ELEMENT_NODE); boolean
				 * posIsElement=( pos.getNodeType()==Node.ELEMENT_NODE);
				 * System.out.println("ID: "+id.getTextContent()+ " "+idIsElement);
				 * System.out.println("cap: "+cap.getTextContent()+ " "+capIsElement);
				 * System.out.println("pos: "+pos.getTextContent()+ " "+posIsElement);
				 * System.out.println();
				 */

			}// for System

		}
		catch (Exception e)
		{
			e.printStackTrace ();
		}

		return results;
	}


	/*
	 * This method modifies parameters so that polygon coordinates are much larger
	 * if needed.
	 */
	public Parameters getModifiedParams (Parameters params)
	{
		fsDump.add ("====Params manipulation=====");
		double maxlat = -90, maxlon = -90, minlat = 90, minlon = 90;

		List<Double> polyCoords = params.getPolygon ();
		double index2 = 0;
		for (int i = 0; i < polyCoords.size (); i++)
		{
			double coordinate = polyCoords.get (i);
			fsDump.add ("polyCoordinate " + i + ": " + coordinate);
			if (i % 2 == 0)
			{// even = latitude

				if (coordinate > maxlat)
					maxlat = coordinate;
				if (coordinate < minlat)
					minlat = coordinate;
				index2 = index2 + 1;
			}
			else
			{// odd =longitude

				if (coordinate > maxlon)
					maxlon = coordinate;
				if (coordinate < minlon)
					minlon = coordinate;
			}
		}

		// evaluate the need for area expansion
		double dlat = maxlat - minlat;
		double dlon = maxlon - minlon;
		final double degreeInKm = 2 * Math.PI * 6371 / 360;
		double lonScaler = Math.cos (2 * Math.PI * (0.5 * maxlat + 0.5 * minlat) / 360); // middle
		// lat
		// used

		double area = (dlat * degreeInKm) * (dlon * lonScaler * degreeInKm);
		double addFactor = 0.0;

		if (area < minAreaInKM2)
		{
			double newLatMax = maxlat;
			double newLatMin = minlat;
			double newLonMax = maxlon;
			double newLonMin = minlon;
			double newDlat;
			double newDlon;

			boolean enough = false;
			while (!enough)
			{// epand the area so that it covers at least the minimum area
				addFactor = addFactor + 0.05;
				// System.out.println("addFactor: "+addFactor);
				newLatMax = maxlat + addFactor * dlat;
				newLatMin = minlat - addFactor * dlat;
				newLonMax = maxlon + addFactor * dlon;
				newLonMin = minlon - addFactor * dlon;

				newDlat = newLatMax - newLatMin;
				newDlon = newLonMax - newLonMin;

				double newA = (newDlat * degreeInKm) * (newDlon * lonScaler * degreeInKm);
				if (newA >= minAreaInKM2)
				{
					enough = true;
					// System.out.println("Enough. newA = "+newA + " km2");
				}
			}

			fsDump.add ("Polygon area was roughly " + area + " km2 => increaseFactor " + addFactor);
			// modify params
			params.getPolygon ().clear ();

			// the following parametrization seems to work although why this exact
			// form is required is not known...
			params.getPolygon ().add (newLatMax);
			params.getPolygon ().add (newLonMax);
			params.getPolygon ().add (newLatMax);
			params.getPolygon ().add (newLonMin);

			params.getPolygon ().add (newLatMin);
			params.getPolygon ().add (newLonMin);
			params.getPolygon ().add (newLatMin);
			params.getPolygon ().add (newLonMax);
			params.getPolygon ().add (newLatMax);
			params.getPolygon ().add (newLonMax);

		}// if area < minArea
		for (int i = 0; i < params.getPolygon ().size (); i++)
		{
			fsDump.add ("Modified coordinates new [" + i + "] " + params.getPolygon ().get (i));
			System.out.println ("Modified coordinates new [" + i + "] " + params.getPolygon ().get (i));
		}

		return params;

	}


	// The user defined polygon is interpreted as a rectangular box and the
	// population count inside that box is calcuated and returned.
	public int getPolygonPopulation (Parameters params) throws NullPointerException
	{
		double maxlat = -90, maxlon = -90, minlat = 90, minlon = 90;

		List<Double> polyCoords = params.getPolygon ();
		double index2 = 0;
		for (int i = 0; i < polyCoords.size (); i++)
		{
			double coordinate = polyCoords.get (i);
			// System.out.println("polyCoordinate " + i + ": " + coordinate);
			if (i % 2 == 0)
			{// even = latitude

				if (coordinate > maxlat)
					maxlat = coordinate;
				if (coordinate < minlat)
					minlat = coordinate;
				index2 = index2 + 1;
			}
			else
			{// odd =longitude

				if (coordinate > maxlon)
					maxlon = coordinate;
				if (coordinate < minlon)
					minlon = coordinate;
			}
		}

		// evaluate the need for area expansion
		double dlat = maxlat - minlat;
		double midlat = (maxlat + minlat) / 2;
		double dlon = maxlon - minlon;
		double midlon = (maxlon + minlon) / 2;
		final double degreeInKm = 2 * Math.PI * 6371 / 360;
		double lonScaler = Math.cos (2 * Math.PI * (0.5 * maxlat + 0.5 * minlat) / 360); // middle
		// lat
		// used

		double area = (dlat * degreeInKm) * (dlon * lonScaler * degreeInKm);
		fsDump.add ("Population count sotring: area = " + area);
		int pop = this.maps.getLocalPopulation (midlat, midlon, area);

		return pop;

	}


	// This method stores a single population count information to KB
	public void storePopulationInfo (int polygonPop, String ontologyIdentifier, String problemURI) throws IOException
	{ // Marco:
		// At fusion time, you can then add the population
		// value with a single assertion (geoAreaXYZ hasGeoAreaPopulation
		// NNN^xsd:integer),
		// where geoAreaXYZ is the individual representing the geographic area
		// attached to the current request.
		final String result = kba.executeQuery (ontologyIdentifier, GEO_AREA_QUERY.replace ("%problemInstance%", problemURI));
		// LOGGER.info (" problemURI " + problemURI);
		// LOGGER.info (" result " + result);

		// Exception Test
		final JSONArray jsonArray = kba.getJSONBinding (result);
		if (jsonArray.isEmpty ())
		{
			final String msg = "Geo reference not found for problem " + problemURI;
			LOGGER.severe ("        " + msg);
			throw new RuntimeException (msg);
		}
		final String geoArea = jsonArray.getJSONObject (0).getJSONObject ("geoarea").getString ("value");

		LOGGER.info ("geoArea " + geoArea + " " + NS_PESCADO + "hasGeoAreaPopulation" + " " + polygonPop + XSD_INTEGER);

		final KnowledgeBaseAccessService kbas = kba.getService ();
		kbas.addABoxStatement (ontologyIdentifier, geoArea, NS_PESCADO + "hasGeoAreaPopulation", polygonPop + XSD_INTEGER);

	}


	/**
	 * Extracts string property from query result.
	 * 
	 * @param property to be extracted
	 * @param jsonArray query result
	 * @return String value
	 */
	private static String extractStringProperty (String property, final JSONArray jsonArray)
	{
		if (jsonArray.isEmpty ())
		{
			final String msg = "Property " + property + " not found!";
			LOGGER.severe ("        " + msg);
			return null;
		}
		final String stringProperty = jsonArray.getJSONObject (0).getJSONObject (property).getString ("value");
		return stringProperty;
	}


	/**
	 * Converts a String form rating to a Double value with the help of ontology
	 * 
	 * @param property
	 * @return Double value. A failed conversion returns null.
	 */
	public Double convertRatingToValue (String ontologyIdentifier, String rating)
	{

		// special cases, when the upper or lower limit just doesn't quite cut it
		if (rating.equals ("dryWeatherRainRating") )
			return 0.0;

		double lowValue = 0.0;
		double highValue = 0.0;
		boolean hasLower = false;
		boolean hasUpper = false;

		String property = "lowerBound";
		String propertyu = "upperBound";
		String ratingTypeForDRSResult = NS_PESCADO + rating;
		// Queries for Ontology:
		final String GET_LOWERBOUND_QUERY;

		{
			final StringBuilder q = new StringBuilder ();
			q.append ("PREFIX pescado: <http://www.pescado-project.eu/ontology/pescado.owl#>\n");
			q.append ("select ?lowerBound where ");
			q.append ("{ ");
			q.append ("<%ratingtype%> pescado:hasLowerBoundValue ?lowerBound . ");
			q.append ("}");
			GET_LOWERBOUND_QUERY = q.toString ();
		}

		final String GET_UPPERBOUND_QUERY;

		{
			final StringBuilder q2 = new StringBuilder ();
			q2.append ("PREFIX pescado: <http://www.pescado-project.eu/ontology/pescado.owl#>\n");
			q2.append ("select ?upperBound where ");
			q2.append ("{ ");
			q2.append ("<%ratingtype%> pescado:hasUpperBoundValue ?upperBound . ");
			q2.append ("}");
			GET_UPPERBOUND_QUERY = q2.toString ();
		}

		String lowerBoundsResult = null;
		try
		{
			String kbasquerylower = GET_LOWERBOUND_QUERY.replace ("%ratingtype%", ratingTypeForDRSResult);
			// LOGGER.info ("LOWER KBASQUERY " + kbasquerylower);
			lowerBoundsResult = this.kba.executeQuery (ontologyIdentifier, GET_LOWERBOUND_QUERY.replace ("%ratingtype%", ratingTypeForDRSResult));
			// LOGGER.info ("LOWER KBASQUERYResult " + lowerBoundsResult);

			final JSONArray jsonArray = this.kba.getJSONBinding (lowerBoundsResult);
			if (jsonArray.isEmpty ())
			{
				final String msg = "Property " + property + " not found!";
				LOGGER.severe ("        " + msg);
			}
			else
			{
				lowValue = jsonArray.getJSONObject (0).getJSONObject (property).getDouble ("value");
				hasLower = true;
			}

		}
		catch (IOException e)
		{
			e.printStackTrace ();
			LOGGER.info ("ERROR");
		}

		String upperBoundsResult = null;
		try
		{
			String kbasqueryupper = GET_UPPERBOUND_QUERY.replace ("%ratingtype%", ratingTypeForDRSResult);
			// LOGGER.info ("UPPER KBASQUERY " + kbasqueryupper);
			upperBoundsResult = this.kba.executeQuery (ontologyIdentifier, GET_UPPERBOUND_QUERY.replace ("%ratingtype%", ratingTypeForDRSResult));
			// LOGGER.info ("UPPER KBASQUERYResult " + upperBoundsResult);

			final JSONArray jsonArray2 = this.kba.getJSONBinding (upperBoundsResult);
			if (jsonArray2.isEmpty ())
			{
				final String msg = "Property " + propertyu + " not found!";
				LOGGER.severe ("        " + msg);
			}
			else
			{

				highValue = jsonArray2.getJSONObject (0).getJSONObject (propertyu).getDouble ("value");
				hasUpper = true;
			}
		}
		catch (IOException e)
		{
			e.printStackTrace ();
			LOGGER.info ("ERROR");

		}

		if (!hasLower && !hasUpper)
		{
			return null;
		}
		else if (hasUpper && hasLower)
		{
			Double avg = (lowValue + highValue) / 2.0;
			// return middle value of lower and upper bounds
			LOGGER.info ("Rating conversion: " + rating + " => " + avg);
			return avg;

		}
		else if (!hasUpper && hasLower)
		{
			LOGGER.info ("Rating conversion (one-sided): " + rating + " => " + lowValue);
			return (Double) lowValue;
		}
		else
		{
			LOGGER.info ("Rating conversion (one-sided): " + rating + " => " + highValue);
			return (Double) highValue;
		}

	}


	/**
	 * This methods retrieves the user selected Environment of Interest TODO
	 * 
	 * @throws IOException
	 */
	public String getEnvironmentOfInterest (String ontologyIdentifier, String problemURI) throws IOException
	{
		String eoi = null;

		final String result = this.kba.executeQuery (ontologyIdentifier, GEO_AREA_QUERY.replace ("%problemInstance%", problemURI));
		// LOGGER.info (" problemURI " + problemURI);
		// LOGGER.info (" result " + result);

		// Exception Test
		final JSONArray jsonArray = this.kba.getJSONBinding (result);
		if (jsonArray.isEmpty ())
		{
			final String msg = "Geo reference not found for problem " + problemURI;
			LOGGER.severe ("        " + msg);
			throw new RuntimeException (msg);
		}
		final String geoArea = jsonArray.getJSONObject (0).getJSONObject ("geoarea").getString ("value");

		final String result2 = this.kba.executeQuery (ontologyIdentifier, ENVIRONMENT_OF_INTEREST_QUERY.replace ("%geoarea%", geoArea));
		final JSONArray jsonArray2 = this.kba.getJSONBinding (result2);
		if (jsonArray2.isEmpty ())
		{
			final String msg = "Environment of interest not found for geo area " + geoArea;
			LOGGER.severe ("        " + msg);
			return null;
		}
		eoi = jsonArray2.getJSONObject (0).getJSONObject ("env").getString ("value");
		LOGGER.info ("Environment of interest: " + eoi);
		return eoi;
	}
	
	
	private ArrayList<Observation> addAirPollutantDataFromAirMerge(ArrayList<Observation> obs, Datetime dt,double lat, double lon) {
	boolean needsSO2 = true;
	boolean needsPM10 = true;
	boolean needsNO2 = true;
	boolean needsO3 = true;
	boolean needsCO = true;
	
	ArrayList<String> types = new ArrayList<String>();
	
	for (int i=0;i<obs.size();i++) {
		
		String type = obs.get(i).type;
		if (type.contains("SO2")) needsSO2 = false;
		if (type.contains("PM10")) needsPM10 = false; 
		if (type.contains("NO2")) needsNO2 = false; 
		if (type.contains("O3")) needsO3 = false; 
		if (type.contains("CO")) needsCO = false; 
	}
	
	if (needsSO2) types.add("SO2");
	if (needsPM10) types.add("PM10");
	if (needsNO2) types.add("NO2");
	if (needsO3) types.add("O3");
	if (needsCO) types.add("CO");
	
	try {
	//ArrayList<Observation> airMObs = AirMergeAccess.getObsAvg(lat, lon, types, dt, this.DB);
		ArrayList<Observation> airMObs = new ArrayList<Observation>(); // AirMerge DISABLED	
	fsDump.add("airMerge query with "+types.size() + " variables.");
	fsDump.add("Observations added: "+airMObs.size());
	
	for (int k=0;k<airMObs.size();k++) {
		Observation ob = airMObs.get(k);
		String temp = "OBSERVATION ADDED :" + ob.type + " " + ob.ID + " " + ob.latitude + " " + ob.longitude + " " + ob.datetime.getStringDate () +", val " + ob.value + ",bV " + ob.base_var + ",isM " +  ",d_off " + ob.d_offset + ",dur" + ob.duration + ",exProf " + ob.env.exProfile;
		// System.out.println (temp);
		fsDump.add (temp);
	}
	
	obs.addAll(airMObs);
	
	
	} catch (Exception e) {
		e.printStackTrace();
	}
		
		
    return obs;
	}
	
	/*This method return a String[] arrayList for the purposes of UI based on fused results
	 * 
	 */
	public String getFusionResultListForUI(ArrayList<FusionResult> list) {
		
		String uiString = "";
		final String s = "_";
		
		for (int i=0;i<list.size();i++) {
			FusionResult fr = list.get(i);
			String type = fr.getType();
			String date = fr.getDt().getStringDate();
			String value = fr.getValueString();
			String unit ="units";
			
			if (type.contains("SO2") || type.contains("NO2") || type.contains("O3") || type.contains("CO") || type.contains("PM10") || type.contains("PM2.5") || type.contains("ollen")) {
			unit = "g/m3";	
			} else if (type.contains("emperature")) {
				unit = "degrees,C";	
			} else if (type.contains("speed") || type.contains("Speed")) {
				unit = "m/s";
			} else if (type.contains("umidity")) {
				unit = "percent";
			} else if (type.contains("rain") || type.contains("Rain")) {
				unit = "mm/h";
			}
		
        uiString = uiString + type + s + date +s + value + s + unit +";";
		}
		
		fsDump.add("====returnedString=======");
		fsDump.add(uiString);
		
		return uiString;
		
	}
	
	

}// CLASS END
