/*
 * To change this template, choose Tools | Templates and open the template in
 * the editor.
 */
package services.pescado.fusionservice.impl.Masks;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.imageio.ImageIO;

import services.pescado.fusionservice.impl.datapack.Boundaries;


/**
 * 
 * @author johanssl
 */
public class MapPack implements Serializable
{
	private static final long				serialVersionUID	= 78391261;
	private ArrayList<LandMask>			landMasks;
	private ArrayList<DensityMask>	densityMasks;
	public boolean									landAvailable			= false;
	public boolean									densityAvailable	= false;
	public String										rootdir;
	public String										region;


	// Constructor 1: one map based on boudaries

	private static ArrayList<String []> getMapsList (String region, String rootdir, boolean LU)
	{
		ArrayList<String []> maps = new ArrayList<String []> ();
		String Z = System.getProperty ("file.separator");
		String name;
		if (LU)
		{
			name = "LUmaps_" + region + ".f";
		}
		else
		{
			name = "Dmaps_" + region + ".f";
		}
		String file = rootdir + "maps" + Z + name;
		try
		{

			BufferedReader br = new BufferedReader (new FileReader (file));
			String line;

			while ((line = br.readLine ()) != null)
			{

				String [] line_values = line.split (";", 10);
				line_values[0] = line_values[0] + ".png";
				maps.add (line_values);

			}
			br.close ();
		}

		catch (FileNotFoundException e)
		{

			e.printStackTrace ();
		}
		catch (IOException e)
		{

			e.printStackTrace ();
		}

		return maps;
	}


	// Constructor 2: map package based on defined region string
	public MapPack (String region, String rootDir)
	{
		String Z = System.getProperty ("file.separator");
		this.rootdir = rootDir;
		this.region = region;
		ArrayList<String []> maps = null;
		this.landMasks = new ArrayList<LandMask> ();

		maps = getMapsList (region, rootDir, true);

		// time to add all maps to arrayList
		for (int i = 0; i < maps.size (); i++)
		{
			String [] mapSpecs = maps.get (i);
			System.out.println ("Adding map " + mapSpecs[0]);

			try
			{
				this.landMasks.add (new LandMask (mapSpecs, rootDir));
			}
			catch (IOException e)
			{
				System.out.print ("LUmap load failed: " + mapSpecs[0] + "...missing?");
			}
		}

		if (this.landMasks.size () > 0)
			this.landAvailable = true;

		// DENSITY MASKS===========================

		this.densityMasks = new ArrayList<DensityMask> ();
		maps = getMapsList (region, rootDir, false);

		// time to add all maps to arrayList
		for (int i = 0; i < maps.size (); i++)
		{
			String [] mapSpecs = maps.get (i);
			System.out.println ("Adding density map " + mapSpecs[0]);

			try
			{
				this.densityMasks.add (new DensityMask (mapSpecs, rootDir));
			}
			catch (IOException e)
			{
				System.out.print ("Dmap load failed: " + mapSpecs[0] + "...missing?");
			}
		}

		if (this.densityMasks.size () > 0)
			this.densityAvailable = true;
	}


	// this method returns the smallest map which contains the two coordinates
	// if no suitable map is found then return null
	public LandMask getSuitableLandMap (double lat1, double lon1)
	{
		double currentSize = 1000000.0;
		int index = -1;
		for (int i = 0; i < this.landMasks.size (); i++)
		{
			double mapSize = this.landMasks.get (i).mapSize;

			boolean covers = this.landMasks.get (i).containsCoordinates (lat1, lon1);
			if (covers)
			{
				if (mapSize < currentSize)
				{
					index = i;
					currentSize = mapSize;
				}
			}

		}// for

		if (index == -1)
			return null;
		// System.err.println("returning the smallest landmask which containts "+lat1+", "+lon1
		// +" = "+this.landMasks.get(index).name);
		return this.landMasks.get (index);
	}


	// this method returns the smallest map which contains the two coordinates
	// if no suitable map is found then return null
	public DensityMask getSuitableDensityMap (double lat1, double lon1)
	{
		double currentSize = 1000000.0;
		int index = -1;
		for (int i = 0; i < this.densityMasks.size (); i++)
		{
			double mapSize = this.densityMasks.get (i).mapSize;

			boolean covers = this.densityMasks.get (i).containsCoordinates (lat1, lon1);
			if (covers)
			{
				if (mapSize < currentSize)
				{
					index = i;
					currentSize = mapSize;
				}
			}

		}// for

		if (index == -1)
			return null;

		return this.densityMasks.get (index);
	}


	public Env getEnvironment (double lat, double lon, int profRadius)
	{

		double pop = this.getLocalPopulation (lat, lon) / 1000;
		double [] suit = this.getSUITprofile (lat, lon, profRadius);
		suit[0] = pop;
		Env env = new Env (suit);
		return env;

	}


	public int getDensity (double lat, double lon)
	{

		double currentSize = 1000000.0;
		int index = -1;
		for (int i = 0; i < this.densityMasks.size (); i++)
		{
			double mapSize = this.densityMasks.get (i).mapSize;

			boolean covers = this.densityMasks.get (i).containsCoordinates (lat, lon);
			if (covers)
			{
				if (mapSize < currentSize)
				{
					index = i;
					currentSize = mapSize;
				}
			}

		}// for

		if (index == -1)
		{
			return 0;
		}
		else
		{
			return this.densityMasks.get (index).getDensity (lat, lon);
		}
	}


	// this method return a SUIT double[]
	public double [] getSUITprofile (double lat, double lon, int lres)
	{
		double [] temp = null;
		double [] suit =
		{
				0,
				0,
				0,
				0,
				0,
				0,
				0,
				0,
				0
		};
		try
		{
			temp = this.getSuitableLandMap (lat, lon).getSUITarray (lat, lon, lres);
		}
		catch (NullPointerException e)
		{

		}
		if (temp == null)
			temp = suit;
		return temp;
	}


	public int getLocalPopulation (double lat, double lon)
	{
		int res = 8;
		double currentSize = 1000000.0;
		int index = -1;
		for (int i = 0; i < this.densityMasks.size (); i++)
		{
			double mapSize = this.densityMasks.get (i).mapSize;

			boolean covers = this.densityMasks.get (i).containsCoordinates (lat, lon);
			if (covers)
			{
				if (mapSize < currentSize)
				{
					index = i;
					currentSize = mapSize;
				}
			}

		}// for

		if (index == -1)
		{
			return 0;
		}
		else
		{
			return this.densityMasks.get (index).getLocalPopulation (lat, lon, res);
		}
	}


	public char getZone (double lat, double lon)
	{

		double currentSize = 1000000.0;
		int index = -1;
		for (int i = 0; i < this.landMasks.size (); i++)
		{

			boolean covers = this.landMasks.get (i).containsCoordinates (lat, lon);
			if (covers)
			{

				if (this.landMasks.get (i).mapSize < currentSize)
				{
					index = i;
					currentSize = this.landMasks.get (i).mapSize;
				}
			}

		}// for

		if (index == -1)
		{
			return 'X';
		}
		else
		{
			return this.landMasks.get (index).get_maskChar (lat, lon);
		}

	}


	public Color getColor (char zonetype)
	{
		Color col;

		if (zonetype == 'U')
		{
			col = new Color (229, 0, 76);
		}
		else if (zonetype == 'S')
		{
			col = new Color (254, 0, 0);
		}
		else if (zonetype == 'W' || zonetype == 'w')
		{
			int rand = (int) (30 * Math.random ());
			col = new Color (180, 215 + rand, 255);
		}
		else if (zonetype == 'T')
		{
			col = new Color (203, 0, 0);

		}
		else if (zonetype == 'I')
		{
			col = new Color (10, 10, 255);

		}
		else if (zonetype == 'V')
		{
			int rand = (int) (30 * Math.random ());
			col = new Color (85, 157 + rand, 30);

		}
		else
		{
			col = Color.WHITE;
		}

		return col;
	}


	private String [] selectBestMap (double [] boundaries, String [][] maps)
	{
		double lonmin = boundaries[0];
		double lonmax = boundaries[1];
		double latmin = boundaries[2];
		double latmax = boundaries[3];
		int length = maps.length;
		double best = 1000;
		int index = 0;
		double test_val;
		for (int i = 0; i < length; i++)
		{ // a good candidate for map: its lonmin is lower than the boundary, its
			// lonmax is larger than the boundary etc

			double lon_min_diff = lonmin - Double.valueOf (maps[i][0]); // we want the
																																	// lonmin of
																																	// pdMap to be
																																	// SMALLER
			if (lon_min_diff < 0)
			{
				lon_min_diff = lon_min_diff * -100;
			}
			double lon_max_diff = Double.valueOf (maps[i][1]) - lonmax;
			if (lon_max_diff < 0)
			{
				lon_max_diff = lon_max_diff * -100;
			}
			double lat_min_diff = latmin - Double.valueOf (maps[i][2]); // we want the
																																	// latmin of
																																	// pdMap to be
																																	// SMALLER
			if (lat_min_diff < 0)
			{
				lat_min_diff = lat_min_diff * -100;
			}
			double lat_max_diff = Double.valueOf (maps[i][3]) - latmax;
			if (lat_max_diff < 0)
			{
				lat_max_diff = lat_max_diff * -100;
			}

			test_val = lon_min_diff + lon_max_diff + lat_min_diff + lat_max_diff;

			// Double.valueOf(maps[i][0])- lonmin + Double.valueOf(maps[i][2]) -
			// lonmax +
			// latmin - Double.valueOf(maps[i][2]) + latmax -
			// Double.valueOf(maps[i][3]);
			System.out.println ("test_val for map " + maps[i][4] + ": " + (int) test_val);

			if (test_val < best)
			{
				index = i;
				best = test_val;
			}
		}

		return maps[index];
	}


	public Color getDensityColor (int dens)
	{

		Color col;
		final int scaler = 16;
		int increment;
		if (dens == 2100 * scaler)
		{
			increment = 25;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 1625 * scaler)
		{
			increment = 24;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 1375 * scaler)
		{
			increment = 23;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 1125 * scaler)
		{
			increment = 22;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 875 * scaler)
		{
			increment = 21;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 625 * scaler)
		{
			increment = 20;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 450 * scaler)
		{
			increment = 19;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 350 * scaler)
		{
			increment = 18;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 250 * scaler)
		{
			increment = 17;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 175 * scaler)
		{
			increment = 16;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 125 * scaler)
		{
			increment = 15;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 95 * scaler)
		{
			increment = 14;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 85 * scaler)
		{
			increment = 13;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 75 * scaler)
		{
			increment = 12;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 65 * scaler)
		{
			increment = 11;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 55 * scaler)
		{
			increment = 9;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 45 * scaler)
		{
			increment = 7;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 35 * scaler)
		{
			increment = 5;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 25 * scaler)
		{
			increment = 4;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 15 * scaler)
		{
			increment = 3;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 8 * scaler)
		{
			increment = 2;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else if (dens == 3 * scaler)
		{
			increment = 1;
			col = new Color (255, 255 - increment * 10, 255 - increment * 10);
		}
		else
		{
			col = new Color (255, 255, 255);

		}

		return col;
	}


	public void saveMapPackToFile ()
	{
		String Z = System.getProperty ("file.separator");
		String fileName = this.rootdir + this.region + "_mapPack.dat";
		System.out.println ("Saving mapapck to: " + fileName);
		// Write to disk with FileOutputStream
		FileOutputStream f_out;
		try
		{
			f_out = new FileOutputStream (fileName);

			// Write object with ObjectOutputStream
			ObjectOutputStream obj_out;
			try
			{
				obj_out = new ObjectOutputStream (f_out);

				// Write object out to disk
				obj_out.writeObject (this);
				obj_out.close ();
				f_out.close ();
			}
			catch (IOException ex)
			{
				Logger.getLogger (MapPack.class.getName ()).log (Level.SEVERE, null, ex);
			}

		}
		catch (FileNotFoundException ex)
		{
			Logger.getLogger (MapPack.class.getName ()).log (Level.SEVERE, null, ex);
		}

	}


	public static MapPack load (String filename)
	{
		// Read from disk using FileInputStream
		FileInputStream f_in;
		try
		{
			f_in = new FileInputStream (filename);

			// Read object using ObjectInputStream
			ObjectInputStream obj_in;
			try
			{
				obj_in = new ObjectInputStream (f_in);
				try
				{
					// Read an object
					Object obj = obj_in.readObject ();
					if (obj instanceof MapPack)
					{
						System.out.println ("MapPack loaded from .dat-file successfully.");
						MapPack mp = (MapPack) obj;
						return mp;
					}
				}
				catch (ClassNotFoundException ex)
				{
					Logger.getLogger (MapPack.class.getName ()).log (Level.SEVERE, null, ex);
				}
			}
			catch (IOException ex)
			{
				Logger.getLogger (MapPack.class.getName ()).log (Level.SEVERE, null, ex);
			}

		}
		catch (FileNotFoundException ex)
		{
			Logger.getLogger (MapPack.class.getName ()).log (Level.SEVERE, null, ex);
		}

		return null;
	}


	public static void produceAndStoreMapsList (Boundaries bounds, int pixelWidth, int res, String dir, boolean LU)
	{
		ArrayList<String> arr = new ArrayList<String> ();
		String n = System.getProperty ("line.separator");
		final double degreeInMeters = 2 * Math.PI * 6371000 / 360;

		double dlat = bounds.latmax - bounds.latmin;
		double dlon = bounds.lonmax - bounds.lonmin;
		double midlat = (bounds.latmax + bounds.latmin) / 2;

		double scaler = Math.cos (midlat / 360 * 2 * Math.PI);

		double dlatM = dlat * degreeInMeters;
		double dlonM = dlon * degreeInMeters * scaler;

		int mapWidth = res * pixelWidth;

		int maxH = (int) (dlatM / mapWidth);
		int maxW = (int) (dlonM / mapWidth);
		System.out.println ("Map width: " + mapWidth + " m");
		System.out.println (dlat + "=> " + maxH);
		System.out.println (dlon + "=> " + maxW);
		System.out.println ("Total of" + maxW * maxH + " maps");

		double dlonMap = mapWidth / degreeInMeters / scaler;
		double dlatMap = mapWidth / degreeInMeters;

		for (int h = 0; h <= maxH; h++)
		{
			double latmax = editPrecision (bounds.latmax - h * dlatMap, 4);
			double latmin = editPrecision (bounds.latmax - (h + 1) * dlatMap, 4);
			for (int w = 0; w <= maxW; w++)
			{
				double lonmax = editPrecision (bounds.lonmin + (w + 1) * dlonMap, 4);
				double lonmin = editPrecision (bounds.lonmin + w * dlonMap, 4);

				String mapName = latmin + "_" + latmax + "_" + lonmin + "_" + lonmax + "_" + res + "_" + pixelWidth;
				System.out.println (mapName);
				String temp = mapName + ";" + lonmin + ";" + lonmax + ";" + latmin + ";" + latmax;
				// {"26.01","27.3","60.3","61.0","kh.png"},

				arr.add (temp);
				String map = "";
				if (LU)
				{
					map = "http://horttanainen.pc.fmi.fi/cgi-bin/mapserv.exe?map=/ms4w/apps/landuse/" + "map/lasse.map&SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&layers=CORINE2006landuse&SRS=EPSG:4326&" + "BBOX=" + lonmin + "," + latmin + "," + lonmax + "," + latmax + "&WIDTH=" + res + "&HEIGHT=" + res + "&format=image/png&styles=default";
				}
				else
				{// pop dens
					map = "http://horttanainen.pc.fmi.fi/cgi-bin/mapserv.exe?map=/ms4w/apps/landuse/" + "map/lasse.map&SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&layers=vaesto&SRS=EPSG:4326&" + "BBOX=" + lonmin + "," + latmin + "," + lonmax + "," + latmax + "&WIDTH=" + res + "&HEIGHT=" + res + "&format=image/png&styles=default";

				}

				System.out.println (map);
				URL imageURL = null;
				try
				{
					imageURL = new URL (map);
				}
				catch (MalformedURLException ex)
				{
					Logger.getLogger (MapPack.class.getName ()).log (Level.SEVERE, null, ex);
				}
				try
				{
					BufferedImage bufferedImage = ImageIO.read (imageURL);
					int htest = bufferedImage.getHeight ();
					int wtest = bufferedImage.getWidth ();
					System.out.println (htest + " x " + wtest);

					try
					{
						// Save as new values_img
						ImageIO.write (bufferedImage, "PNG", new File (dir + mapName + ".PNG"));
					}
					catch (IOException ex)
					{

					}

				}
				catch (IOException ex)
				{
					Logger.getLogger (MapPack.class.getName ()).log (Level.SEVERE, null, ex);
				}

				System.out.println ();
				System.out.println ();
			}
		}
		String listName = "luMaps.f";
		if (!LU)
			listName = "DMaps.f";
		printArrayString (dir, arr, listName, true);
	}


	// this tool method edits the number of decimals of d2 double
	// int precision is the number of decimals wanted
	public static double editPrecision (double value, int precision)
	{
		int prec = (int) Math.pow (10, precision);
		int temp = (int) (prec * value);
		double result = (double) temp / prec;
		return result;
	}


	// This method prints recieved messages hourly counters to message.csv file
	public static void printArrayString (String resultsDir, ArrayList<String> lines, String name, boolean append)
	{
		String n = System.getProperty ("line.separator");

		try
		{
			BufferedWriter out = new BufferedWriter (new FileWriter (resultsDir + name, append));
			for (int i = 0; i < lines.size (); i++)
			{
				String line = lines.get (i);
				out.write (line + n);
			}

			out.close ();
		}
		catch (IOException e)
		{
		}

	}


	public int getLocalPopulation (double midlat, double midlon, double area)
	{
		int pop = 0;

		int n = Math.max (0, (int) ((Math.sqrt (area / 0.25 / 0.25) / 2) - 0.5));
	//	double scaler = area / (2 * n * 0.25 + 1) * (2 * n * 0.25 + 1);
        
		DensityMask pdm = this.getSuitableDensityMap (midlat, midlon);
		if (pdm == null) return 0;
		
		pop = pdm.getLocalPopulation (midlat, midlon, n);
		pop = (int) (pop);

		return pop;

	}

}// CLASS END
