package services.pescado.fusionservice.impl;

import java.net.*;
import java.util.*;
import javax.xml.namespace.*;
import services.pescado.kbas.*;
import services.pescado.relatedaspectscomputationservice.*;


/**
 * Manages the services used by the Fusion Service. Note: This should be put
 * into a registry (if there would be one).
 * 
 * @author J&uuml;rgen Mo&szlig;graber
 */
public final class Services
{
	final static QName															RELATED_ASPECTS_COMPUTATION_SERVICE_NAME	= new QName ("http://pescado.services/RelatedAspectsComputationService/", "RelatedAspectsComputationService");
	final static QName															KNOWLEDGE_BASE_ACCESS_SERVICE_NAME				= new QName ("http://pescado.services/KnowledgeBaseAccessService/", "KnowledgeBaseAccessService");

	private static RelatedAspectsComputationService	racs;

	private static URL															racsWsdlLocation;

	private final static KBAccess										KBAS																			= new KBAccess ();


	/**
	 * Initialise the web service access.
	 * 
	 * @param properties Properties configuration with service URLs
	 * @throws MalformedURLException The WSDL location is not an URL
	 */
	public static void init (final Properties properties) throws MalformedURLException
	{
		KBAS.setWSDLLocation (properties.getProperty (KNOWLEDGE_BASE_ACCESS_SERVICE_NAME.getLocalPart ()));
		Services.racsWsdlLocation = new URL (properties.getProperty (RELATED_ASPECTS_COMPUTATION_SERVICE_NAME.getLocalPart ()));
	}


	/**
	 * Get the port for the knowledge base access service.
	 * 
	 * @return The service port
	 */
	public synchronized static KBAccess getKnowledgeBaseAccess ()
	{
		return KBAS;
	}


	/**
	 * Get the port for the related aspects computation service.
	 * 
	 * @return The service port
	 */
	public synchronized static RelatedAspectsComputationService getRelatedAspectsComputationWS ()
	{
		if (racs == null)
		{
			final RelatedAspectsComputationService_Service s = new RelatedAspectsComputationService_Service (Services.racsWsdlLocation, RELATED_ASPECTS_COMPUTATION_SERVICE_NAME);
			racs = s.getRelatedAspectsComputationServiceSOAP ();
		}
		return racs;
	}


	/**
	 * Helper class.
	 */
	private Services ()
	{
		// Empty by intention
	}
}
