package services.pescado.fusionservice.impl;

/**
 * Informations about the value types to use from the ontology.
 * 
 * @author J&uuml;rgen Mo&szlig;graber
 */
public enum ValueTypes
{
	/** HumidityValue. */
	HUMIDITYVALUE("HumidityValue", "http://www.pescado-project.eu/ontology/pescado.owl#humidity", "percent", "^^xsd:double"),

	/** PressureValue. */
	PRESSUREVALUE("PressureValue", "http://www.pescado-project.eu/ontology/pescado.owl#pressure", "ectoPascal", "^^xsd:double"),

	/** RainValue. */
	RAINVALUE("RainValue", "http://www.pescado-project.eu/ontology/pescado.owl#rain", "millimeterPerHour", "^^xsd:double"),

	/** RainValue. */
	SECONDPROTRAINVALUE("RainValue", "http://www.pescado-project.eu/ontology/pescado.owl#rain_RainValue", "millimeterPerHour", "^^xsd:double"),

	/** SkyConditionValue. */
	SKYCONDITIONVALUE("SkyConditionValue", "http://www.pescado-project.eu/ontology/pescado.owl#skyCondition", "eights", "^^xsd:integer"),

	/** SnowValue. */
	SNOWVALUE("SnowValue", "http://www.pescado-project.eu/ontology/pescado.owl#snow", "centimeterPerHour", "^^xsd:double"),

	/** TemperatureValue. */
	TEMPERATUREVALUE("TemperatureValue", "http://www.pescado-project.eu/ontology/pescado.owl#temperature", "degreeC", "^^xsd:double"),

	/** TemperatureValue. */
	SECONDPROTTEMPERATUREVALUE("TemperatureValue", "http://www.pescado-project.eu/ontology/pescado.owl#temperature_TemperatureValue", "degreeC", "^^xsd:double"),

	/** UVIndexValue. */
	UVINDEXVALUE("UVIndexValue", "http://www.pescado-project.eu/ontology/pescado.owl#UVIndex", "TODO", "^^xsd:integer"),

	/** WindSpeedValue. */
	WINDSPEEDVALUE("WindSpeedValue", "http://www.pescado-project.eu/ontology/pescado.owl#windSpeed", "meterPerSecond", "^^xsd:double"),

	/** WindSpeedValue. */
	SECONDPROTWINDSPEEDVALUE("WindSpeedValue", "http://www.pescado-project.eu/ontology/pescado.owl#windSpeed_WindSpeedValue", "meterPerSecond", "^^xsd:double"),

	/** WindDirectionValue. */
	WINDDIRECTIONVALUE("WindDirectionValue", "http://www.pescado-project.eu/ontology/pescado.owl#windDirection", "degree", "^^xsd:integer");

	private final String	className;
	private final String	uri;
	private final String	unit;
	private final String	xsdType;


	/**
	 * Constructor.
	 * 
	 * @param className The class name of the value
	 * @param uri The URI of the value
	 * @param unit The unit of the value
	 * @param xsdType The XSD type of the value
	 */
	private ValueTypes (final String className, final String uri, final String unit, final String xsdType)
	{
		this.className = className;
		this.uri = uri;
		this.unit = unit;
		this.xsdType = xsdType;
	}


	/**
	 * Get the class name of the value.
	 * 
	 * @return The class name
	 */
	public String getClassName ()
	{
		return this.className;
	}


	/**
	 * Get the URI of the value.
	 * 
	 * @return The uri
	 */
	public String getUri ()
	{
		return this.uri;
	}


	/**
	 * Get the unit of the value.
	 * 
	 * @return The unit
	 */
	public String getUnit ()
	{
		return this.unit;
	}


	/**
	 * Get the XSD type of the value.
	 * 
	 * @return The xsdType
	 */
	public String getXsdType ()
	{
		return this.xsdType;
	}


	/**
	 * Get the ValueType which belongs to the given URI.
	 * 
	 * @param uri The URI
	 * @return The ValueType or null if it does not exist
	 */
	public static ValueTypes getByURI (final String uri)
	{
		String testUri = "";
		final String [] nsarray = uri.split ("#");
		if (nsarray.length == 1)
		{
			testUri = "http://www.pescado-project.eu/ontology/pescado.owl#" + uri;
		}
		else
		{
			testUri = uri;
		}
		for (final ValueTypes vt: ValueTypes.values ())
			if (vt.getUri ().equals (testUri))
				return vt;
		return null;
	}
}
