package services.pescado.fusionservice.impl.database;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;

import services.pescado.fusionservice.impl.datapack.Observation;
import services.pescado.fusionservice.impl.fuser.Datetime;

/**
 * 
 * @author johanssl
 */
public class DataBase {
	ArrayList<DBline> db;
	ArrayList<String[]> locations;

	public EqProfiles eqProfiles;
	private Hashtable<String, double[]> defaultBaseImpr;

	public DataBase(String rootdir) {

		this.eqProfiles = new EqProfiles(rootdir);

		// default imprecision metrics Hash
		this.defaultBaseImpr = new Hashtable<String, double[]>();
		defaultBaseImpr.put("PM2.5", new double[] { 2.0, 8.0 });
		defaultBaseImpr.put("PM10", new double[] { 4.0, 16.0 });
		defaultBaseImpr.put("SO2", new double[] { 1.0, 4.0 });
		defaultBaseImpr.put("NO2", new double[] { 4.0, 16.0 });
		defaultBaseImpr.put("CO", new double[] { 10.0, 40.0 });
		defaultBaseImpr.put("O3", new double[] { 8.0, 32.0 });
		defaultBaseImpr.put("rain", new double[] { 0.5, 2 });
		defaultBaseImpr.put("windSpeed", new double[] { 0.5, 2.0 });
		defaultBaseImpr.put("temperature", new double[] { 0.5, 2.0 });
		defaultBaseImpr.put("windDirection", new double[] { 5, 20.0 });

		this.db = new ArrayList<DBline>();
		String Z = System.getProperty("file.separator");
		int lineNumber = 0;

		String file = rootdir + "data" + Z + "db.csv";
		System.out.println("DB reading: file is " + file);

		try {
			BufferedReader br = new BufferedReader(new FileReader(file));

			String line;
			while ((line = br.readLine()) != null) {

				lineNumber++;
				if (lineNumber == 1)
					continue;
				try {

					this.db.add(new DBline(line));
					// System.out.println("Adding ID: "+line_values[0]);
				} catch (ArrayIndexOutOfBoundsException e) {
					System.out
							.println("DB-file: Array index out of bounds while reading!");
				}

			}
			br.close();
		}

		catch (FileNotFoundException e) {
			System.out.println("DB-file: file NOT found!");

			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		this.locations = this.readLocations(rootdir);

	}

	public double getGenericVarEstimate(String type, boolean model) {
		int index = 0;
		if (model)
			index = 1;
		Double impr = null;
		try {
			impr = this.defaultBaseImpr.get(type)[index];
		} catch (NullPointerException e) {
			return 2.0 * (index + 1) * (index + 1);
		}
		if (impr == null)
			impr = 2.0 * (index + 1) * (index + 1); // 4 times for models,
													// generic unknown type
		return impr;
	}
	
	
	

	public ArrayList<String> getLocationList() {
		ArrayList<String> locList = new ArrayList<String>();
		for (int i = 0; i < this.locations.size(); i++) {

			String loc = this.locations.get(i)[0];
			locList.add(loc);

		}

		return locList;

	}

	public ArrayList<double[]> getCoordsFromLocations() {
		ArrayList<double[]> locs = new ArrayList<double[]>();

		double lat = 0;
		double lon = 0;
		for (int i = 0; i < this.locations.size(); i++) {

			lat = Double.valueOf(this.locations.get(i)[3]);
			lon = Double.valueOf(this.locations.get(i)[4]);

			double[] temp = { lat, lon };
			locs.add(temp);

		}

		return locs;

	}

	public double[] getCoordsFromLocation(String identifier) {
		boolean found = false;
		double lat = 0;
		double lon = 0;
		for (int i = 0; i < this.locations.size(); i++) {
			if (this.locations.get(i)[0].equals(identifier)) {
				found = true;
				lat = Double.valueOf(this.locations.get(i)[3]);
				lon = Double.valueOf(this.locations.get(i)[4]);
			}
		}

		if (found) {
			double[] temp = { lat, lon };
			return temp;
		} else {
			return null;
		}

	}

	private DBline getIndex(String ID) {

		for (int i = 0; i < this.db.size(); i++) {
			if (this.db.get(i).ID1.equals(ID) || this.db.get(i).ID2.equals(ID))
				return db.get(i);
		}
		return null;
	}

	public double getLatitude(String ID) {

			return this.getIndex(ID).lat;
		
	}

	public double getLongitude(String ID) {
		
			return this.getIndex(ID).lon;
		
	}

    public boolean hasExactCoords(String ID) {
    	DBline dbl = this.getIndex(ID);
    	if (dbl == null) {
    		return false;
    	} else {
    		return dbl.exact;
    	}
    }

	public double getDislocalization(String ID) {
		return this.getIndex(ID).d_offset;

	}

	public boolean isModel(String ID) {
		return this.getIndex(ID).model;

	}

	/**
	 * 
	 * @param ID is the reference name that is going to be searched from DB.
	 * @param type is the variable type, e.g. temperature or NO2.
	 * @param t_offset is the maturity of information in hours in respect the time of validity.
	 * @return base variance estimate for ID node with the given type and temporal offset.
	 */
	public double getBaseVar(String ID, String type, int t_offset, double duration, boolean model) {
		DBline dbl = this.getIndex(ID);
		double var;
		
	//	if (dbl == null) { // was not listed
		
			var = this.getGenericVarEstimate(type, model);
	      
			double tMultiplier = (1.0 + (double)t_offset/24.0);
			tMultiplier = Math.sqrt(tMultiplier);
			
			double durMultiplier = Math.sqrt(duration);
			if (durMultiplier < 1) durMultiplier =1;
			return var*tMultiplier*durMultiplier;	
	/*		
		} else {
		
		double generic = this.getGenericVarEstimate(type, dbl.model);
        var = 1.5*dbl.imprecision*generic; //TODO this is dummy
		double tMultiplier = (1.0 + (double)t_offset/24.0);
		tMultiplier = Math.sqrt(tMultiplier);
		return var*tMultiplier;
		}
		
*/

	}

	public ArrayList<String[]> readLocations(String rootdir) {
		ArrayList<String[]> locations = new ArrayList<String[]>();
		String Z = System.getProperty("file.separator");

		String file = rootdir + "data" + Z + "location.txt";
		System.out.println("DB reading:  location file is " + file);

		try {
			int lineNumber = 0;
			BufferedReader br = new BufferedReader(new FileReader(file));
			String line;

			while ((line = br.readLine()) != null) {

				lineNumber++;

				String[] line_values = line.split(";", 10);

				boolean isValid = false;
				try {// validity test - 4th element should be a number
						// double lat = Double.valueOf(line_values[3]);
					isValid = true;
				} catch (NumberFormatException e) {

				} catch (ArrayIndexOutOfBoundsException ex) {

				}
				if (isValid)
					locations.add(line_values);

			}
			br.close();
		}

		catch (FileNotFoundException e) {
			System.out.println("DB-file: file NOT found!");

			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		System.out.println("Locations added: " + locations.size());
		return locations;
	}



	public double calculateVariance(String type, Observation obs, double lat,
			double lon, Datetime ens_datetime) {

		double d2, d1, t4, t3, t2, t1, t_cutoff, d_switch; // parameters for
															// t-polynomial and
															// distance function

		double dist = obs.calc_spatial_distance(lat, lon) / 1000; // in km
		double t = Math
				.abs(obs.datetime.hourCounter - ens_datetime.hourCounter);
		boolean useCompatibilty = false;

		double a;
		double total_var = 0;
		double t_var;
		double dist_var;
		dist = Math.max(dist, obs.d_offset);

		if (type.equals("SO2") || type.equals("SO2_AirPollutantValue")) {
			useCompatibilty = true;
			t4 = 0;
			t3 = 0.0015;
			t2 = -0.0934;
			t1 = 1.874;
			t_cutoff = 40;

			d1 = 1;
			double b = 1 / 3;
			d2 = d1 * b;
			d_switch = 20;
			a = (1 - b) * d_switch * d1;

		} else if (type.equals("PM2.5")
				|| type.equals("PM2.5_AirPollutantValue")) {
			useCompatibilty = true;
			t4 = 0;
			t3 = 0.004;
			t2 = -0.2799;
			t1 = 7.1896;

			t_cutoff = 31;

			d1 = 3;
			double b = 1 / 3;
			d2 = d1 * b;
			d_switch = 20;
			a = (1 - b) * d_switch * d1;

		} else if (type.equals("CO") || type.equals("CO_AirPollutantValue")) {
			useCompatibilty = true;
			t4 = 0.3374; // about 17000 increase in 10h. R^2 = 0.998
			t3 = -16.084;
			t2 = 145.44;
			t1 = 1651.3;
			t_cutoff = 31;

			d1 = 100;
			double b = 1 / 3;
			d2 = d1 * b;
			d_switch = 20;
			a = (1 - b) * d_switch * d1;

		} else if (type.equals("O3") || type.equals("O3_AirPollutantValue")) {
			useCompatibilty = true;
			t4 = 0.0105; // about 800 increase in 10h. R^2 = 0.9987
			t3 = -0.4687;
			t2 = 2.799;
			t1 = 85.227;
			t_cutoff = 33;

			d1 = 23.4;
			double b = 1 / 3;
			d2 = d1 * b;
			d_switch = 20;
			a = (1 - b) * d_switch * d1;

		} else if (type.equals("NO2") || type.equals("NO2_AirPollutantValue")) {
			useCompatibilty = true;
			// d2=35.65; // about 800 increase in 10h. R^2 = 0.9987
			// d1=-7.16;
			d1 = 23.4;
			double b = 1 / 3;
			d2 = d1 * b;
			d_switch = 20;
			a = (1 - b) * d_switch * d1;

			t4 = 0.0023 * 2;
			t3 = -0.05 * 2;
			t2 = 72.9 * 2;
			t1 = 37.41 * 2;
			t_cutoff = 33;

		} else if (type.equals("PM10") || type.equals("PM10_AirPollutantValue")) {
			useCompatibilty = true;
			t4 = 0.0;// about 200 increase in 10h. R^2 = 0.9944
			t3 = 0.0302;
			t2 = -1.7555;
			t1 = 32.061;

			t_cutoff = 33;

			d1 = 10;
			double b = 1 / 3;
			d2 = d1 * b;
			d_switch = 20;
			a = (1 - b) * d_switch * d1;

		} else if (type.equals("temperature")) {
			d2 = 0;// about 7 increase in 10h. R^2 = 0.9987
			d1 = 0.1129;

			t4 = 0.00005;
			t3 = -0.00036;
			t2 = 0.0711;
			t1 = 0.3737;

			t_cutoff = 50;
			d_switch = 40000;// no switch

			a = (1 - 1 / 3) * d_switch;

		} else if (type.equals("windSpeed")) {
			d2 = 0;// about 7 increase in 10h. R^2 = 0.9993
			d1 = 0.14;

			t4 = 0.0000005;
			t3 = -0.00033;
			t2 = 0.0247;
			t1 = 0.5793;

			t_cutoff = 50;
			d_switch = 40000;// no switch

			a = (1 - 1 / 3) * d_switch;

		} else if (type.equals("windDirection")) {
			d2 = 0;
			d1 = 120 / 4;

			t4 = 0;
			t3 = 0;
			t2 = -16.823 / 4;
			t1 = 1444 / 4;

			t_cutoff = 50;
			d_switch = 40000;// no switch

			a = (1 - 1 / 3) * d_switch;

		} else if (type.equals("rain")) {
			d2 = 0;
			d1 = 0.1129;

			t4 = 0.00005;
			t3 = -0.00036;
			t2 = 0.0711;
			t1 = 0.3737;

			t_cutoff = 50;
			d_switch = 40000;// no switch

			a = (1 - 1 / 3) * d_switch;

		} else if (type.equals("skyCondition")) {
			d2 = 0;
			d1 = 0.1129 * 6;

			t4 = 0;
			t3 = 0.0011;
			t2 = -0.0762;
			t1 = 1.7829;

			t_cutoff = 35;
			d_switch = 40000;// no switch

			a = (1 - 1 / 3) * d_switch;

		} else {// GENERIC
			t4 = 0;
			t3 = 0.004;
			t2 = -0.2799;
			t1 = 7.1896;

			d2 = 3.15;
			d1 = 1.36;

			t_cutoff = 40;
			d_switch = 2;

			a = (1 - 1 / 3) * d_switch;

		}
		// //General var calculation///////////////////////////////////
		if (t > t_cutoff)
			t = t_cutoff;
		t_var = t4 * t * t * t * t + t3 * t * t * t + t2 * t * t + t1 * t; // about
																			// 10
																			// increase
																			// in
																			// 10h.
																			// R^2
																			// =
																			// 0.952

		if (dist < d_switch) {
			dist_var = d1 * dist;
		} else {
			dist_var = d2 * dist + a;
		}
		// dist_var = dist_var/3.0;
		double compatibility = 1.0;

		total_var = compatibility * dist_var + t_var + obs.base_var;

		return total_var;
		// ////////////////////////////////////////////////////////////////////////
	}

public void replaceAndStore(ArrayList<String> response) {
	
	this.db.clear();
	
	for (int i=0;i<response.size();i++) {
		DBline dbl = new DBline(response.get(i));
		this.db.add(dbl);
	}
	
	this.storeDB(response);
	
}

private void storeDB(ArrayList<String> response) {
	
	java.io.File file = new java.io.File(""); // Dummy file
	String abspath = file.getAbsolutePath();
	String Z = System.getProperty("file.separator");
	
	String filename = abspath + Z + "data" + Z + "db.csv";
	
	String n = System.getProperty("line.separator");
	  
	   
boolean append = false;
	 try { 
	        BufferedWriter out = new BufferedWriter(new FileWriter(filename, append));
	       for (int i=0;i<response.size();i++) {
	         String line = response.get(i);    
	         out.write(line+ n); 
	       } 
	       
	        out.close(); 
	        } catch (IOException e) { 
	        }  

	
}

}// class END

