package services.pescado.fusionservice.impl.database;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import services.pescado.fusionservice.impl.Masks.Env;
import services.pescado.fusionservice.impl.fuser.Datetime;

public class EqProfiles {


    public ArrayList<EqProfile> eqProfiles;
         
    private double cutoff = 0.0;
    
    public EqProfiles(String rootdir) {
    
      this.eqProfiles = new ArrayList<EqProfile>();
      this.readEqProfiles(rootdir);
      
      
      
    }
    
    public void readEqProfiles(String rootdir) {

               
       String Z = System.getProperty("file.separator");
       String file = rootdir + "data"+Z+"eqProfiles.csv";
	System.out.println("Profiles db reading: file is "+file);			

	try {
           int lineNumber =0;
	BufferedReader br = new BufferedReader( new FileReader(file));
	String line;
       String lastType ="";
       ArrayList<String[]> profData = new ArrayList<String[]>();
	while( (line = br.readLine()) != null)  {
           
		lineNumber++;
               
               if (lineNumber ==1) continue;      
               String[] line_values = line.split(";",30);
               if (lineNumber ==2) lastType = line_values[0];
                String currType = line_values[0];
                //String prof = line_values[1];
               
                if (currType.equals(lastType)) {
                    profData.add(line_values);
                    //System.out.println("Adding "+line_values[0] + "\t" +line_values[1] + "\t" + line_values[2]);
                
                } else {//new type
                 EqProfile eqProfile = new EqProfile(profData);
                 this.eqProfiles.add(eqProfile);
                 System.out.println("Added eq.profile set:  "+ lastType);
                 profData.clear();
                 profData.add(line_values);
                 lastType = currType;
                 
                }
              				 
		 
		}
       br.close();
}

	catch (FileNotFoundException e) {
		System.out.println("DB-file: file NOT found!");
	
	//e.printStackTrace();
	} catch (IOException e) {
	// TODO Auto-generated catch block
	e.printStackTrace();
	}

    }

public EqProfile getEqProfile(int typeInt) {
   
   for (int i=0;i<this.eqProfiles.size();i++) {
       if (eqProfiles.get((i)).typeInt == typeInt) {
           return eqProfiles.get(i);
       }
   }
   return null;
   
}

public EqProfile getEqProfile(String type) {
   
   for (int i=0;i<this.eqProfiles.size();i++) {
       if (eqProfiles.get((i)).type.equals(type)) {
           return eqProfiles.get(i);
       }
   }
   return null;
   
}


//this method returns a corrected observation value which is a non-biased estimator for the target location.
//Example: target profile average at 15.00 is 20.0, estimator location profile average at 14.00 is 35.0. Estimator value at 14.00 is 50.0. Then a nonbiased estimator is 50 + (20.0 -35.0) = 35.0 
public double getProfileCorrectedValue(double value, String type, Env valueEnv,Env esEnv,Datetime valueDt, Datetime esDt, double distance, double correctionStrength) {
   
   EqProfile eqp = this.getEqProfile(type);
   if (eqp == null) {
       //System.out.println(type +" not available in eqProfs.");
       return value;
   } else if (esEnv.override || valueEnv.override) {
     return value;  
   }
   
   double correction =0;
   
  
   double valueAve = eqp.getAverage(valueEnv, valueDt);
   //System.out.println("target ave:"+valueAve);
   double esAve = eqp.getAverage(esEnv, esDt);
   //System.out.println("estimator ave:"+esAve);
   correction =esAve - valueAve;
   
   correction = correction*correctionStrength;
   //System.out.println("Correction after scaling: "+correction);
   //System.out.println("target = " +targetProf + ", estimator = "+obs.relevance + " => correction term = "+correction);
   if (distance <this.cutoff) correction = distance/cutoff*correction;
 
   double correctedVal = correction + value;
 
    if (!eqp.allowNegative) {
        if (correctedVal < 0) correctedVal = 0;       
    }
    return correctedVal;
}


	
}
