package services.pescado.fusionservice.impl.datapack;

import java.util.ArrayList;

import services.pescado.fusionservice.impl.fuser.Datetime;

/**
 * 
 * @author johanssl
 */
public class GridSource extends Source {
	ArrayList<Layer> layers;
	boolean layered = false;
	int avgLayerSize;

	// public ArrayList<Observation> source_obs;
	// public String ID;

	// public double baseSTD;
	// public String profile;
	// public String type;
	// public int amountOfObs;

	// Constructor
	public GridSource(ArrayList<Observation> obsList, int res,
			double[] boundaries) {
		super(obsList);

		this.createLayers(res, boundaries);

	}

	public GridSource(Observation obs) {
		super(obs);
		this.layers = new ArrayList<Layer>();
	}

	@Override
	public void createSinglelayer(int resolution, double[] boundaries,
			Datetime dt) {

		if (this.layered) {
			System.out.println(this.ID + " - already fully layered.");
			return;
		}
		boolean alreadyContains = false;
		for (int i = 0; i < this.layers.size(); i++) {
			if (this.layers.get(i).layerHour == dt.hourCounter)
				alreadyContains = true;
		}
		if (alreadyContains)
			return;

		ArrayList<String> timelayers = new ArrayList<String>();

		// get layers to be formed
		for (int i = 0; i < this.source_obs.size(); i++) {
			String timestamp = this.source_obs.get(i).datetime.getStringDate();
			if (!timelayers.contains(timestamp))
				timelayers.add(timestamp);
		}

		if (timelayers.isEmpty()) {
			System.out.println("No time layers!");
			return;
		}
		// lets search for the closest layer
		String closest = "";
		int minSeparation = Integer.MAX_VALUE;
		long hourCounter = dt.hourCounter;
		for (int i = 0; i < timelayers.size(); i++) {

			Datetime ldt = new Datetime(timelayers.get(i));
			long layerCounter = ldt.hourCounter;
			int sep = (int) Math.abs(hourCounter - layerCounter);
			if (sep < minSeparation) {
				minSeparation = sep;
				closest = timelayers.get(i);
			}
		}

		// form a layer

		ArrayList<Observation> layerObs = new ArrayList<Observation>();
		for (int j = 0; j < this.source_obs.size(); j++) {

			if (closest.equals(this.source_obs.get(j).datetime.getStringDate())) {
				layerObs.add(this.source_obs.get(j));

				// System.out.println("...adding "+obs.get(j).value);
			}

		}
		System.out.println("GridSource - " + this.ID
				+ " -forming a single layer: " + closest);
		this.layers.add(new Layer(layerObs, resolution, boundaries));

	}

	// this method creates separate and continuous layers from gridded data
	@Override
	public void createLayers(int resolution, double[] boundaries) {
		if (this.layered) {
			System.out.println(this.ID + " - already layered.");
			return;
		}

		ArrayList<String> timelayers = new ArrayList<String>();
		this.layers = new ArrayList<Layer>();

		// get layers to be formed
		for (int i = 0; i < this.source_obs.size(); i++) {
			String timestamp = this.source_obs.get(i).datetime.getStringDate();
			if (!timelayers.contains(timestamp))
				timelayers.add(timestamp);
		}
		System.out.println("GridSource - " + this.ID
				+ " -time layers to be formed: " + timelayers.size());

		// form layers
		for (int i = 0; i < timelayers.size(); i++) {
			// System.out.println("Layer "+i);
			ArrayList<Observation> layerObs = new ArrayList<Observation>();
			for (int j = 0; j < this.source_obs.size(); j++) {

				if (timelayers.get(i).equals(
						this.source_obs.get(j).datetime.getStringDate())) {
					layerObs.add(this.source_obs.get(j));

					// System.out.println("...adding "+obs.get(j).value);
				}

			}
			this.layers.add(new Layer(layerObs, resolution, boundaries));

		}
		this.avgLayerSize = this.source_obs.size() / this.layers.size();
		this.layered = true;
	}

	public double getValue(double lat, double lon, Datetime date) {
		int minSeparation = Integer.MAX_VALUE;
		long hour = date.hourCounter;
		double value = 0;
		int index = 0;
		for (int i = 0; i < this.layers.size(); i++) {
			int separation = (int) Math
					.abs(hour - this.layers.get(i).layerHour);
			if (separation < minSeparation) {
				index = i;
				minSeparation = separation;
				if (minSeparation == 0) {
					break;
				}
			}
		}
		value = this.layers.get(index).getValue(lat, lon)[0];
		return value;
	}

	@Override
	// this method seeks the closest layer and then gets the closest observation
	// in that layer based on layer location indexing
	public Observation get_best_obs(Datetime date, double lat, double lon,
			boolean smooth) {
		int minSeparation = Integer.MAX_VALUE;
		long hour = date.hourCounter;
		// double[] value = new double[2];

		int index = -1;
		for (int i = 0; i < this.layers.size(); i++) {
			int separation = (int) Math
					.abs(hour - this.layers.get(i).layerHour);
			if (separation < minSeparation) {
				index = i;
				minSeparation = separation;

				if (minSeparation == 0) {
					break;
				}
			}
		}

		if (index == -1) {// no layer was found
			Observation obs2 = this.getBestObsBruteForce(date, lat, lon);
			return obs2;

		}

		// Observation obs;
		// if (this.layers.get(index).layerSize*2 < this.avgLayerSize ) { //
		// sometimes layers do not contain equal amount of observations.
		// obs = this.getBestObsBruteForce(date, lat, lon);
		// System.out.println(this.ID+": brute force for" +
		// date.getStringDate());
		// } else {
		Observation obs = this.layers.get(index).getBestLayerObs(lat, lon,
				smooth);
		// }

		// value = this.layers.get(index).getValue(lat, lon);
		// int t_offset = this.layers.get(index).t_offset;
		// String type, String ID, double latitude, double longitude, Datetime
		// dt, double value,double baseVar, boolean model, String relevance
		// Observation obs = new Observation(this.type, this.ID, lat, lon, date,
		// value[0],value[1], true, "Override",value[1],t_offset);
		// obs = this.source_obs.get((int)value[1]);
		// obs.value = value[0];
		return obs;

	}

}// class end
