package services.pescado.fusionservice.impl.datapack;

import java.util.ArrayList;

/**
 * 
 * @author johanssl
 */
public class Layer {
	ArrayList<Observation> layerObs;
	int t_offset;
	String layerID;
	long layerHour;
	double latmin;
	double lonmin;
	double latmax;
	double lonmax;
	double dlat;
	double dlon;
	double[][] smoothedValues;
	double[][] obsIndex;
	int res;
	int layerSize;

	public Layer(ArrayList<Observation> layerObs, int resolution,
			double[] boundaries) {
		this.layerObs = layerObs;
		this.layerSize = layerObs.size();
		this.latmin = boundaries[2];
		this.latmax = boundaries[3];
		this.lonmin = boundaries[0];
		this.lonmax = boundaries[1];
		this.dlat = (latmax - latmin) / resolution;
		this.dlon = (lonmax - lonmin) / resolution;

		this.t_offset = layerObs.get(0).t_offset;
		this.layerID = layerObs.get(0).datetime.getStringDate();
		this.layerHour = layerObs.get(0).datetime.hourCounter;
		this.smoothedValues = new double[resolution][resolution];
		this.obsIndex = new double[resolution][resolution];
		this.res = resolution;
		for (int w = 0; w < resolution; w++) {
			for (int h = 0; h < resolution; h++) {
				double lat = this.latmax - dlat * h;
				double lon = this.lonmin + dlon * w;

				double[] value = this.getExtrapolation(lat, lon);
				this.smoothedValues[h][w] = value[0];
				this.obsIndex[h][w] = value[1];
				// System.out.println(lat+", "+lon+", "+value[0]);

			}
		}

	}

	public double[] getExtrapolation(double lat, double lon) {
		double shortestDist = Double.MAX_VALUE;
		int size = this.layerObs.size();
		int obsIndex = 0;
		double[][] vals = new double[size][2];
		for (int i = 0; i < size; i++) {
			double dist = this.layerObs.get(i).calc_spatial_distance(lat, lon) / 1000.0;// dislocalization
																						// added
			vals[i][0] = this.layerObs.get(i).value;
			vals[i][1] = 1 / (dist * dist);
			if (dist < shortestDist) {
				shortestDist = dist;
				obsIndex = i;
			}
		}
		double weightSum = 0;
		for (int i = 0; i < size; i++) {
			weightSum = weightSum + vals[i][1];
		}
		// System.out.println("weightsum: "+weightSum);
		double value = 0;
		for (int i = 0; i < size; i++) {
			value = value + vals[i][0] * vals[i][1] / weightSum;
		}

		double bv = obsIndex;
		double[] temp = { value, bv };
		return temp;
	}

	public double[] getValue(double lat, double lon) {
		int[] indices = this.getIndex(lat, lon);
		int h = indices[0];
		int w = indices[1];
		double[] temp = { this.smoothedValues[h][w], this.obsIndex[h][w] };
		return temp;
	}

	public int[] getIndex(double lat, double lon) {

		// upper left corner is maxlat minlon;
		int h = (int) ((this.latmax - lat) / this.dlat);
		int w = (int) ((lon - this.lonmin) / this.dlon);

		int[] returnable = { h, w };
		return returnable;
	}

	public Observation getBestLayerObs(double lat, double lon, boolean smooth) {
		int h = (int) ((this.latmax - lat) / this.dlat);
		int w = (int) ((lon - this.lonmin) / this.dlon);

		int index = (int) this.obsIndex[h][w];
		Observation obs = this.layerObs.get(index);

		if (smooth) {
			Observation newObs = new Observation(obs.type, obs.ID,
					obs.latitude, obs.longitude, obs.datetime,
					this.smoothedValues[h][w], obs.base_var, obs.model,
					obs.env, obs.d_offset, obs.t_offset, obs.opDataType, obs.duration);
			// obs.value = this.smoothedValues[h][w];
			return newObs;
		} else {
			return obs;
		}
	}

}// class end
