package services.pescado.fusionservice.impl.datapack;

/**
 *
 * @author johanssl
 */

import java.util.ArrayList;

import services.pescado.fusionservice.impl.Masks.Env;
import services.pescado.fusionservice.impl.fuser.Datetime;

public class Source {

	public ArrayList<Observation> source_obs;
	public String ID;
	public double baseSTD;
	public Env env;
	public String type;
	public int amountOfObs;
	public boolean model;
    public boolean banned = false;

	public double popDensity;

	// Constructor
	public Source(Observation obs) {
		this.source_obs = new ArrayList<Observation>();
		this.amountOfObs = 0;
		this.model = obs.model;
		this.ID = obs.ID;

		this.env= obs.env;
		this.baseSTD = obs.base_var;

		this.add_observation(obs);
	}

	// Constructor
	public Source(ArrayList<Observation> obsList) {
		this.source_obs = obsList;
		this.amountOfObs = 0;
		this.model = obsList.get(0).model;
		this.ID = obsList.get(0).ID;

		this.env = obsList.get(0).env;
		this.baseSTD = obsList.get(0).base_var;

	}



	// this methods is inteded for large source_obs arrays - for extensive time
	// series
	// returns hourly (00-23) averages based on measurements
	public double[] getHourAverages() {
		double[] returnable = new double[24];
		double[][] hourAve = new double[24][2];

		for (int i = 0; i < source_obs.size(); i++) {
			int hour = source_obs.get(i).datetime.hour;

			hourAve[hour][0] = hourAve[hour][0] + source_obs.get(i).value; // summing
																			// values
			if (source_obs.get(i).value > 0.0)
				hourAve[hour][1] = hourAve[hour][1] + 1; // amount of
															// observations

		}

		for (int i = 0; i < 24; i++) {
			returnable[i] = hourAve[i][0] / hourAve[i][1]; // get average
		}

		for (int i = 1; i < 23; i++) { // this fixes a curious bug. TODO
										// investigate the actual BUG!
			if (returnable[i] == 0 && returnable[i + 1] != 0) {
				returnable[i] = returnable[i + 1];
			} else if (returnable[i] == 0 && returnable[i - 1] != 0) {
				returnable[i] = returnable[i - 1];
			}
		}

		return returnable;
	}

	public void add_observation(Observation obs) {
		this.source_obs.add(obs);
		this.amountOfObs++;
	}

	public boolean itsMe(String ID) {
		if (this.ID == ID) {
			return true;
		} else {
			return false;
		}
	}

	// This method presumes that each source has fixed coordinates and therefore
	// fixed zone. The only variable is then the time
	// This method returns the observation which is timewise the closest.
	public Observation get_best_obs(Datetime ens_datetime, double lat,
			double lon, boolean smooth) {
		int index = 0;
		double smallestVar = Double.MAX_VALUE;
		for (int i = 0; i < this.source_obs.size(); i++) {
			Observation obs = source_obs.get(i);

			double var = this.calcGenericVar(obs, lat, lon, ens_datetime);

			if (var < smallestVar) {
				smallestVar = var;
				index = i;
			}
		}
		return this.source_obs.get(index);

	}

	// This method presumes that each source has fixed coordinates and therefore
	// fixed zone. The only variable is then the time
	// This method returns the observation which is timewise the closest.
	public Observation getBestObsBruteForce(Datetime ens_datetime, double lat,
			double lon) {
		int index = 0;
		double smallestVar = Double.MAX_VALUE;
		for (int i = 0; i < this.source_obs.size(); i++) {
			Observation obs = source_obs.get(i);

			double var = this.calcGenericVar(obs, lat, lon, ens_datetime);

			if (var < smallestVar) {
				smallestVar = var;
				index = i;
			}
		}
		return this.source_obs.get(index);

	}

	public String printoutObs() {
		String n = System.getProperty("line.separator");
		String output = "";
		Observation[] obs;

		int length = this.source_obs.size();
		obs = new Observation[length];
		source_obs.toArray(obs);

		for (int i = 0; i < length; i++) {
			if (!this.model) {
				output = output + obs[i].datetime.getStringDate() + ":  "
						+ obs[i].value + n;
			} else {
				output = output + obs[i].datetime.getStringDate() + ":  "
						+ obs[i].value + ", " + obs[i].latitude + ", "
						+ obs[i].longitude + ", " + obs[i].t_offset + ", "
						+ obs[i].base_var + n;
			}
		}

		return output;

	}

	public double calcGenericVar(Observation obs, double lat, double lon,
			Datetime ens_datetime) {
		double d2, d1, t4, t3, t2, t1, t_cutoff; // parameters for t-polynomial
													// and distance function

		double dist = obs.calc_spatial_distance(lat, lon);
		double t = obs.calc_temporal_distance(ens_datetime);

		double total_var = 0;
		double baseVar = obs.base_var;
		double t_var;
		double dist_var;

		// COMMON modifications////
		t = t / 3600;// in hours
		if (t < 0)
			t = 0;

		d2 = 0;
		d1 = 23.4;
		t4 = 0.0023;
		t3 = -0.05;
		t2 = 72.9;
		t1 = 37.41;

		t_cutoff = 33;

		// //General var calculation///////////////////////////////////
		if (t > t_cutoff)
			t = t_cutoff;
		t_var = t4 * t * t * t * t + t3 * t * t * t + t2 * t * t + t1 * t; // about
																			// 10
																			// increase
																			// in
																			// 10h.
																			// R^2
																			// =
																			// 0.952

		dist_var = d2 * dist * dist + d1 * dist;

		total_var = dist_var + t_var + baseVar;
		return total_var;
	}

	// this is a dummy method which does nothing - GridSource Overrides this
	public void createLayers(int resolution, double[] boundaries) {

	}

	public double getLat() {
		return -1.0;
	}

	public double getLon() {
		return -1.0;
	}

	public void createSinglelayer(int resolution, double[] boundaries,
			Datetime dt) {

	}

}// Class end bracket
