package services.pescado.fusionservice.impl.datapack;


import java.util.ArrayList;
import services.pescado.fusionservice.impl.fuser.Datetime;

/**
 * 
 * @author johanssl
 */
public class VariableData {
	public ArrayList<Source> sources;
	public int size;
	public String type;
	public Datetime start;
	public Datetime end;
	public String unit;
	public double[] boundaries;
	public boolean evalProfiles = false;

	// pescado constructor
	public VariableData(Observation obs) {
		this.type = obs.type;
		this.sources = new ArrayList<Source>();
		this.unit = this.getUnit(type);

		if (type.equals("NO2") || type.equals("SO2") || type.equals("PM2.5")
				|| type.equals("PM10") || type.equals("CO")
				|| type.equals("O3")) {
			this.evalProfiles = true;
		} else {
			this.evalProfiles = false;
		}

		this.sources.add(new Source(obs));
		this.size = 1;
		
		
			this.start = obs.datetime;	
			this.end = obs.datetime;


	}

	public void addObservation(Observation obs) {

		if (!obs.type.equals(this.type)) {
			System.out.println("Observation couldn't be added -wrong TYPE!");
			return;
		}

		boolean newSource = true;
		for (int i = 0; i < this.sources.size(); i++) {

			if (obs.ID.equals(this.sources.get(i).ID)) {
				this.sources.get(i).add_observation(obs);
				newSource = false;
				break;
			}
		}

		if (newSource) {
			this.sources.add(new Source(obs));
		}
		this.size++;
		// update vd start and end dates
		if (this.start == null)
			this.start = obs.datetime;
		if (this.end == null) {
			this.end = new Datetime(obs.datetime.getStringDate());
			this.end.addSeconds((int)(3600*obs.duration));
		}

		if (this.start.hourCounter > obs.datetime.hourCounter)
			this.start = new Datetime(obs.datetime.getStringDate());
		
		if (this.end.hourCounter < obs.datetime.hourCounter + obs.duration) {
			this.end = new Datetime(obs.datetime.getStringDate());
			this.end.addSeconds((int)(3600*obs.duration));
		}
	}

	

	

	double[] getCornerCoordinatesFromData(double excess) {
		double[] common_boundary = { 180.0, -180.0, 90.0, -90.0 };

		int i;
		// double lonmin=0,lonmax=0,latmin =0, latmax =0;
		for (i = 0; i < this.sources.size(); i++) {
			for (int j = 0; j < this.sources.get(i).source_obs.size(); j++) {
				Observation obs = this.sources.get(i).source_obs.get(j);

				if (common_boundary[0] > obs.longitude) {// lonmin gets smaller
					common_boundary[0] = obs.longitude;
				}
				if (common_boundary[1] < obs.longitude) {// lonmax gets bigger
					common_boundary[1] = obs.longitude;
				}
				if (common_boundary[2] > obs.latitude) {// latmin gets smaller
					common_boundary[2] = obs.latitude;
				}
				if (common_boundary[3] < obs.latitude) {// latmax gets bigger
					common_boundary[3] = obs.latitude;
				}

			}
		}

		double lon_excess = (common_boundary[0] - common_boundary[1]) * excess; // we
																				// want
																				// to
																				// make
																				// d2
																				// slighly
																				// larger
																				// grid
																				// than
																				// the
																				// smallest
																				// possible
		double lat_excess = (common_boundary[2] - common_boundary[3]) * excess;
		common_boundary[0] = common_boundary[0] + lon_excess;
		common_boundary[1] = common_boundary[1] - lon_excess;
		common_boundary[2] = common_boundary[2] + lat_excess;
		common_boundary[3] = common_boundary[3] - lat_excess;

		return common_boundary;
	}

	public void createSingleLayerforGridSources(int res, Datetime dt) {
		for (int i = 0; i < this.sources.size(); i++) {
			this.sources.get(i).createSinglelayer(res, this.boundaries, dt);
		}
	}

	public void createlayersForGridSources(int res, double[] boundaries) {
		for (int i = 0; i < this.sources.size(); i++) {
			this.sources.get(i).createLayers(res, boundaries);
		}
	}

	private String getUnit(String type) {
		String unit = "N/A";

		String[][] options = { { "PM2.5", "g/m3" }, { "PM10", "g/m3" },
				{ "SO2", "g/m3" }, { "NO2", "g/m3" }, { "CO", "g/m3" },
				{ "O3", "g/m3" }, { "temperature", "C" }, { "rain", "mm/h" },
				{ "windSpeed", "m/s" }, { "windDirection", "degree" } };

		for (int i = 0; i < options.length; i++) {
			if (type.equals(options[i][0]))
				unit = options[i][1];
		}

		return unit;

	}

	

	// this method tries to parse a number out of String. If this cannot be
	// done, then a value is crunched based on the variable type
	public double formatValue(String type, String strValue, String source) {
		double val = Double.NaN;
		boolean success = false;
		try {
			val = Double.valueOf(strValue);
			success = true;

			if (type.equals("cloudiness")
					&& (source.equals("Wund") || source.equals("yrno"))) {
				int temp = (int) (val * 8 / 100);
				val = temp;
			}

		} catch (NumberFormatException e) {
			// could not parse directly to value
		}

		if (success) {
			return val;
		} else { // we need to process the String a bit further

			// perhaps the String is a wind direction...
			if (strValue.equals("N")) {
				val = 0.0;
			} else if (strValue.equals("NE") || strValue.equals("EN")) {
				val = 45.0;
			} else if (strValue.equals("E")) {
				val = 90.0;
			} else if (strValue.equals("ES") || strValue.equals("SE")) {
				val = 135.0;
			} else if (strValue.equals("S")) {
				val = 180.0;
			} else if (strValue.equals("SW") || strValue.equals("WS")) {
				val = 225.0;
			} else if (strValue.equals("W")) {
				val = 270.0;
			} else if (strValue.equals("NW") || strValue.equals("WN")) {
				val = 315.0;
			}

			return val;
		}

	}

	

	public boolean hasData(String ID) {
		for (int i = 0; i < this.sources.size(); i++) {
			if (ID.equals(sources.get(i).ID))
				return true;
		}
		return false;
	}

	private void addSourceToPack(Observation obs) {
		boolean station = true;
		String inf = "stationary";
		if (obs.model) {
			station = false;
			inf = "grid";
		}
		System.out.println("Adding " + inf + " source: " + obs.ID);
		if (station) {
			this.sources.add(new StationarySource(obs));
		} else {
			this.sources.add(new GridSource(obs));
		}

	}

	// this method searches the proper source from list and adds obs to it. If
	// the correct source is not found then a new source is added.
	public void addObsToPack(Observation obs) {
		boolean added = false;
		for (int i = 0; i < this.sources.size(); i++) {
			if (obs.ID.equals(sources.get(i).ID)) {
				sources.get(i).add_observation(obs); // add to correct source
				added = true;
				break;
			}
			// source was not found and thus it is added to arrayList

		}
		if (!added)
			this.addSourceToPack(obs);
	}

	// This method adjusts read values if necessary - if this returns null then
	// no "observation" isntance is created for this value
	public double correctValues(double value, String type) {

		// boolean trash = false;

		if (!type.equals("temperature")) {
			if (value < 0)
				value = 0;
		}

		return value;
	}

	public void addSource(Source dats) {
		this.sources.add(dats);
	}

}// end class
