package services.pescado.fusionservice.impl.fuser;

/**
 *
 * @author johanssl
 */

//

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;

import services.pescado.fusionservice.impl.Masks.Env;
import services.pescado.fusionservice.impl.Masks.MapPack;
import services.pescado.fusionservice.impl.database.DataBase;
import services.pescado.fusionservice.impl.database.Options;
import services.pescado.fusionservice.impl.datapack.DataPack;
import services.pescado.fusionservice.impl.datapack.Observation;
import services.pescado.fusionservice.impl.datapack.Source;
import services.pescado.fusionservice.impl.datapack.VariableData;

public class Ensemble {
	/*
	 * Lasse Johansson 2011 FMI This class controls model instances and combines
	 * their estimation values etc.
	 */
	// TODO it should be possible to unify slurped observations and
	// model-classified lists.

	// ============================//fusion results, zones and densities
	private Double[][] ensemble;
	private Double[][] ensembleSTD;
	private int[][] density;
	private char[][] zones;
	private Double[][] source_w;

	// ==============================

	// measurements list for comparison
	ArrayList<Observation> obsList;
	boolean hasMeasurements = false;

	public double totExposure;
	public double cellArea;
	public int totPersons;
	public boolean useDensity = true;

	public int profileRadius = 150;
	private int height, width;
	// timeseries
	public double[][] timeseries;

	// public String rootdir;
	public String favoredColorScheme = "basic";
	public String output;
	public double ens_maximum, ens_minimum, ens_STD_maximum, ens_STD_minimum,
			ens_w_maximum, ens_w_minimum;
	public String dateNow;
	// public double[] boundaries; // 0 lonmin,1 lonmax,2 latmin,3 latmax

	public MapPack mapPack;
	public Datetime datetimeStart;
	public Datetime datetimeEnd;
	public Datetime datetimeEns;// current date for fusion
	public DataBase DB = null;

	// public String type;
	// public String unit;

	// public Observation[] obs; // TODO obsolete
	// public Datasource[] list_of_sources; // TODO obsolete
	public DataPack datapack;

	public int source_index;// if weight printing, then this index points to the
							// correct source
	public ArrayList<String> bannedSources = null;
	public boolean someIsBanned = false;
	// fusion methods and options============

	public boolean trust_measurements = false;
	public boolean totalNerf = false;
	public boolean imbaNerf = false;
	public boolean autoCorrect = false;
	public boolean averages = false;
	public boolean transparency;
	public boolean sectorizeAll = false;
	// =====================================
	// private Hashtable<String, String[]> database; //db hash

	public Color[] cols = new Color[40];
	public String[] vals = new String[40];

	public Ensemble(DataPack dp, DataBase DB, MapPack maps) {
		this.DB = DB;
		this.datapack = dp;
		this.datetimeStart = dp.start;
		this.datetimeEnd = dp.end;
		this.mapPack = maps;
		// this.profiles = profiles;
		// this.type = dp.type;
		// this.rootdir = rootdir;
		// this.unit = this.getUnit(type);
		// TODO test
		// DataPack test = new DataPack(rootdir,this.datetimeStart,
		// this.datetimeEnd, false,maps,DB);

		Calendar cal = Calendar.getInstance();
		SimpleDateFormat formatter = new SimpleDateFormat("(MMMdd_HHmm_ss)");
		this.dateNow = formatter.format(cal.getTime());
		System.out.println(this.dateNow);

	}

	/*
	 * This method produces an arrayList from fused values for 1) all variables
	 * 2) all datetimes between vd.start and vd.endcentlat and cent lon
	 * specifies the location for fusiontimeIncrement specifies duration and how
	 * sparsely the fusion is going to be doen between vd.start and vd.end (1 =>
	 * 1h increments)
	 */
	public ArrayList<FusionResult> getCompleteFusionResult(double centlat,
			double centlon, int timeIncrement, Env forcedEnv,Datetime startDate, Datetime endDate) {
		ArrayList<FusionResult> frList = new ArrayList<FusionResult>();


		for (int i = 0; i < this.datapack.variables.size(); i++) {
			String type = this.datapack.variables.get(i).type;
			
			if (type == null) {
				System.out.println("getCompleteFusion: Null type with variable");
				continue;
			}
			
			Datetime start = this.datapack.variables.get(i).start;
			Datetime end = this.datapack.variables.get(i).end;
			
			if (end.hourCounter > endDate.hourCounter) end = endDate; // fusion is only done for the user defined time span
			if (start.hourCounter < startDate.hourCounter) start = startDate; // fusion is only done for the user defined time span
			
			if (start == null || end == null) {
				System.out.println("FUSER: Null dates with" + type);
				continue;
			}
			
			int hours = (int) (end.hourCounter - start.hourCounter);
			Datetime current = new Datetime(start.getStringDate());

			for (int j = 0; j <= hours; j++) {

				ArrayList<Double> vals = this.calculateCombinationValue(
						this.datapack.variables.get(i), centlat, centlon,
						forcedEnv, current, false, Options.method);
				double value = vals.get(0);
				double STD = vals.get(1);
				String opDataType = this.datapack.variables.get(i).sources
						.get(0).source_obs.get(0).opDataType; // TODO does this
																// always work
																// correctly?
				Datetime cdt = new Datetime(current.getStringDate());
				frList.add(new FusionResult(this.editPrecision(value, 3), STD, type, timeIncrement,
						cdt, centlat, centlon, opDataType));

				System.out.println(type + "\t" + this.editPrecision(value, 3) + "\t"
						+ current.getStringDate() + "\t" + opDataType);

				current.addSeconds(3600 * timeIncrement);
			}

		}
		return frList;
	}

	// This method updates the String-form date stamp
	public void updateDateNow() {
		Calendar cal = Calendar.getInstance();
		SimpleDateFormat formatter = new SimpleDateFormat("(MMMdd_HHmm_ss)");
		this.dateNow = formatter.format(cal.getTime());
	}

	// This method prints recieved messages hourly counters to message.csv file
	public void dumpALtoFile(ArrayList<String> lines) {
		String n = System.getProperty("line.separator");
		boolean append;
		append = false;
		
		java.io.File file = new java.io.File(""); // Dummy file
		String abspath = file.getAbsolutePath();
		String Z = System.getProperty("file.separator");
		String filename = abspath + Z + "data" + Z + "dump" + Z + "dataDump_lastRun.txt";

		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(filename, append));
			for (int i = 0; i < lines.size(); i++) {
				String temp = lines.get(i);
				
				out.write(temp + n);
			}

			out.close();
		} catch (IOException e) {
		}

	}
	
	//this method dumps datapack to file with a simple structure so that later on it will be possible to evaluate datanode's historical performance,
	public void dumpObsToFile() {
		// create mapPack and DB
		java.io.File file = new java.io.File(""); // Dummy file
		String Z = System.getProperty("file.separator");
		String abspath = file.getAbsolutePath();
	    String dir = abspath +Z + "data" + Z + "dump"+Z;
	 int vars = this.datapack.variables.size();
	 
	 for (int i =0; i<vars;i++) {
	     
	    ArrayList<String> forcs = new ArrayList<String>();//forecast dump for each var
	    ArrayList<String> mes = new ArrayList<String>(); //pseudo measurements for each var
	    String first = "ID;TYPE;TIMELOC;VALUE;T_OFFSET;";
	    forcs.add(first);
	    mes.add(first);
	     VariableData vd = this.datapack.variables.get(i);
	        for (int s =0;s<vd.sources.size();s++) {
	           Source so = vd.sources.get(s);
	            for (int o =0;o<so.source_obs.size();o++) {
	               Observation ob = so.source_obs.get(o);
	               int latInt = (int)(ob.latitude*10);
	               int lonInt = (int)(ob.longitude*10);
	               String timeLoc = latInt + "_" + lonInt + "_"+ob.datetime.hourCounter;
	               
	               String line = ob.ID + ";" + ob.datetime.getStringDate() + ";" + timeLoc + ";" + ob.value + ";" + ob.t_offset;
	               if (ob.t_offset < 2 && ob.t_offset >= 0) {
	                   mes.add(line);
	               } else {
	                  forcs.add(line); 
	               }
	            }//for obs
	            
	        }//for sources
	   
	  if (mes.size()> 1) this.printOutALtoFile2(dir, mes, "mes_"+vd.type+ ".csv",false);
	  System.out.println("Dumping "+ (int)(mes.size()-1) +" references to file.");
	 if (forcs.size()> 1)  this.printOutALtoFile2(dir, forcs, "forcs_"+vd.type+ ".csv",false);
	 System.out.println("Dumping "+ (int)(forcs.size()-1) +" forcs to file.");
	   
	 }//for vars
	       
	}

	// This method prints recieved messages hourly counters to message.csv file
	public void printOutALtoFile2(String resultsDir, ArrayList<String> lines,
			String name, boolean append) {
		String n = System.getProperty("line.separator");
		
		append = false;

		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(resultsDir
					+ name, append));
			for (int i = 0; i < lines.size(); i++) {
				String line = lines.get(i);
				out.write(line + n);
			}

			out.close();
		} catch (IOException e) {
		}

	}

	/*
	 * This method produces geographical ensemble. Output: this method produces
	 * public Double[][] ensemble, Double[][] ensembleSTD and Double[][]
	 * source_w;
	 * 
	 * @params: resolution defines the output grid resolution (height component,
	 * width is scaled automatically) time (e.g. "YYYY-MM-DD-HH-mm") defines the
	 * ensembling time
	 */
	public void calcAreaEnsemble(String type, String time, int resolution) {

		ArrayList<Double> combVal;
		VariableData vd = this.datapack.getVarData(type);

		this.totExposure = 0;
		this.totPersons = 0;
		int selectedSource = this.source_index;
		double lonmin, lonmax, latmin, latmax;

		Datetime ens_datetime = new Datetime(time);
		this.datetimeEns = ens_datetime;

		vd.createSingleLayerforGridSources(80, ens_datetime);
		// if (this.sectorizeAll) this.sectorizeAll(25, ens_datetime);

		lonmin = vd.boundaries[0];
		lonmax = vd.boundaries[1];
		latmin = vd.boundaries[2];
		latmax = vd.boundaries[3];

		double temp2 = ((lonmax - lonmin) / (latmax - latmin) / 2);

		this.height = resolution;
		this.width = (int) (temp2 * height);
		this.cellArea = this.calcCellArea(latmax, latmin, lonmax, lonmin);

		this.ensemble = new Double[height][width];

		this.ensembleSTD = new Double[height][width];
		this.zones = new char[height][width];
		this.source_w = new Double[height][width];
		this.density = new int[height][width];

		System.out
				.println("Geographical area for ensemble (latmin,lonmin)(latmax,lonmax): ("
						+ latmin
						+ ","
						+ lonmin
						+ "), ("
						+ latmax
						+ ","
						+ lonmax + ")");

		double lon_increment = (lonmax - lonmin) / width;
		double lat_increment = (latmax - latmin) / height;

		// ////OUTER height-loop [j][]
		for (int j = 0; j < height; j++) {
			double clat = latmax - (j - 1) * lat_increment;
			// ////////inner width loop [][i]
			for (int i = 0; i < width; i++) {
				double clon = lonmin + (i - 1) * lon_increment;
				char bg_zone = 'X'; // this is for background image layer

				if (this.mapPack.landAvailable) {

					bg_zone = mapPack.getZone(clat, clon);

				}

			
			
					combVal = this.calculateCombinationValue(vd, clat, clon,
							null, ens_datetime, false, Options.method);
			

				this.ensemble[j][i] = this.editPrecision(combVal.get(0), 2);// the
																			// value;
				this.zones[j][i] = bg_zone;
				this.ensembleSTD[j][i] = this.editPrecision(combVal.get(1), 2);
				this.source_w[j][i] = this.editPrecision(
						combVal.get(selectedSource + 2), 3);
				// to file -output. For this we must make d2 csv-compatible line
				// to be flushed

				// if (this.density[j][i] == 1100*16)
				// System.out.println("Density max-category warning!!");
				if (this.useDensity) {
					this.density[j][i] = this.mapPack.getDensity(clat, clon);
					this.totExposure = this.totExposure + this.density[j][i]
							* this.cellArea * this.ensemble[j][i];
					this.totPersons = (int) (this.totPersons + this.density[j][i]
							* this.cellArea);
				}
			}

		}

		this.generate_output(vd); // after the ensembling we generate output
									// based on the results

	}

	public String calcAreaMaxMin(String time, String type) {
		final int resolution = 100;
		String output = "";

		Datetime ens_datetime = new Datetime(time);
		double temp2;

		double lonmin, lonmax, latmin, latmax;
		VariableData vd = this.datapack.getVarData(type);
		// if (this.sectorizeAll) this.sectorizeAll(10, ens_datetime);

		lonmin = vd.boundaries[0];
		lonmax = vd.boundaries[1];
		latmin = vd.boundaries[2];
		latmax = vd.boundaries[3];

		temp2 = ((lonmax - lonmin) / (latmax - latmin) / 2); // lonmax -lonmin
																// per latmax -
																// latmin. This
																// is divided by
																// 2 to correct
																// the usual
																// difference
																// between d2
																// lat and lon
																// degree in
																// finland
		this.height = resolution;
		this.width = (int) (temp2 * this.height);
		double lon_increment = (lonmax - lonmin) / this.width;
		double lat_increment = (latmax - latmin) / this.height;

		double min = Integer.MAX_VALUE;
		double max = Integer.MIN_VALUE;
		double minSTD = 0.0, maxSTD = 0.0;
		Env maxEnv = null;
		Env minEnv = null;
		double lonMax = 0, lonMin = 0, latMax = 0, latMin = 0;

		double[] contributionMax = new double[vd.sources.size()], contributionMin = new double[vd.sources
				.size()];

		// ////OUTER height-loop [j][]
		for (int j = 1; j <= this.height; j++) {
			// ////////inner width loop [][i]
			for (int i = 1; i <= width; i++) {
				double clat = latmax - (j - 1) * lat_increment;
				double clon = lonmin + (i - 1) * lon_increment;
Env env = this.mapPack.getEnvironment(clat, clon, 250);

				ArrayList<Double> value_temp = this.calculateCombinationValue(
						vd, clat, clon, env, ens_datetime, false, Options.method);
				if (value_temp.get(0) > max) {
					max = value_temp.get(0);
					maxSTD = value_temp.get(1);
					maxEnv = env;
					latMax = this.editPrecision(clat, 5);
					lonMax = this.editPrecision(clon, 5);
					for (int k = 2; k < value_temp.size(); k++) {
						contributionMax[k - 2] = value_temp.get(k);
					}
				}
				if (value_temp.get(0) < min) {
					min = value_temp.get(0);
					minSTD = value_temp.get(0);
					minEnv = env;
					latMin = this.editPrecision(clat, 5);
					lonMin = this.editPrecision(clon, 5);
					for (int k = 2; k < value_temp.size(); k++) {
						contributionMin[k - 2] = value_temp.get(k);
					}
				}

			}

		}
		String n = System.getProperty("line.separator");
		output = type + "-extremalities in the selected area in " + time + n;
		output = output + "MINIMUM value: " + this.editPrecision(min, 3)
				+ " (STD = " + this.editPrecision(minSTD, 3) + ")" + n;
		output = output + minEnv.exProfile + " at " + latMin + ", " + lonMin + n + n;
		output = output + "Contribution:" + n;
		for (int i = 0; i < vd.sources.size(); i++) {
			output = output + vd.sources.get(i).ID + "  " + "\t"
					+ this.editPrecision(contributionMin[i] * 100, 1) + "%" + n;
		}
		output = output + "==============================" + n + n;
		output = output + "MAXIMUM value: " + this.editPrecision(max, 3)
				+ " (STD = " + this.editPrecision(maxSTD, 3) + ")" + n;
		output = output + maxEnv.exProfile + " at " + latMax + ", " + lonMax + n + n;
		output = output + "Contribution:" + n;
		for (int i = 0; i < vd.sources.size(); i++) {
			output = output + vd.sources.get(i).ID + "  " + "\t"
					+ this.editPrecision(contributionMax[i] * 100, 1) + "%" + n;
		}
		return output;

	}

	public ArrayList<Double> calculateCombinationWD(VariableData vd, double lat, double lon, Datetime ens_datetime, boolean oldSchool) {
        
		ArrayList<Double> returnable = new ArrayList<Double>();
		ArrayList<Double> offsets = new ArrayList<Double>();
		double[] scores;
		double[] normalized_scores;
		double[] values_cos,values_sin;

		//scores
		double score;
		double scoresum=0;
		double ensembled_value_cos =0;
		double ensembled_value_sin =0;

		double fusedVar =0;

		int size = vd.sources.size();


		    scores = new double[size];
		    normalized_scores = new double[size];
		    values_cos = new double[size];                                 
		    values_sin = new double[size]; 
		                                                                  

		    for (int i=0;i < scores.length;i++ ) {
		             //zonetype analysis -both obs profile and the profile of current grid coordinates are needed
		        
		    Observation best;
		    if (oldSchool) {
		    best= vd.sources.get(i).getBestObsBruteForce(ens_datetime,lat,lon);
		    } else {
		    best= vd.sources.get(i).get_best_obs(ens_datetime,lat,lon,true);    
		    }
		    
		    offsets.add((double)best.t_offset); 
		    
		       if (vd.sources.get(i).banned) {       
		                scores[i]=0.00000001;
		                normalized_scores[i]=0.0;
		                values_cos[i]=0;                                 
		                values_sin[i]=0; 
		             continue;
		         
		       }                        

		             
		                score = this.DB.calculateVariance(vd.type, best, lat, lon, ens_datetime);

		                values_cos[i] = Math.cos(2*Math.PI*best.value/360);     
		                values_sin[i] = Math.sin(2*Math.PI*best.value/360); 

		                scores[i] = 1/score; // = 1/VAR
		 
		                scoresum = scoresum +scores[i]; //=SUM(1/VAR_i) 
		                //System.out.println("i: "+i+" ID: "+best.ID+", distance: "+distance+"timesptamp: "+best.timestamp+" unscaled weight: "+scores[i] +"scoresum: " +scoresum);
		                
		        }//END OF FOR SOURCES

		        //score normalization (2):
		        for (int i=0;i < scores.length;i++ ) {
		                normalized_scores[i] = scores[i]/scoresum;
		        }

		        
		        //value calculation phase (3):
		        for (int i=0;i < scores.length;i++ ) {
		                ensembled_value_sin =  ensembled_value_sin + normalized_scores[i]*values_sin[i];
		                ensembled_value_cos = ensembled_value_cos + normalized_scores[i]*values_cos[i]; 
		                
		                fusedVar = fusedVar + 1/scores[i]*normalized_scores[i]*normalized_scores[i]; // this sums every d2^2*VAR(obs)
		                  
		            }

		                double angle;
		                double cos = Math.acos(ensembled_value_cos)*(360/2/Math.PI);
		                double sin = Math.asin(ensembled_value_sin)*(360/2/Math.PI);
		                if (sin < 0) { //names are a bit twisted here I suppose, but this still seem to work
		                    angle = 360-cos;
		                } else {
		                    angle = cos; 
		                }
		                //System.out.println("angle: "+angle);  
		                

						returnable.add(angle);
		                                
		                                //2nd element = estimated STD
						returnable.add(Math.sqrt(fusedVar)); //back to total STD
		                                
		                                // rest elements - source weights
		                                for (int k=0;k<normalized_scores.length;k++)
		                                returnable.add(normalized_scores[k]);
		                                
		                                double offSum =0;
		                                for (int k=0;k<offsets.size();k++) {
		                                    offSum = offSum + offsets.get(k);
		                                }
		                                offSum = offSum/offsets.size();
		                                returnable.add(offSum);
						return returnable;
		} 
	

	// ensemble calculation method for d2 single point in time and space.
	// Uses d2 fast method for time separation calculation
	// option: this method uses also the second best datapoint for each source
	// if boolean use_two is true.
	// This feature is handy e.g. if datapoints 13.00 and 14.00 exists while
	// 13.30 is the ensembling time
	public ArrayList<Double> calculateCombinationValue(VariableData vd, double lat, double lon,Env fixedEnv, Datetime ens_datetime, boolean oldSchool, int method) {
		 
		if (vd.type.equals("windDirection")) return this.calculateCombinationWD(vd, lat, lon, ens_datetime, oldSchool);   
		    
		ArrayList<Double> returnable = new ArrayList<Double>();
		ArrayList<Double> offsets = new ArrayList<Double>();

		ArrayList<double[]> weightVals = new ArrayList<double[]>();
		ArrayList<Double> vars = new ArrayList<Double>();


		//Environment evaluation - if the parameter is null, a new Env is evaluated
		Env env;
		if (fixedEnv != null) {
		env = fixedEnv;    
		} else {
		env= this.mapPack.getEnvironment(lat, lon, Options.profRadius);
		}

		                             
		    //score calculation phase (1):
		 
		    double scoresum=0;
		    for (int i=0;i < vd.sources.size();i++ ) {
		             //zonetype analysis -both obs profile and the profile of current grid coordinates are needed
		    if (vd.sources.get(i).banned) continue;
		        
		        
		    Observation best;
		    if (oldSchool) {
		    best= vd.sources.get(i).getBestObsBruteForce(ens_datetime,lat,lon);
		    } else {
		    best= vd.sources.get(i).get_best_obs(ens_datetime,lat,lon,true);    
		    }
		    
		    offsets.add((double)best.t_offset);             
		                                     

		               // System.out.println("baseVar: "+ base_var);
		                //System.out.println("Var calc: d,t,bv: " +distance + ", " + temporal_diff + ", "+ base_var);
		                double dist = best.calc_spatial_distance(lat, lon)/1000.0;
		               if (method == Options.METHOD_IWD) {
		                   double var = dist + 0.001;
		                   double score = 1/var;
		                   vars.add(score);
		                   double[] temp = {score,best.value};
		                   weightVals.add(temp);
		                   scoresum = scoresum +score;
		                   
		               
		               }else if (method == Options.METHOD_AVERAGES) {
		               
		                  //same var for every value will lead to average
		                   double score = 1;
		                   vars.add(score);
		                   double[] temp = {score,best.value};
		                   weightVals.add(temp);
		                   scoresum = scoresum +score;
		               
		                  }else if (method == Options.METHOD_CLOSEST) {
		                                         
		                   double score = dist;
		                   vars.add(score);
		                   double[] temp = {score,best.value};
		                   weightVals.add(temp);
		                   scoresum = scoresum +score;
		                                
		                   
		               } else {//fusion
		                double var = this.DB.calculateVariance(vd.type, best, lat, lon, ens_datetime);   
		                double score = 1/var;
		                //System.out.println("sennding " + best.value);
		                double val = this.DB.eqProfiles.getProfileCorrectedValue(best.value, vd.type, best.env,env,ens_datetime, best.datetime,dist, Options.profCorrStr);
		                
		                double[] temp = {score,val};
		                //System.out.println("recieved:" + val+ "\t" + score);
		                weightVals.add(temp);
		                vars.add(var);
		                scoresum = scoresum +score;
		                 
		               }
		                
		               
		               //if method.closest then scores and values are needed to be reprocessed
		               if (method == Options.METHOD_CLOSEST) {
		                      double best_score = Double.MAX_VALUE;
		                      int index =0;
		                for (int k =0;k<weightVals.size();k++) {
		                    if (weightVals.get(k)[0] < best_score) {
		                        best_score = weightVals.get(k)[0];
		                        index = k;
		                    }
		                }
		                
		                for (int k =0;k<weightVals.size();k++) {
		                    if (k== index) {
		                       weightVals.get(k)[0] = 1; 
		                        
		                    } else {
		                       weightVals.get(k)[0] = 0; 
		                    }
		                }
		               scoresum = 1;
		               }// if Method.closest             
		                
		        }//for sources

		    
		    
		        //score normalization (2):
		        for (int i=0;i < weightVals.size();i++ ) {          
		                weightVals.get(i)[0] = weightVals.get(i)[0]/scoresum;            
		        }

		        
		        double ensembled_value=0;
		        double fusedVar =0;
		        //value calculation phase (3):
		        for (int i=0;i < weightVals.size();i++ ) {
		            
		                ensembled_value = ensembled_value + weightVals.get(i)[0]*weightVals.get(i)[1];	// this sums every d2*OBS
		                fusedVar = fusedVar + vars.get(i)*weightVals.get(i)[0]*weightVals.get(i)[0]; // this sums every d2^2*VAR(obs)

		      
		            }

						returnable.add(ensembled_value);
		                                
		                                //2nd element = estimated STD
						returnable.add(Math.sqrt(fusedVar)); //back to total STD
		                                
		                                // rest elements - source weights
		                                for (int k=0;k<weightVals.size();k++)
		                                returnable.add(weightVals.get(k)[0]);
		                                
		                                double offSum =0;
		                                for (int k=0;k<offsets.size();k++) {
		                                    offSum = offSum + offsets.get(k);
		                                }
		                                offSum = offSum/offsets.size();
		                                returnable.add(offSum);
						return returnable;

		}

	public void generate_output(VariableData vd) {
		String temp = "";
		double max_ens_val = Double.MIN_VALUE;
		double min_ens_val = Double.MAX_VALUE;

		double max_ens_w = 0;
		double min_ens_w = 1;
		double max_ens_STD = Double.MIN_VALUE;
		double min_ens_STD = Double.MAX_VALUE;

		// double maxlat,maxlon,minlat,minlon; //TODO get coordinates for
		// maxima/minima
		// Check the ensemble
		for (int j = 0; j < this.height; j++) { // Note that we do not want to
												// check lat and lons here which
												// are included in the ensemble
												// array! thus we start at 1
			for (int i = 0; i < this.width; i++) {

				if (this.ensemble[j][i] <= min_ens_val) {
					min_ens_val = this.ensemble[j][i];
					// this.boundaries[3] - j*lat_increment,
					// lonmin+i*lon_increment

				}

				if (this.ensemble[j][i] >= max_ens_val) {
					max_ens_val = this.ensemble[j][i];
				}

				// same for var

				if (this.ensembleSTD[j][i] <= min_ens_STD) {
					min_ens_STD = this.ensembleSTD[j][i];
					// this.boundaries[3] - j*lat_increment,
					// lonmin+i*lon_increment

				}

				if (this.ensembleSTD[j][i] >= max_ens_STD) {
					max_ens_STD = this.ensembleSTD[j][i];
				}

				if (this.source_w[j][i] <= min_ens_w) {
					min_ens_w = this.source_w[j][i];
					// this.boundaries[3] - j*lat_increment,
					// lonmin+i*lon_increment

				}

				if (this.source_w[j][i] >= max_ens_w) {
					max_ens_w = this.source_w[j][i];
				}

			}
		}

		this.ens_maximum = max_ens_val;
		this.ens_minimum = min_ens_val;
		this.ens_STD_maximum = max_ens_STD;
		this.ens_STD_minimum = min_ens_STD;
		this.ens_w_maximum = max_ens_w;
		this.ens_w_minimum = min_ens_w;
		String n = System.getProperty("line.separator");

		temp = temp + "Pollutant type: " + vd.type + n;
		temp = temp + "Area ensemble in: " + this.datetimeEns.getStringDate()
				+ n;
		temp = temp + "Resolution : " + this.height + " X " + this.width
				+ " cells. " + n;
		temp = temp + "Amount of input data nodes : " + vd.sources.size() + n
				+ "============================================" + n;

		temp = temp + "Ensemble maximum value :" + n + max_ens_val + vd.unit
				+ n;
		temp = temp + "Ensemble minimum value : " + n + min_ens_val + vd.unit
				+ n;

		temp = temp + "Expected lowest standard deviation: " + n + "(+-)"
				+ min_ens_STD + vd.unit + n;
		temp = temp + "Expected highest standard deviation: " + n + "(+-)"
				+ max_ens_STD + vd.unit + n;

		temp = temp + "(One STD reflects roughly a " + n
				+ "70% confidence range.)" + n + n;

		if (this.useDensity) {
			if (vd.unit.equals("g/m3")) {
				// It would be a bit silly to present exposures to cloudiness
				// etc
				temp = temp
						+ "Average exposure for residents: "
						+ n
						+ this.editPrecision(
								this.totExposure / this.totPersons, 3)
						+ vd.unit + n;
			}
			temp = temp + "Number of residents in the area (2010): " + n
					+ this.totPersons + n + n;
		}

		temp = temp + "Selected source ("
				+ (vd.sources.get(this.source_index).ID) + ")" + n
				+ "minmum weight : " + min_ens_w + n;
		temp = temp + "Selected source ("
				+ (vd.sources.get(this.source_index).ID) + ") " + n
				+ "maxmum weight : " + max_ens_w + n + n;

		if (this.trust_measurements) {
			temp = temp + "Base variance for measurements" + n
					+ " was set to zero." + n;
		}

		this.output = temp;
	}

	public BufferedImage draw_ensemble(String type, String STDorValue,
			String color_set, boolean automatic_sc, boolean transparency) {
		Double[][] data;
		double max;
		double min;
		double range;
		Color cc40, cc39, cc38, cc37, cc36, cc35, cc34, cc33, cc32, cc31, cc30, cc29, cc28, cc27, cc26, cc25, cc24, cc23, cc22, cc21, cc20, cc19, cc18, cc17, cc16, cc15, cc14, cc13, cc12, cc11, cc10, cc9, cc8, cc7, cc6, cc5, cc4, cc3, cc2, cc1;
		Boolean automatic_scaling = automatic_sc;
		String colorset = color_set;

		int tr;
		if (transparency) {
			tr = 0; // in some cases this will trigger almost totally
					// transparent colors using Math.max(tr,a)
		} else {
			tr = 255; // no transparency
		}

		if (STDorValue.equals("STD")) {
			data = this.ensembleSTD;

			max = this.ens_STD_maximum;
			min = this.ens_STD_minimum;

		} else if (STDorValue.equals("weight")) {
			data = this.source_w;
			// max = this.ens_w_maximum;
			// min= this.ens_w_minimum;
			max = 1.0;
			min = 0.0;

		} else { // AQ pollutants===================================
			data = this.ensemble;
			if (type.equals("PM2.5") || type.equals("PM2.5_AirPollutantValue")) {
				max = 51;
				min = 10;
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}

			} else if (type.equals("SO2")
					|| type.equals("SO2_AirPollutantValue")) {
				max = 351;
				min = 20;
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}
			} else if (type.equals("PM10")
					|| type.equals("PM10_AirPollutantValue")) {
				max = 201;
				min = 20;

				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}
			} else if (type.equals("CO") || type.equals("CO_AirPollutantValue")) {
				max = 31000;
				min = 400;
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}

			} else if (type.equals("NO2")
					|| type.equals("NO2_AirPollutantValue")) {
				max = 201;
				min = 40;
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}
			} else if (type.equals("O3") || type.equals("O3_AirPollutantValue")) {
				max = 181;
				min = 60;
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}
			} else if (type.equals("temperature")) {// Meteorological===============================
				max = 35; // degrees C
				min = -40;
				colorset = "Temperature";
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}
			} else if (type.equals("windSpeed")) {// Meteorological===============================
				max = 32;// average of 1h (m/s)
				min = 0;
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}
			} else if (type.equals("rain")) {// Meteorological===============================
				max = 3; // mm/1h
				min = 0.3;
				colorset = "BlueToWhite";
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}
			} else if (type.equals("skyCondition")) {// Meteorological===============================
				max = 9; // mm/1h
				min = 0;
				colorset = "BlueToWhite";
				if (automatic_scaling) {
					max = this.ens_maximum;
					min = this.ens_minimum;
				}

			} else {
				System.out
						.println("Color scaling: unknown type => default max to lowest scaling");
				max = this.ens_maximum;
				min = this.ens_minimum;
			}

		}

		range = max - min;
		Color zero = new Color(254, 254, 254, 255);// transparent white
		int height = this.height;
		int width = this.width;

		if (colorset.equals("greenToViolet")
				|| colorset.equals("TruePollutant")) {
			// 40 steps green to violet
			this.cols[39] = cc40 = new Color(156, 1, 241, Math.max(tr, 230));
			this.cols[38] = cc39 = new Color(177, 2, 241, Math.max(tr, 230));
			this.cols[37] = cc38 = new Color(197, 2, 241, Math.max(tr, 230));
			this.cols[36] = cc37 = new Color(217, 2, 241, Math.max(tr, 230));
			this.cols[35] = cc36 = new Color(237, 2, 241, Math.max(tr, 230));
			this.cols[34] = cc35 = new Color(241, 2, 224, Math.max(tr, 230));
			this.cols[33] = cc34 = new Color(240, 2, 204, Math.max(tr, 230));
			this.cols[32] = cc33 = new Color(240, 2, 183, Math.max(tr, 230));
			this.cols[31] = cc32 = new Color(240, 2, 163, Math.max(tr, 230));
			this.cols[30] = cc31 = new Color(240, 2, 143, Math.max(tr, 230));
			this.cols[29] = cc30 = new Color(240, 2, 123, Math.max(tr, 230));
			this.cols[28] = cc29 = new Color(240, 2, 102, Math.max(tr, 230));
			this.cols[27] = cc28 = new Color(239, 2, 82, Math.max(tr, 230));
			this.cols[26] = cc27 = new Color(239, 3, 62, Math.max(tr, 230));
			this.cols[25] = cc26 = new Color(239, 3, 42, Math.max(tr, 230));
			this.cols[24] = cc25 = new Color(239, 3, 22, Math.max(tr, 230));
			this.cols[23] = cc24 = new Color(239, 3, 3, Math.max(tr, 230));
			this.cols[22] = cc23 = new Color(238, 23, 3, Math.max(tr, 230));
			this.cols[21] = cc22 = new Color(238, 43, 3, Math.max(tr, 230));
			this.cols[20] = cc21 = new Color(238, 63, 3, Math.max(tr, 230));

			this.cols[19] = cc20 = new Color(238, 83, 3, Math.max(tr, 230));
			this.cols[18] = cc19 = new Color(238, 103, 3, Math.max(tr, 230));
			this.cols[17] = cc18 = new Color(238, 123, 3, Math.max(tr, 230));
			this.cols[16] = cc17 = new Color(237, 143, 3, Math.max(tr, 230));
			this.cols[15] = cc16 = new Color(237, 162, 3, Math.max(tr, 230));
			this.cols[14] = cc15 = new Color(237, 182, 3, Math.max(tr, 230));
			this.cols[13] = cc14 = new Color(237, 202, 4, Math.max(tr, 230));
			this.cols[12] = cc13 = new Color(237, 221, 4, Math.max(tr, 230));
			this.cols[11] = cc12 = new Color(232, 236, 4, Math.max(tr, 230));
			this.cols[10] = cc11 = new Color(212, 236, 4, Math.max(tr, 230));
			this.cols[9] = cc10 = new Color(192, 236, 4, Math.max(tr, 230));
			this.cols[8] = cc9 = new Color(172, 236, 4, Math.max(tr, 230));
			this.cols[7] = cc8 = new Color(153, 236, 4, Math.max(tr, 230));
			this.cols[6] = cc7 = new Color(133, 236, 4, Math.max(tr, 230));
			this.cols[5] = cc6 = new Color(113, 235, 4, Math.max(tr, 230));
			this.cols[4] = cc5 = new Color(94, 235, 4, Math.max(tr, 230));
			this.cols[3] = cc4 = new Color(74, 235, 4, Math.max(tr, 230));
			this.cols[2] = cc3 = new Color(54, 235, 4, Math.max(tr, 230));
			this.cols[1] = cc2 = new Color(35, 235, 4, Math.max(tr, 230));
			this.cols[0] = cc1 = new Color(15, 235, 5, Math.max(tr, 230));

		} else if (colorset.equals("Temperature")
				|| colorset.equals("ColdToWarm")) {
			// 40 steps green to violet
			this.cols[39] = cc40 = new Color(191, 0, 1, Math.max(tr, 200));
			this.cols[38] = cc39 = new Color(191, 17, 1, Math.max(tr, 200));
			this.cols[37] = cc38 = new Color(192, 37, 1, Math.max(tr, 200));
			this.cols[36] = cc37 = new Color(193, 57, 0, Math.max(tr, 200));
			this.cols[35] = cc36 = new Color(194, 77, 0, Math.max(tr, 200));
			this.cols[34] = cc35 = new Color(195, 97, 0, Math.max(tr, 200));
			this.cols[33] = cc34 = new Color(196, 117, 0, Math.max(tr, 200));
			this.cols[32] = cc33 = new Color(197, 137, 0, Math.max(tr, 200));
			this.cols[31] = cc32 = new Color(198, 157, 0, Math.max(tr, 200));
			this.cols[30] = cc31 = new Color(199, 177, 0, Math.max(tr, 200));
			this.cols[29] = cc30 = new Color(200, 198, 0, Math.max(tr, 200));
			this.cols[28] = cc29 = new Color(182, 201, 0, Math.max(tr, 200));
			this.cols[27] = cc28 = new Color(162, 202, 0, Math.max(tr, 200));
			this.cols[26] = cc27 = new Color(143, 203, 0, Math.max(tr, 200));
			this.cols[25] = cc26 = new Color(123, 203, 0, Math.max(tr, 200));
			this.cols[24] = cc25 = new Color(103, 204, 0, Math.max(tr, 200));
			this.cols[23] = cc24 = new Color(83, 205, 0, Math.max(tr, 200));
			this.cols[22] = cc23 = new Color(63, 206, 0, Math.max(tr, 200));
			this.cols[21] = cc22 = new Color(43, 207, 0, Math.max(tr, 200));
			this.cols[20] = cc21 = new Color(1, 209, 0, Math.max(tr, 200));

			this.cols[19] = cc20 = new Color(1, 209, 0, Math.max(tr, 200));
			this.cols[18] = cc19 = new Color(1, 210, 21, Math.max(tr, 200));
			this.cols[17] = cc18 = new Color(1, 210, 37, Math.max(tr, 200));
			this.cols[16] = cc17 = new Color(1, 211, 57, Math.max(tr, 200));
			this.cols[15] = cc16 = new Color(1, 212, 87, Math.max(tr, 200));
			this.cols[14] = cc15 = new Color(1, 213, 107, Math.max(tr, 200));
			this.cols[13] = cc14 = new Color(1, 214, 129, Math.max(tr, 200));
			this.cols[12] = cc13 = new Color(1, 215, 151, Math.max(tr, 200));
			this.cols[11] = cc12 = new Color(1, 216, 173, Math.max(tr, 200));
			this.cols[10] = cc11 = new Color(1, 217, 195, Math.max(tr, 200));
			this.cols[9] = cc10 = new Color(1, 218, 218, Math.max(tr, 200));
			this.cols[8] = cc9 = new Color(1, 197, 219, Math.max(tr, 200));
			this.cols[7] = cc8 = new Color(1, 176, 220, Math.max(tr, 200));
			this.cols[6] = cc7 = new Color(1, 155, 221, Math.max(tr, 200));
			this.cols[5] = cc6 = new Color(1, 133, 222, Math.max(tr, 200));
			this.cols[4] = cc5 = new Color(1, 112, 223, Math.max(tr, 200));
			this.cols[3] = cc4 = new Color(1, 90, 224, Math.max(tr, 200));
			this.cols[2] = cc3 = new Color(1, 68, 225, Math.max(tr, 200));
			this.cols[1] = cc2 = new Color(1, 46, 226, Math.max(tr, 200));
			this.cols[0] = cc1 = new Color(1, 24, 227, Math.max(tr, 200));

		} else if (colorset.equals("BlueToWhite")
				|| colorset.equals("WhiteToBlue")) {
			// 40 steps green to violet
			this.cols[39] = cc40 = new Color(49, 2, 130, Math.max(tr, 230));//
			this.cols[38] = cc39 = new Color(50, 6, 132, Math.max(tr, 230));
			this.cols[37] = cc38 = new Color(51, 11, 135, Math.max(tr, 230));
			this.cols[36] = cc37 = new Color(52, 15, 138, Math.max(tr, 230));
			this.cols[35] = cc36 = new Color(53, 18, 141, Math.max(tr, 230));
			this.cols[34] = cc35 = new Color(54, 22, 144, Math.max(tr, 230));
			this.cols[33] = cc34 = new Color(56, 25, 147, Math.max(tr, 230));
			this.cols[32] = cc33 = new Color(58, 28, 150, Math.max(tr, 230));
			this.cols[31] = cc32 = new Color(59, 31, 153, Math.max(tr, 230));
			this.cols[30] = cc31 = new Color(61, 35, 156, Math.max(tr, 230));
			this.cols[29] = cc30 = new Color(63, 38, 159, Math.max(tr, 230));//
			this.cols[28] = cc29 = new Color(66, 43, 161, Math.max(tr, 230));
			this.cols[27] = cc28 = new Color(69, 48, 164, Math.max(tr, 230));
			this.cols[26] = cc27 = new Color(72, 55, 167, Math.max(tr, 230));
			this.cols[25] = cc26 = new Color(75, 61, 170, Math.max(tr, 230));
			this.cols[24] = cc25 = new Color(79, 67, 173, Math.max(tr, 230));
			this.cols[23] = cc24 = new Color(83, 72, 176, Math.max(tr, 230));
			this.cols[22] = cc23 = new Color(87, 78, 179, Math.max(tr, 230));
			this.cols[21] = cc22 = new Color(90, 83, 182, Math.max(tr, 230));
			this.cols[20] = cc21 = new Color(93, 88, 185, Math.max(tr, 230));

			this.cols[19] = cc20 = new Color(96, 93, 188, Math.max(tr, 230));//
			this.cols[18] = cc19 = new Color(100, 100, 191, Math.max(tr, 230));
			this.cols[17] = cc18 = new Color(104, 107, 194, Math.max(tr, 230));
			this.cols[16] = cc17 = new Color(108, 113, 197, Math.max(tr, 230));
			this.cols[15] = cc16 = new Color(116, 123, 201, Math.max(tr, 230));
			this.cols[14] = cc15 = new Color(124, 133, 205, Math.max(tr, 230));
			this.cols[13] = cc14 = new Color(132, 143, 208, Math.max(tr, 230));
			this.cols[12] = cc13 = new Color(140, 151, 212, Math.max(tr, 230));
			this.cols[11] = cc12 = new Color(148, 158, 215, Math.max(tr, 230));
			this.cols[10] = cc11 = new Color(164, 173, 222, Math.max(tr, 230));
			this.cols[9] = cc10 = new Color(172, 181, 225, Math.max(tr, 230));//
			this.cols[8] = cc9 = new Color(180, 189, 228, Math.max(tr, 230));
			this.cols[7] = cc8 = new Color(188, 197, 231, Math.max(tr, 230));
			this.cols[6] = cc7 = new Color(197, 205, 235, Math.max(tr, 230));
			this.cols[5] = cc6 = new Color(205, 213, 238, Math.max(tr, 230));
			this.cols[4] = cc5 = new Color(214, 221, 241, Math.max(tr, 180));
			this.cols[3] = cc4 = new Color(223, 229, 244, Math.max(tr, 140));
			this.cols[2] = cc3 = new Color(232, 236, 247, Math.max(tr, 120));
			this.cols[1] = cc2 = new Color(241, 244, 250, Math.max(tr, 50));
			this.cols[0] = cc1 = new Color(255, 255, 255, Math.max(tr, 0));//

		} else {// light yellow to dark red

			this.cols[39] = cc40 = new Color(96, 0, 3, Math.max(tr, 240));
			this.cols[38] = cc39 = new Color(100, 1, 2, Math.max(tr, 240));
			this.cols[37] = cc38 = new Color(104, 5, 3, Math.max(tr, 240));
			this.cols[36] = cc37 = new Color(108, 10, 5, Math.max(tr, 240));
			this.cols[35] = cc36 = new Color(112, 15, 8, Math.max(tr, 240));
			this.cols[34] = cc35 = new Color(116, 20, 11, Math.max(tr, 240));
			this.cols[33] = cc34 = new Color(120, 25, 14, Math.max(tr, 240));
			this.cols[32] = cc33 = new Color(124, 30, 17, Math.max(tr, 240));
			this.cols[31] = cc32 = new Color(128, 35, 20, Math.max(tr, 240));
			this.cols[30] = cc31 = new Color(132, 43, 21, Math.max(tr, 240));
			this.cols[29] = cc30 = new Color(136, 49, 24, Math.max(tr, 240));
			this.cols[28] = cc29 = new Color(140, 55, 27, Math.max(tr, 240));
			this.cols[27] = cc28 = new Color(144, 61, 31, Math.max(tr, 240));
			this.cols[26] = cc27 = new Color(148, 68, 34, Math.max(tr, 240));
			this.cols[25] = cc26 = new Color(152, 75, 38, Math.max(tr, 240));
			this.cols[24] = cc25 = new Color(156, 81, 42, Math.max(tr, 240));
			this.cols[23] = cc24 = new Color(160, 87, 47, Math.max(tr, 240));
			this.cols[22] = cc23 = new Color(164, 94, 51, Math.max(tr, 240));
			this.cols[21] = cc22 = new Color(168, 101, 55, Math.max(tr, 240));
			this.cols[20] = cc21 = new Color(172, 109, 59, Math.max(tr, 240));

			this.cols[19] = cc20 = new Color(177, 117, 63, Math.max(tr, 240));
			this.cols[18] = cc19 = new Color(181, 124, 68, Math.max(tr, 240));
			this.cols[17] = cc18 = new Color(185, 131, 73, Math.max(tr, 240));
			this.cols[16] = cc17 = new Color(189, 138, 78, Math.max(tr, 240));
			this.cols[15] = cc16 = new Color(193, 146, 83, Math.max(tr, 240));
			this.cols[14] = cc15 = new Color(197, 153, 89, Math.max(tr, 240));
			this.cols[13] = cc14 = new Color(201, 160, 94, Math.max(tr, 240));
			this.cols[12] = cc13 = new Color(205, 168, 100, Math.max(tr, 240));
			this.cols[11] = cc12 = new Color(209, 175, 105, Math.max(tr, 240));
			this.cols[10] = cc11 = new Color(213, 182, 111, Math.max(tr, 240));
			this.cols[9] = cc10 = new Color(217, 189, 117, Math.max(tr, 240));
			this.cols[8] = cc9 = new Color(221, 196, 124, Math.max(tr, 240));
			this.cols[7] = cc8 = new Color(225, 203, 130, Math.max(tr, 240));
			this.cols[6] = cc7 = new Color(229, 210, 136, Math.max(tr, 240));
			this.cols[5] = cc6 = new Color(233, 217, 143, Math.max(tr, 240));
			this.cols[4] = cc5 = new Color(237, 224, 150, Math.max(tr, 240));
			this.cols[3] = cc4 = new Color(241, 230, 157, Math.max(tr, 240));
			this.cols[2] = cc3 = new Color(245, 237, 164, Math.max(tr, 240));
			this.cols[1] = cc2 = new Color(249, 243, 171, Math.max(tr, 240));
			this.cols[0] = cc1 = new Color(254, 250, 179, Math.max(tr, 240));

		}
		// VALUES of the scale
		double v40 = max - 0.025 * range;
		int temp = (int) (100 * v40);
		double temp2 = (double) (temp) / 100;
		this.vals[39] = Double.toString(temp2);
		double v39 = max - 0.05 * range;
		temp = (int) (100 * v39);
		temp2 = (double) (temp) / 100;
		this.vals[38] = Double.toString(temp2);
		double v38 = max - 0.075 * range;
		temp = (int) (100 * v38);
		temp2 = (double) (temp) / 100;
		this.vals[37] = Double.toString(temp2);
		double v37 = max - 0.1 * range;
		temp = (int) (100 * v37);
		temp2 = (double) (temp) / 100;
		this.vals[36] = Double.toString(temp2);
		double v36 = max - 0.125 * range;
		temp = (int) (100 * v36);
		temp2 = (double) (temp) / 100;
		this.vals[35] = Double.toString(temp2);
		double v35 = max - 0.15 * range;
		temp = (int) (100 * v35);
		temp2 = (double) (temp) / 100;
		this.vals[34] = Double.toString(temp2);
		double v34 = max - 0.175 * range;
		temp = (int) (100 * v34);
		temp2 = (double) (temp) / 100;
		this.vals[33] = Double.toString(temp2);
		double v33 = max - 0.2 * range;
		temp = (int) (100 * v33);
		temp2 = (double) (temp) / 100;
		this.vals[32] = Double.toString(temp2);
		double v32 = max - 0.225 * range;
		temp = (int) (100 * v32);
		temp2 = (double) (temp) / 100;
		this.vals[31] = Double.toString(temp2);
		double v31 = max - 0.25 * range;
		temp = (int) (100 * v31);
		temp2 = (double) (temp) / 100;
		this.vals[30] = Double.toString(temp2);
		double v30 = max - 0.275 * range;
		temp = (int) (100 * v30);
		temp2 = (double) (temp) / 100;
		this.vals[29] = Double.toString(temp2);
		double v29 = max - 0.3 * range;
		temp = (int) (100 * v29);
		temp2 = (double) (temp) / 100;
		this.vals[28] = Double.toString(temp2);
		double v28 = max - 0.325 * range;
		temp = (int) (100 * v28);
		temp2 = (double) (temp) / 100;
		this.vals[27] = Double.toString(temp2);
		double v27 = max - 0.35 * range;
		temp = (int) (100 * v27);
		temp2 = (double) (temp) / 100;
		this.vals[26] = Double.toString(temp2);
		double v26 = max - 0.375 * range;
		temp = (int) (100 * v26);
		temp2 = (double) (temp) / 100;
		this.vals[25] = Double.toString(temp2);
		double v25 = max - 0.4 * range;
		temp = (int) (100 * v25);
		temp2 = (double) (temp) / 100;
		this.vals[24] = Double.toString(temp2);
		double v24 = max - 0.425 * range;
		temp = (int) (100 * v24);
		temp2 = (double) (temp) / 100;
		this.vals[23] = Double.toString(temp2);
		double v23 = max - 0.45 * range;
		temp = (int) (100 * v23);
		temp2 = (double) (temp) / 100;
		this.vals[22] = Double.toString(temp2);
		double v22 = max - 0.475 * range;
		temp = (int) (100 * v22);
		temp2 = (double) (temp) / 100;
		this.vals[21] = Double.toString(temp2);
		double v21 = max - 0.5 * range;
		temp = (int) (100 * v21);
		temp2 = (double) (temp) / 100;
		this.vals[20] = Double.toString(temp2);

		double v20 = max - 0.525 * range;
		temp = (int) (100 * v20);
		temp2 = (double) (temp) / 100;
		this.vals[19] = Double.toString(temp2);
		double v19 = max - 0.55 * range;
		temp = (int) (100 * v19);
		temp2 = (double) (temp) / 100;
		this.vals[18] = Double.toString(temp2);
		double v18 = max - 0.575 * range;
		temp = (int) (100 * v18);
		temp2 = (double) (temp) / 100;
		this.vals[17] = Double.toString(temp2);
		double v17 = max - 0.6 * range;
		temp = (int) (100 * v17);
		temp2 = (double) (temp) / 100;
		this.vals[16] = Double.toString(temp2);
		double v16 = max - 0.625 * range;
		temp = (int) (100 * v16);
		temp2 = (double) (temp) / 100;
		this.vals[15] = Double.toString(temp2);
		double v15 = max - 0.65 * range;
		temp = (int) (100 * v15);
		temp2 = (double) (temp) / 100;
		this.vals[14] = Double.toString(temp2);
		double v14 = max - 0.675 * range;
		temp = (int) (100 * v14);
		temp2 = (double) (temp) / 100;
		this.vals[13] = Double.toString(temp2);
		double v13 = max - 0.7 * range;
		temp = (int) (100 * v13);
		temp2 = (double) (temp) / 100;
		this.vals[12] = Double.toString(temp2);
		double v12 = max - 0.725 * range;
		temp = (int) (100 * v12);
		temp2 = (double) (temp) / 100;
		this.vals[11] = Double.toString(temp2);
		double v11 = max - 0.75 * range;
		temp = (int) (100 * v11);
		temp2 = (double) (temp) / 100;
		this.vals[10] = Double.toString(temp2);
		double v10 = max - 0.775 * range;
		temp = (int) (100 * v10);
		temp2 = (double) (temp) / 100;
		this.vals[9] = Double.toString(temp2);
		double v9 = max - 0.8 * range;
		temp = (int) (100 * v9);
		temp2 = (double) (temp) / 100;
		this.vals[8] = Double.toString(temp2);
		double v8 = max - 0.825 * range;
		temp = (int) (100 * v8);
		temp2 = (double) (temp) / 100;
		this.vals[7] = Double.toString(temp2);
		double v7 = max - 0.85 * range;
		temp = (int) (100 * v7);
		temp2 = (double) (temp) / 100;
		this.vals[6] = Double.toString(temp2);
		double v6 = max - 0.875 * range;
		temp = (int) (100 * v6);
		temp2 = (double) (temp) / 100;
		this.vals[5] = Double.toString(temp2);
		double v5 = max - 0.9 * range;
		temp = (int) (100 * v5);
		temp2 = (double) (temp) / 100;
		this.vals[4] = Double.toString(temp2);
		double v4 = max - 0.925 * range;
		temp = (int) (100 * v4);
		temp2 = (double) (temp) / 100;
		this.vals[3] = Double.toString(temp2);
		double v3 = max - 0.95 * range;
		temp = (int) (100 * v3);
		temp2 = (double) (temp) / 100;
		this.vals[2] = Double.toString(temp2);
		double v2 = max - 0.975 * range;
		temp = (int) (100 * v2);
		temp2 = (double) (temp) / 100;
		this.vals[1] = Double.toString(temp2);
		double v1 = max - 1.0 * range;
		temp = (int) (100 * v1);
		temp2 = (double) (temp) / 100;
		this.vals[0] = Double.toString(temp2);

		int[] pixels = new int[width * height];

		for (int j = 0; j < height; j++) {
			for (int i = 0; i < width; i++) {

				double value = Double.valueOf(data[j][i]);

				if (value >= v40) {

					pixels[(j) * width + (i)] = cc40.getRGB();

				} else if (value < v40 && value >= v39) {

					pixels[(j) * width + (i)] = cc39.getRGB();

				} else if (value < v39 && value >= v38) {

					pixels[(j) * width + (i)] = cc38.getRGB();

				} else if (value < v38 && value >= v37) {

					pixels[(j) * width + (i)] = cc37.getRGB();

				} else if (value < v37 && value >= v36) {

					pixels[(j) * width + (i)] = cc36.getRGB();

				} else if (value < v36 && value >= v35) {

					pixels[(j) * width + (i)] = cc35.getRGB();

				} else if (value < v35 && value >= v34) {

					pixels[(j) * width + (i)] = cc34.getRGB();

				} else if (value < v34 && value >= v33) {

					pixels[(j) * width + (i)] = cc33.getRGB();

				} else if (value < v33 && value >= v32) {

					pixels[(j) * width + (i)] = cc32.getRGB();

				} else if (value < v32 && value >= v31) {

					pixels[(j) * width + (i)] = cc31.getRGB();

				} else if (value < v31 && value >= v30) {

					pixels[(j) * width + (i)] = cc30.getRGB();

				} else if (value < v30 && value >= v29) {

					pixels[(j) * width + (i)] = cc29.getRGB();

				} else if (value < v29 && value >= v28) {

					pixels[(j) * width + (i)] = cc28.getRGB();

				} else if (value < v28 && value >= v27) {

					pixels[(j) * width + (i)] = cc27.getRGB();

				} else if (value < v27 && value >= v26) {

					pixels[(j) * width + (i)] = cc26.getRGB();

				} else if (value < v26 && value >= v25) {

					pixels[(j) * width + (i)] = cc25.getRGB();

				} else if (value < v25 && value >= v24) {

					pixels[(j) * width + (i)] = cc24.getRGB();

				} else if (value < v24 && value >= v23) {

					pixels[(j) * width + (i)] = cc23.getRGB();

				} else if (value < v23 && value >= v22) {

					pixels[(j) * width + (i)] = cc22.getRGB();

				} else if (value < v22 && value >= v21) {

					pixels[(j) * width + (i)] = cc21.getRGB();

				} else if (value < v21 && value >= v20) {

					pixels[(j) * width + (i)] = cc20.getRGB();

				} else if (value < v20 && value >= v19) {

					pixels[(j) * width + (i)] = cc19.getRGB();

				} else if (value < v19 && value >= v18) {

					pixels[(j) * width + (i)] = cc18.getRGB();

				} else if (value < v18 && value >= v17) {

					pixels[(j) * width + (i)] = cc17.getRGB();

				} else if (value < v17 && value >= v16) {

					pixels[(j) * width + (i)] = cc16.getRGB();

				} else if (value < v16 && value >= v15) {

					pixels[(j) * width + (i)] = cc15.getRGB();

				} else if (value < v15 && value >= v14) {

					pixels[(j) * width + (i)] = cc14.getRGB();

				} else if (value < v14 && value >= v13) {

					pixels[(j) * width + (i)] = cc13.getRGB();

				} else if (value < v13 && value >= v12) {

					pixels[(j) * width + (i)] = cc12.getRGB();

				} else if (value < v12 && value >= v11) {

					pixels[(j) * width + (i)] = cc11.getRGB();

				} else if (value < v11 && value >= v10) {

					pixels[(j) * width + (i)] = cc10.getRGB();

				} else if (value < v10 && value >= v9) {

					pixels[(j) * width + (i)] = cc9.getRGB();

				} else if (value < v9 && value >= v8) {

					pixels[(j) * width + (i)] = cc8.getRGB();

				} else if (value < v8 && value >= v7) {

					pixels[(j) * width + (i)] = cc7.getRGB();

				} else if (value < v7 && value >= v6) {

					pixels[(j) * width + (i)] = cc6.getRGB();

				} else if (value < v6 && value >= v5) {

					pixels[(j) * width + (i)] = cc5.getRGB();

				} else if (value < v5 && value >= v4) {

					pixels[(j) * width + (i)] = cc4.getRGB();

				} else if (value < v4 && value >= v3) {

					pixels[(j) * width + (i)] = cc3.getRGB();

				} else if (value < v3 && value >= v2) {

					pixels[(j) * width + (i)] = cc2.getRGB();

				} else if (value < v2) {

					pixels[(j) * width + (i)] = cc1.getRGB();

				} else { // something went wrong
					/*
					 * if(this.use_low_cols) { pixels[(j)*width + (i)] =
					 * cc40.getRGB(); //for some reason this fixes d2 bug } else
					 * { pixels[(j)*width + (i)] = cc1.getRGB(); }
					 */

					pixels[(j) * width + (i)] = zero.getRGB();
				}
			}
		}

		BufferedImage pixelImage = new BufferedImage(width, height,
				BufferedImage.TYPE_INT_ARGB);
		pixelImage.setRGB(0, 0, width, height, pixels, 0, width);

		/*
		 * JFrame frame = new JFrame(); JLabel label = new JLabel(new
		 * ImageIcon(pixelImage)); frame.getContentPane().add(label,
		 * BorderLayout.CENTER); frame.pack(); frame.setVisible(true);
		 */

		return pixelImage;
	}

	public void setSource_weightPrinting(int index) {

		this.source_index = index;
	}


	public String getNominal(String type, double value, boolean forImprecision) {
		String returnable = "N/A";
		double[] scale_vals;
		String[] scale;
		if (type.equals("PM2.5") || type.equals("PM2.5_AirPollutantValue")) { // OK
			if (!forImprecision) {
				double[] temp = { 0, 5, 10, 20, 50 };
				scale_vals = temp;
				String[] temp2 = { "Good", "Satisfactory", "Fair", "Poor",
						"Very Poor" };
				scale = temp2;
			} else {
				double[] temp = { 0, 1.5, 2.5, 5, 8, 12 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}

		} else if (type.equals("SO2") || type.equals("SO2_AirPollutantValue")) {// OK
			if (!forImprecision) {
				double[] temp = { 0, 20, 80, 250, 350 };
				scale_vals = temp;
				String[] temp2 = { "Good", "Satisfactory", "Fair", "Poor",
						"Very Poor" };
				scale = temp2;
			} else {
				double[] temp = { 0, 2, 5, 10, 20, 50 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}
		} else if (type.equals("PM10") || type.equals("PM10_AirPollutantValue")) {// OK
			if (!forImprecision) {
				double[] temp = { 0, 20, 50, 100, 200 };
				scale_vals = temp;
				String[] temp2 = { "Good", "Satisfactory", "Fair", "Poor",
						"Very Poor" };
				scale = temp2;
			} else {
				double[] temp = { 0, 2, 4, 8, 12, 18 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}
		} else if (type.equals("CO") || type.equals("CO_AirPollutantValue")) {// OK
			if (!forImprecision) {
				double[] temp = { 0, 4000, 8000, 20000, 30000 };
				scale_vals = temp;
				String[] temp2 = { "Good", "Satisfactory", "Fair", "Poor",
						"Very Poor" };
				scale = temp2;
			} else {
				double[] temp = { 0, 100, 200, 400, 800, 2000 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}
		} else if (type.equals("NO2") || type.equals("NO2_AirPollutantValue")) {// OK
			if (!forImprecision) {
				double[] temp = { 0, 40, 70, 150, 200 };
				scale_vals = temp;
				String[] temp2 = { "Good", "Satisfactory", "Fair", "Poor",
						"Very Poor" };
				scale = temp2;
			} else {
				double[] temp = { 0, 2, 4, 8, 12, 18 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}
		} else if (type.equals("O3") || type.equals("O3_AirPollutantValue")) {
			if (!forImprecision) {
				double[] temp = { 0, 60, 100, 140, 180 };
				scale_vals = temp;
				String[] temp2 = { "Good", "Satisfactory", "Fair", "Poor",
						"Very Poor" };
				scale = temp2;
			} else {
				double[] temp = { 0, 2, 4, 8, 12, 18 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}
		} else if (type.equals("temperature")) {// Meteorological===============================
			if (!forImprecision) {
				double[] temp = { -40, -25, -15, -5, -0.1, 10, 18, 23, 28, 35 };
				scale_vals = temp;
				String[] temp2 = { "Extreme frost", "Severe Frost", "Cold",
						"Weak frost", "Chilly", "Mild", "Warm", "Hot",
						"Remarkably hot", "Extremely hot" };
				scale = temp2;
			} else {
				double[] temp = { 0, 0.5, 1, 2, 3, 5 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}
		} else if (type.equals("windSpeed")) {// Meteorological===============================
			if (!forImprecision) {
				double[] temp = { 0, 1, 3, 7, 13, 20, 32 };
				scale_vals = temp;
				String[] temp2 = { "Calm", "Weak wind", "Moderate wind",
						"Hard wind", "Storm", "Hurricane" };
				scale = temp2;
			} else {
				double[] temp = { 0, 0.5, 1, 2, 3, 5 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}
		} else if (type.equals("rain")) {// Meteorological===============================
			if (!forImprecision) {
				double[] temp = { 0, 0.3, 0.9, 4.4, 20 };
				scale_vals = temp;
				String[] temp2 = { "Dry weather", "Light rain",
						"Moderate rain", "Heavy rain",
						"Heavy rain with flooding" };
				scale = temp2;
			} else {
				double[] temp = { 0, 0.1, 0.5, 1, 2, 5 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}

		} else if (type.equals("windDirection")) {// Meteorological===============================
			if (!forImprecision) {
				double[] temp = { 0, 22.5, 67.5, 112.5, 157.5, 202.5, 247.5,
						292.5, 337.5 };
				scale_vals = temp;
				String[] temp2 = { "North", "North-East", "East", "South-East",
						"South", "South-West", "West", "North-West", "North" };
				scale = temp2;
			} else {
				double[] temp = { 0, 2, 4, 10, 20, 40, 60 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}

		} else if (type.equals("skyCondition")) {// Meteorological===============================
			if (!forImprecision) {
				double[] temp = { 0, 2, 4, 5, 6 };
				scale_vals = temp;
				String[] temp2 = { "Clear", "Quite clear", "Partly cloudy",
						"Quite cloudy", "Cloudy" };
				scale = temp2;
			} else {
				double[] temp = { 0, 0.3, 0.6, 1.2, 2, 3, 4 };
				scale_vals = temp;
				String[] temp2 = { "Very Good", "Good", "Satisfactory", "Fair",
						"Poor", "Very Poor" };
				scale = temp2;
			}

		} else {
			if (!forImprecision) {
				double[] temp = { 10, 20, 30, 40, 50 };
				scale_vals = temp;
				String[] temp2 = { "N/A", "N/A", "N/A", "N/A", "N/A" };
				scale = temp2;
			} else {
				double[] temp = { 0, 2, 3, 5, 8, 12 };
				scale_vals = temp;
				String[] temp2 = { "N/A", "N/A", "N/A", "N/A", "N/A", "N/A" };
				scale = temp2;
			}
		}
		// lets search the correct index describing value
		int index = 0;
		boolean found = false;
		for (int i = 0; i < scale_vals.length - 1; i++) {
			if (value <= scale_vals[0]) {
				found = true;
			} else if (value > scale_vals[i] && value < scale_vals[i + 1]) {
				index = i;
				found = true;
			}
		}
		if (!found) {
			index = scale_vals.length - 1;
		}

		try {
			returnable = scale[index];
			returnable = "(" + returnable + ")";
		} catch (ArrayIndexOutOfBoundsException e) {
			String error = "Index error! (type,value,index): (" + type + ", "
					+ value + ", " + index + ")";

			System.out.println(error);
		}

		return returnable;
	}

	// this tool method edits the number of decimals of d2 double
	// int precision is the number of decimals wanted
	public double editPrecision(double value, int precision) {
		int prec = (int) Math.pow(10, precision);
		int temp = (int) (prec * value);
		double result = (double) temp / prec;
		return result;
	}

	public double calcCellArea(double lat2, double lat1, double lon2,
			double lon1) {
		double area = 0;
		final double degreeInMeters = 2 * Math.PI * 6371000 / 360;

		double dlat = (lat2 - lat1) / this.height;
		double dlon = (lon2 - lon1) / this.width;
		double lonScaler = Math.cos(2 * Math.PI * (0.5 * lat2 + 0.5 * lat1)
				/ 360); // middle lat used

		double dlatInKm = dlat * degreeInMeters / 1000;
		double dlonInKm = dlon * degreeInMeters * lonScaler / 1000;
		area = dlatInKm * dlonInKm;
		System.out.println("Cell area: " + area + "km");
		return area;
	}

	public String getUnit(String type) {
		String unit = "N/A";

		String[][] options = { { "PM2.5", "g/m3" }, { "PM10", "g/m3" },
				{ "SO2", "g/m3" }, { "NO2", "g/m3" }, { "CO", "g/m3" },
				{ "O3", "g/m3" }, { "temperature", "C" }, { "rain", "mm/h" },
				{ "windSpeed", "m/s" }, { "windDirection", "degree" } };

		for (int i = 0; i < options.length; i++) {
			if (type.equals(options[i][0]))
				unit = options[i][1];
		}

		return unit;
	}

	public double estimateBaseVar(String type, double maturity, boolean isAmodel) {

		double basev = 0;
		double multiplier = 1;
		if (isAmodel)
			multiplier = 3;

		if (type.equals("PM2.5")) {
			basev = 6;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("SO2")) {
			basev = 3;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("PM10")) {
			basev = 16;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("CO")) {
			basev = 1000;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("NO2")) {
			basev = 16;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("O3")) {
			basev = 100;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("temperature")) {
			basev = 2;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("windSpeed")) {
			basev = 2;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("rain")) {
			basev = 0.2;
			basev = basev * multiplier + basev * maturity / 2;
		} else if (type.equals("skyCondition")) {
			basev = 0.4;
			basev = basev * multiplier + basev * maturity / 2;
		} else {
			basev = 4;
			basev = basev * multiplier + basev * maturity / 2;
		}

		System.out.println("basevar estimated: " + basev
				+ "with (type/maturity/model)" + type + "/" + maturity + "/"
				+ isAmodel);
		return basev;
	}

}// Class bracket end
