package services.pescado.fusionservice.impl;

import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.junit.Test;

import services.pescado.common.PESCaDOException;
import services.pescado.common.ParameterType;
import de.fraunhofer.iosb.sosclient.OWSException;


/**
 * Tests functionality of Fusion Service.
 * 
 * @author Desiree Hilbring
 */
public class TestFS
{

	/**
	 * A demo FS request.
	 * 
	 * @throws IOException Could not retrieve the data
	 * @throws ParseException Could not parse result values
	 * @throws OWSException The SOS server returned an exception
	 */

	@Test
	public void testFS () throws IOException, ParseException, OWSException
	{
		final FusionServiceImpl fs = new FusionServiceImpl ();

		final List<ParameterType> parameter = new ArrayList<ParameterType> ();
		final ParameterType p1 = new ParameterType ();
		p1.setProperty ("timeStart");
		p1.getContent ().add ("2011-04-28T00:00:00+03:00");
		parameter.add (p1);
		final ParameterType p2 = new ParameterType ();
		p2.setProperty ("timeEnd");
		p2.getContent ().add ("2011-04-29T13:00:00+03:00");
		parameter.add (p2);
		final ParameterType p3 = new ParameterType ();
		p3.setProperty ("geoType");
		p3.getContent ().add ("area");
		parameter.add (p3);
		final ParameterType p4 = new ParameterType ();
		p4.setProperty ("geoPolygon");
		p4.getContent ().add ("60.46060328697707,24.92523193359375 60.46060328697707,24.171295166015625 60.13740224199459,24.171295166015625 60.13740224199459,24.92523193359375");
		parameter.add (p4);

		// final Parameters params = new Parameters (parameter);
		// Old version of input param for retrieveData
		final Map<String, Collection<String>> relevantAspects = new HashMap<String, Collection<String>> ();
		final Collection<String> properties = new ArrayList<String> ();
		properties.add ("http://www.pescado-project.eu/ontology/pescado.owl#O3");
		properties.add ("http://www.pescado-project.eu/ontology/pescado.owl#NO2");
		relevantAspects.put ("http://www.pescado-project.eu/ontology/pescado.owl#AirQualityDataType", properties);
		// Old version of input param for retrieveData end

		// New version of input param for retrieveData
		final List<ParameterType> racsParameters = new ArrayList<services.pescado.common.ParameterType> ();
		final Set<Entry<String, Collection<String>>> entrySet = relevantAspects.entrySet ();
		final Iterator<Entry<String, Collection<String>>> entrySetIterator = entrySet.iterator ();
		while (entrySetIterator.hasNext ())
		{
			final Entry<String, Collection<String>> racsValues = entrySetIterator.next ();
			final ParameterType racsParameter = new ParameterType ();
			final String key = racsValues.getKey ();
			racsParameter.setProperty (key);
			final Collection<String> racsValuesV = racsValues.getValue ();
			final Iterator<String> racsValuesIterator = racsValuesV.iterator ();
			String combinedRacsValue = "";
			while (racsValuesIterator.hasNext ())
			{
				final String singleRacsValue = racsValuesIterator.next ();
				combinedRacsValue = combinedRacsValue.concat (singleRacsValue);
				combinedRacsValue = combinedRacsValue.concat (" ");
			}
			racsParameter.getContent ().add (combinedRacsValue);
			racsParameters.add (racsParameter);
		}
		// New version of input param for retrieveData end

		try
		{
			String problemURI = "http://www.pescado-project.eu/ontology/pescadoInstances.owl#problem_beb1349d-e6f3-4d74-85b4-005aa5095437";

			try
			{
				fs.readCapabilities ();
			}
			catch (final Throwable t)
			{
				t.printStackTrace ();
				System.out.println ("Could not retrieve offerings from DRS. Retrying ...");

				try
				{
					Thread.sleep (10000);
				}
				catch (InterruptedException e)
				{
					System.out.println ("Sleep for retry interrupted.");
				}
				fs.readCapabilities ();
				return;
			}

			final String fusionResult = fs.fuseData (null, problemURI, parameter, racsParameters);
			System.out.println ("Result of FS: " + fusionResult);
		}
		catch (final PESCaDOException e)
		{
			e.printStackTrace ();
		}
	}

}
