
package services.pescado.knowledgebaseaccessservice;

import java.net.URL;
import javax.xml.namespace.QName;
import javax.xml.ws.Service;
import javax.xml.ws.WebEndpoint;
import javax.xml.ws.WebServiceClient;
import javax.xml.ws.WebServiceException;
import javax.xml.ws.WebServiceFeature;


/**
 * The Knowledge Base Access Service supports the access to the Knowledge Base on which the whole PESCaDO platform relies to provide user decision support. This service provides, via the KBAccess interface, access, manipulation and storage of ontologies (or ontology modules) stored in the Knowledge Base. Besides the common KBAccess interface retrieval methods, the KBQuery interface offers a generic mechanism for retrieval.
 * Some typical uses of this service are:
 * - Storing, updating or deleting available ontology modules;
 * - Retrieving (partially or fully) a stored ontology module;
 * - Getting high-level information about some ontology module, such as the list of concepts, or the list of supported properties for a given concept;
 * - Inserting in the knowledge base actual data, e.g. those retrieved from the Data Repository, which will be used to fulfil the user problem request;
 * - Querying one of more ontology modules.
 * The Knowledge Base Access Service provides the functionality through the following interfaces:
 * - ServiceCapabilities: Informs the client about the common and specific capabilities of the Knowledge Base Access Service.
 * - KBAccess: Supports the management (storage, retrieval, and deletion) of ontology modules, provides a high-level view on ontologies, and allows for adding factual knowledge (data from the data repository, inferred data) to the knowledge base.
 * - KBQuery: Providing operations to query the ontology modules contained in the knowledge base.
 * 
 * This class was generated by the JAX-WS RI.
 * JAX-WS RI 2.2.5-b01 
 * Generated source version: 2.2
 * 
 */
@WebServiceClient(name = "KnowledgeBaseAccessService", targetNamespace = "http://pescado.services/KnowledgeBaseAccessService/", wsdlLocation = "KnowledgeBaseAccessService.wsdl")
public class KnowledgeBaseAccessService_Service
    extends Service
{

    private final static URL KNOWLEDGEBASEACCESSSERVICE_WSDL_LOCATION;
    private final static WebServiceException KNOWLEDGEBASEACCESSSERVICE_EXCEPTION;
    private final static QName KNOWLEDGEBASEACCESSSERVICE_QNAME = new QName("http://pescado.services/KnowledgeBaseAccessService/", "KnowledgeBaseAccessService");

    static {
        KNOWLEDGEBASEACCESSSERVICE_WSDL_LOCATION = services.pescado.knowledgebaseaccessservice.KnowledgeBaseAccessService_Service.class.getResource("KnowledgeBaseAccessService.wsdl");
        WebServiceException e = null;
        if (KNOWLEDGEBASEACCESSSERVICE_WSDL_LOCATION == null) {
            e = new WebServiceException("Cannot find 'KnowledgeBaseAccessService.wsdl' wsdl. Place the resource correctly in the classpath.");
        }
        KNOWLEDGEBASEACCESSSERVICE_EXCEPTION = e;
    }

    public KnowledgeBaseAccessService_Service() {
        super(__getWsdlLocation(), KNOWLEDGEBASEACCESSSERVICE_QNAME);
    }

    public KnowledgeBaseAccessService_Service(WebServiceFeature... features) {
        super(__getWsdlLocation(), KNOWLEDGEBASEACCESSSERVICE_QNAME, features);
    }

    public KnowledgeBaseAccessService_Service(URL wsdlLocation) {
        super(wsdlLocation, KNOWLEDGEBASEACCESSSERVICE_QNAME);
    }

    public KnowledgeBaseAccessService_Service(URL wsdlLocation, WebServiceFeature... features) {
        super(wsdlLocation, KNOWLEDGEBASEACCESSSERVICE_QNAME, features);
    }

    public KnowledgeBaseAccessService_Service(URL wsdlLocation, QName serviceName) {
        super(wsdlLocation, serviceName);
    }

    public KnowledgeBaseAccessService_Service(URL wsdlLocation, QName serviceName, WebServiceFeature... features) {
        super(wsdlLocation, serviceName, features);
    }

    /**
     * 
     * @return
     *     returns KnowledgeBaseAccessService
     */
    @WebEndpoint(name = "KnowledgeBaseAccessServiceSOAP")
    public KnowledgeBaseAccessService getKnowledgeBaseAccessServiceSOAP() {
        return super.getPort(new QName("http://pescado.services/KnowledgeBaseAccessService/", "KnowledgeBaseAccessServiceSOAP"), KnowledgeBaseAccessService.class);
    }

    /**
     * 
     * @param features
     *     A list of {@link javax.xml.ws.WebServiceFeature} to configure on the proxy.  Supported features not in the <code>features</code> parameter will have their default values.
     * @return
     *     returns KnowledgeBaseAccessService
     */
    @WebEndpoint(name = "KnowledgeBaseAccessServiceSOAP")
    public KnowledgeBaseAccessService getKnowledgeBaseAccessServiceSOAP(WebServiceFeature... features) {
        return super.getPort(new QName("http://pescado.services/KnowledgeBaseAccessService/", "KnowledgeBaseAccessServiceSOAP"), KnowledgeBaseAccessService.class, features);
    }

    private static URL __getWsdlLocation() {
        if (KNOWLEDGEBASEACCESSSERVICE_EXCEPTION!= null) {
            throw KNOWLEDGEBASEACCESSSERVICE_EXCEPTION;
        }
        return KNOWLEDGEBASEACCESSSERVICE_WSDL_LOCATION;
    }

}
